Package: ReDaMoR
Type: Package
Title: Relational Data Modeler
Version: 1.0.0
Authors@R: c(
    person(
        "Patrice", "Godard",
        role = c("aut", "cre","cph"),
        email = "patrice.godard@gmail.com"
    ),
    person(
        "Kai", "Lin",
        role = c("ctb")
    )
    )
Maintainer: Patrice Godard <patrice.godard@gmail.com>
Description: The aim of this package is to manipulate relational
   data models in R.
   It provides functions to create, modify and export data models
   in json format.
   It also allows importing models created
   with 'MySQL Workbench' (<https://www.mysql.com/products/workbench/>).
   These functions are accessible through a graphical user
   interface made with 'shiny'.
   Constraints such as types, keys, uniqueness and mandatory fields are
   automatically checked and corrected when editing a model.
   Finally, real data can be confronted to a model to check their compatibility.
URL: https://patzaw.github.io/ReDaMoR/,
        https://github.com/patzaw/ReDaMoR/
BugReports: https://github.com/patzaw/ReDaMoR/issues
Depends: R (>= 4.1), dplyr, visNetwork
Imports: readr, shiny, shinyjs, jsonlite, DT, colourpicker, rintrojs,
        markdown, rstudioapi, crayon, utils, graphics, stats, Matrix
Suggests: knitr, rmarkdown, igraph
License: GPL-3
Encoding: UTF-8
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-18 15:14:50 UTC; pgodard
Author: Patrice Godard [aut, cre, cph],
  Kai Lin [ctb]
Repository: CRAN
Date/Publication: 2026-05-19 05:10:30 UTC
Built: R 4.5.2; ; 2026-05-19 06:37:28 UTC; unix

Package: ReDaMoR
Type: Package
Title: Relational Data Modeler
Version: 1.0.0
Authors@R: c(
    person(
        "Patrice", "Godard",
        role = c("aut", "cre","cph"),
        email = "patrice.godard@gmail.com"
    ),
    person(
        "Kai", "Lin",
        role = c("ctb")
    )
    )
Maintainer: Patrice Godard <patrice.godard@gmail.com>
Description: The aim of this package is to manipulate relational
   data models in R.
   It provides functions to create, modify and export data models
   in json format.
   It also allows importing models created
   with 'MySQL Workbench' (<https://www.mysql.com/products/workbench/>).
   These functions are accessible through a graphical user
   interface made with 'shiny'.
   Constraints such as types, keys, uniqueness and mandatory fields are
   automatically checked and corrected when editing a model.
   Finally, real data can be confronted to a model to check their compatibility.
URL: https://patzaw.github.io/ReDaMoR/,
        https://github.com/patzaw/ReDaMoR/
BugReports: https://github.com/patzaw/ReDaMoR/issues
Depends: R (>= 4.1), dplyr, visNetwork
Imports: readr, shiny, shinyjs, jsonlite, DT, colourpicker, rintrojs,
        markdown, rstudioapi, crayon, utils, graphics, stats, Matrix
Suggests: knitr, rmarkdown, igraph
License: GPL-3
Encoding: UTF-8
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-18 15:14:50 UTC; pgodard
Author: Patrice Godard [aut, cre, cph],
  Kai Lin [ctb]
Repository: CRAN
Date/Publication: 2026-05-19 05:10:30 UTC
Built: R 4.6.0; ; 2026-05-19 07:09:32 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",RelDataModel)
S3method("[<-",RelDataModel)
S3method("[[<-",RelDataModel)
S3method("names<-",RelDataModel)
S3method(c,RelDataModel)
S3method(format,RelTableModel)
S3method(get_foreign_keys,RelDataModel)
S3method(get_foreign_keys,RelTableModel)
S3method(length,RelDataModel)
S3method(length,RelTableModel)
S3method(lengths,RelTableModel)
S3method(lengths,default)
S3method(names,RelDataModel)
S3method(plot,RelDataModel)
S3method(print,RelTableModel)
export(RelDataModel)
export(RelTableModel)
export(SUPPTYPES)
export(add_field)
export(add_foreign_key)
export(add_index)
export(add_table)
export(as_type)
export(auto_layout)
export(check_foreign_keys)
export(check_types)
export(clean_autosaved_RelDataModels)
export(col_types)
export(confront_data)
export(confront_table_data)
export(conv_type_ref)
export(copy_fields)
export(correct_constraints)
export(df_to_model)
export(format_confrontation_report)
export(format_confrontation_report_md)
export(fromDBM)
export(get_foreign_keys)
export(guess_constraints)
export(identical_RelDataModel)
export(identical_RelTableModel)
export(index_table)
export(is.MatrixModel)
export(is.RelDataModel)
export(is.RelTableModel)
export(is_MM)
export(lengths)
export(list_autosaved_RelDataModel)
export(list_type_ref)
export(model_relational_data)
export(norm_type_ref)
export(order_fields)
export(order_indexes)
export(readSQLDataModel)
export(read_SQL_data_model)
export(read_json_data_model)
export(read_named_MM)
export(read_named_MM_header)
export(recover_RelDataModel)
export(remove_field)
export(remove_foreign_key)
export(remove_index)
export(remove_table)
export(rename_field)
export(rename_table)
export(set_primary_key)
export(set_unique_index)
export(toDBM)
export(update_field)
export(update_foreign_key)
export(update_table_display)
export(view_confrontation_report)
export(write_json_data_model)
import(dplyr)
import(readr)
import(rintrojs)
import(shiny)
import(visNetwork)
importFrom(DT,DTOutput)
importFrom(DT,dataTableProxy)
importFrom(DT,datatable)
importFrom(DT,formatStyle)
importFrom(DT,renderDT)
importFrom(DT,replaceData)
importFrom(Matrix,sparseMatrix)
importFrom(colourpicker,colourInput)
importFrom(crayon,bgGreen)
importFrom(crayon,bgMagenta)
importFrom(crayon,bgRed)
importFrom(crayon,bold)
importFrom(crayon,green)
importFrom(crayon,underline)
importFrom(crayon,white)
importFrom(crayon,yellow)
importFrom(graphics,plot)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,write_json)
importFrom(markdown,renderMarkdown)
importFrom(rstudioapi,viewer)
importFrom(shinyjs,disable)
importFrom(shinyjs,enable)
importFrom(shinyjs,useShinyjs)
importFrom(stats,complete.cases)
importFrom(stats,rbeta)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(utils,object.size)
importFrom(utils,packageName)

## Read the model ----
hpo_from_sql <- read_SQL_data_model(
  system.file("examples/HPO-model.sql", package = "ReDaMoR")
)
## Confront to data ----
confrontation_report <- confront_data(
  hpo_from_sql,
  path = list.files(
    system.file("examples/HPO-subset", package = "ReDaMoR"),
    full.names = TRUE
  ),
  returnData = TRUE
)

## Read the model ----
hpo_model <- read_json_data_model(
  system.file("examples/HPO-model.json", package = "ReDaMoR")
)
## Confront to data ----
confrontation_report <- confront_data(
  hpo_model,
  path = list.files(
    system.file("examples/HPO-subset", package = "ReDaMoR"),
    full.names = TRUE
  ),
  returnData = TRUE
)

## Read the model ----
hpo_from_sql <- read_SQL_data_model(
  system.file("examples/HPO-model.sql", package = "ReDaMoR")
)
## Confront to data ----
confrontation_report <- confront_data(
  hpo_from_sql,
  path = list.files(
    system.file("examples/HPO-subset", package = "ReDaMoR"),
    full.names = TRUE
  ),
  verbose = FALSE,
  returnData = TRUE
)
## Show the report in console ----
format_confrontation_report(confrontation_report) |> cat()
## Format the report using markdown ----
format_confrontation_report_md(confrontation_report) |> cat()

## Read data files ----
to_read <- list.files(
  system.file("examples/HPO-subset", package = "ReDaMoR"),
  full.names = TRUE
)
hpo_tables <- list()
for (f in to_read) {
  hpo_tables[[sub("[.]txt$", "", basename(f))]] <- readr::read_tsv(f)
}
## Build the model from a list of data frames ----
new_model <- df_to_model(
  list = names(hpo_tables),
  envir = as.environment(hpo_tables)
)
## Guess constraints and auto layout ----
new_model <- guess_constraints(new_model, data = hpo_tables) |>
  auto_layout(lengthMultiplier = 250)

## Plot the model ----
new_model |>
  plot()

## ----setup, include = FALSE---------------------------------------------------
library(knitr)
library(ReDaMoR)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
cranRef <- function(x){
  sprintf(
    "[%s](https://CRAN.R-project.org/package=%s): %s",
    x, x, packageDescription(x)$Title
  )
}

## ----eval=FALSE---------------------------------------------------------------
# install.packages("ReDaMoR")

## ----eval=FALSE---------------------------------------------------------------
# devtools::install_github("patzaw/ReDaMoR")

## ----eval=FALSE---------------------------------------------------------------
# library(ReDaMoR)
# m <- model_relational_data()

## ----eval=FALSE---------------------------------------------------------------
# m <- model_relational_data(recover_RelDataModel())

## -----------------------------------------------------------------------------
hpo_model <- read_json_data_model(
  system.file("examples/HPO-model.json", package="ReDaMoR")
)
plot(hpo_model)

## -----------------------------------------------------------------------------
## Edit the model
# m <- model_relational_data(hpo_model)

## -----------------------------------------------------------------------------
confrontation_report <- confront_data(
  hpo_model,
  path=list.files(
    system.file("examples/HPO-subset", package="ReDaMoR"),
    full.names=TRUE
  ),
  returnData=TRUE
)

## ----results='asis'-----------------------------------------------------------
# view_confrontation_report(confrontation_report) # Use RStudio viewer
format_confrontation_report_md(
  confrontation_report,
  title="Example: Confrontation with original data",
  level=1, numbered=FALSE
) |>
  cat()

## -----------------------------------------------------------------------------
hpo_tables <- confrontation_report$data

## ----results='asis'-----------------------------------------------------------
hpo_tables$HPO_diseases <- hpo_tables$HPO_diseases |> slice(1:100)
hpo_tables$HPO_synonyms[1:10, "synonym"] <- NA
hpo_tables$HPO_hp <- hpo_tables$HPO_hp |> mutate(level=as.character(level))
confront_data(hpo_model, hpo_tables, verbose=FALSE) |>
  format_confrontation_report_md(
    title="Example: Confrontation with altered data",
    level=1, numbered=FALSE
  ) |>
    cat()

## -----------------------------------------------------------------------------
hpo_tables <- confrontation_report$data

## Build the model from a list of data frames ----
new_model <- df_to_model(
   list=names(hpo_tables), envir=as.environment(hpo_tables)
)
## Guess constraints and auto layout ----
new_model <- guess_constraints(new_model, data = hpo_tables) |>
   auto_layout(lengthMultiplier=250)

## Plot the model ----
new_model |>
   plot()

## -----------------------------------------------------------------------------
# model_relational_data(new_model)

## -----------------------------------------------------------------------------
ge_model <- read_json_data_model(
  system.file("examples/GE-model.json", package="ReDaMoR")
)
plot(ge_model)

#' @import dplyr
#' @import shiny
#' @import rintrojs
#' @import visNetwork
#' @import readr
#' @importFrom Matrix sparseMatrix
#' @importFrom utils packageName object.size
#' @importFrom stats rbeta runif complete.cases setNames
#' @importFrom graphics plot
#' @importFrom markdown renderMarkdown
#' @importFrom rstudioapi viewer
#' @importFrom jsonlite fromJSON write_json
#' @importFrom colourpicker colourInput
#' @importFrom DT DTOutput renderDT datatable dataTableProxy replaceData
#' @importFrom DT formatStyle
#' @importFrom crayon bgGreen white bgRed bgMagenta green yellow bold underline
#' @importFrom shinyjs useShinyjs enable disable
#'
NULL
# igraph is required by the auto_layout() function but not imported

###############################################################################@
#' Identify if a file is in MatrixMarket text format
#'
#' @param file the file to read
#'
#' @return A logical. If FALSE, the first line of the file is returned as
#' an attribute named "r1": `attr(is_MM, "r1")`
#'
#' @export
#'
is_MM <- function(file) {
  r1 <- readLines(file, n = 1)
  sms <- "%%MatrixMarket"
  toRet <- substr(r1, 1, nchar(sms)) == sms
  if (!toRet) {
    attr(toRet, "r1") <- r1
  }
  return(toRet)
}


###############################################################################@
#' Read the header of a named sparse matrix in MatrixMarket text format
#'
#' @param file the file to read
#' @param guess_max the number of lines to read to find the header.
#' (4 should be sufficient. Default: 20)
#'
#' @return A list with the following fields:
#' - rownames: a character vector with the matrix row names
#' - colnames: a character vector with the matrix column names
#' - rows: the number of matrix rows
#' - columns: the number of matrix columns
#' - values: the number of values in the matrix
#' - header_length: the number of lines in the header
#'
#' @export
#'
read_named_MM_header <- function(file, guess_max = 20) {
  h <- readLines(file, n = guess_max)
  rn <- strsplit(
    sub("^%%Rownames: ", "", h[grep("^%%Rownames: ", h)]),
    split = "\t"
  )[[1]]
  cn <- strsplit(
    sub("^%%Colnames: ", "", h[grep("^%%Colnames: ", h)]),
    split = "\t"
  )[[1]]
  hl <- grep("^%%", h, invert = TRUE)[1]
  n <- as.numeric(strsplit(
    h[hl],
    split = " "
  )[[1]])
  nr <- n[1]
  nc <- n[2]
  nv <- n[3]
  stopifnot(nr == length(rn), nc == length(cn))
  return(list(
    rownames = rn,
    colnames = cn,
    rows = nr,
    columns = nc,
    values = nv,
    header_length = hl
  ))
}

###############################################################################@
#' Read a named sparse matrix in MatrixMarket text format
#'
#' @param file the file to read
#' @param skip the number of records to skip (default: 0)
#' @param n_max the maximum number of records to read (default: Inf)
#' @param class the class of object to return. By default a "dgCMatrix".
#' If "tibble" is chosen, the sparse matrix is returned as a tibble with
#' 3 columns: i (row index), j (column index) and x (values) and
#' an "header" attribute containing the matrix rownames and colnames.
#' @param guess_max the number of lines to read to find the header.
#' (see [read_named_MM_header()])
#'
#' @return By default a dgCMatrix.
#' If the "tibble" class is chosen, the sparse matrix is returned as
#' a tibble with 3 columns: i (row index), j (column index) and x (values) and
#' an "header" attribute containing the matrix rownames and colnames.
#'
#' @export
#'
read_named_MM <- function(
  file,
  skip = 0,
  n_max = Inf,
  class = c("dgCMatrix", "tibble"),
  guess_max = 20
) {
  class <- match.arg(class)
  h <- read_named_MM_header(file, guess_max = guess_max)

  ## The following chunck is about dealing with many rownames or colnames
  ## in gz compressed files
  ovcs <- Sys.getenv("VROOM_CONNECTION_SIZE")
  on.exit(Sys.setenv("VROOM_CONNECTION_SIZE" = ovcs))
  Sys.setenv(
    "VROOM_CONNECTION_SIZE" = max(
      c(
        utils::object.size(h$rownames),
        utils::object.size(h$colnames),
        as.numeric(ovcs)
      ),
      na.rm = TRUE
    )
  )
  ##

  td <- readr::read_delim(
    file,
    delim = "\t",
    col_names = c("i", "j", "x"),
    col_types = "iin",
    skip = h$header_length + skip,
    n_max = n_max
  )
  if (class == "dgCMatrix") {
    toRet <- Matrix::sparseMatrix(
      i = td$i,
      j = td$j,
      x = td$x,
      dimnames = list(h$rownames[1:max(td$i)], h$colnames[1:max(td$j)])
    )
  }
  if (class == "tibble") {
    toRet <- td
    attr(toRet, "header") <- h
  }
  return(toRet)
}

###############################################################################@
#' Format confrontation report for printing in console
#'
#' @param cr the confrontation report from [confront_data]
#' @param title a character with a single value corresponding to the report
#' title (e.g. database/model name)
#'
#' @example inst/examples/ex_confront.R
#'
#' @export
#'
format_confrontation_report <- function(
  cr,
  title = "Model"
) {
  ## Helpers ----
  successTag <- function(s) {
    if (s) {
      crayon::bgGreen(crayon::white('SUCCESS'))
    } else {
      crayon::bgRed(crayon::white('FAILURE'))
    }
  }
  messageTag <- function(m) {
    crayon::bgMagenta(crayon::white(m))
  }

  toRet <- c()

  ## Title ----
  toRet <- c(toRet, crayon::bold(crayon::underline(title)))

  ## Global success ----
  toRet <- c(toRet, successTag(cr$success))

  ## Global config ----
  toRet <- c(
    toRet,
    '',
    crayon::bold('Check configuration'),
    sprintf('   - Optional checks: %s', paste(cr$checks, collapse = ", ")),
    sprintf('   - Maximum number of records: %s', cr$n_max),
    ''
  )

  ## Missing tables ----
  if (length(cr$missingTables) > 0) {
    toRet <- c(
      toRet,
      crayon::bold('Missing tables'),
      'The following tables are missing:',
      paste("   -", cr$missingTables),
      ''
    )
  }

  ## Supplementary tables ----
  if (length(cr$suppTables) > 0) {
    toRet <- c(
      toRet,
      crayon::bold('Not supported tables'),
      'The following tables are not supported by the model:',
      '',
      paste("   -", cr$suppTables),
      ''
    )
  }

  ## Table information ----
  ttoRet <- unlist(lapply(
    names(cr$constraints),
    function(tn) {
      tcr <- cr$constraints[[tn]]
      toRet <- c()

      ### Missing fields ----
      if (length(tcr$missingFields) > 0) {
        toRet <- c(
          toRet,
          crayon::underline('Missing fields'),
          'The following fields are missing:',
          paste("   -", tcr$missingFields),
          ''
        )
      }

      ### Supplementary fields ----
      if (length(tcr$suppFields) > 0) {
        toRet <- c(
          toRet,
          crayon::underline('Not supported fields'),
          'The following fields are not supported by the model:',
          paste("   -", tcr$suppFields),
          ''
        )
      }

      ### Field information ----
      ftoRet <- unlist(lapply(
        names(tcr$fields),
        function(fn) {
          s <- tcr$fields[[fn]]$success
          m <- tcr$fields[[fn]]$message
          if (!s || (!is.null(m) && !is.na(m) && m != "")) {
            return(paste0(
              '   - ',
              fn,
              ': ',
              successTag(s),
              ' ',
              messageTag(m)
            ))
          } else {
            return(NULL)
          }
        }
      ))
      if (length(ftoRet) > 0) {
        toRet <- c(
          toRet,
          crayon::underline('Field issues or warnings'),
          ftoRet,
          ''
        )
      }

      ### Index information ----
      if (length(tcr$indexes) > 0) {
        itoRet <- unlist(lapply(
          1:length(tcr$indexes),
          function(i) {
            idx <- paste(cr$model[[tn]]$indexes[[i]]$fields, collapse = "+")
            if (cr$model[[tn]]$indexes[[i]]$unique) {
              idx <- paste(idx, '(unique)')
            }
            s <- tcr$indexes[[i]]$success
            m <- tcr$indexes[[i]]$message
            if (!s || (!is.null(m) && !is.na(m) && m != "")) {
              return(paste0(
                '   - ',
                idx,
                ': ',
                successTag(s),
                ' ',
                messageTag(m)
              ))
            } else {
              return(NULL)
            }
          }
        ))
        if (length(itoRet) > 0) {
          toRet <- c(
            toRet,
            crayon::underline('Index issues or warnings'),
            itoRet,
            ''
          )
        }
      }

      ### Foreign key information ----
      if (length(tcr$foreignKey) > 0) {
        fktoRet <- unlist(lapply(
          1:length(tcr$foreignKey),
          function(i) {
            fk <- paste(
              cr$model[[tn]]$foreignKeys[[i]]$key$from,
              cr$model[[tn]]$foreignKeys[[i]]$key$to,
              sep = "->"
            ) |>
              paste(collapse = " + ")
            fk <- paste0(
              cr$model[[tn]]$foreignKeys[[i]]$refTable,
              ' [',
              fk,
              ']'
            )
            s <- tcr$foreignKey[[i]]$success
            m <- tcr$foreignKey[[i]]$message
            if (!s || (!is.null(m) && !is.na(m) && m != "")) {
              return(paste0(
                '   - ',
                fk,
                ': ',
                successTag(s),
                ' ',
                messageTag(m)
              ))
            } else {
              return(NULL)
            }
          }
        ))
        if (length(fktoRet) > 0) {
          toRet <- c(
            toRet,
            crayon::underline('Foreign keys issues or warnings'),
            fktoRet,
            ''
          )
        }
      }

      ### Results if anything to show ----
      if (length(toRet) > 0 || !tcr$success) {
        toRet <- c(
          crayon::bold(tn),
          successTag(tcr$success),
          toRet,
          ''
        )
      }
    }
  ))

  if (length(ttoRet) > 0) {
    toRet <- c(
      toRet,
      ttoRet
    )
  }

  ## Concatenate the result ----
  return(paste(toRet, collapse = "\n"))
}

###############################################################################@
#' Format confrontation report in markdown format
#'
#' @param cr the confrontation report from [confront_data]
#' @param title a character with a single value corresponding to the report
#' @param level rmarkdown level in document hierarchy (default:0 ==> highest).
#' It should be an integer between 0 and 4.
#' @param numbered a logical. If TRUE (default) the sections are part of
#' document numbering.
#' @param bgSuccess background color for SUCCESS
#' @param txSuccess text color for SUCCESS
#' @param bgFailure background color for FAILURE
#' @param txFailure text color for FAILURE
#' @param bgMessage background color for a warning message
#' @param txMessage text color for a warning message
#'
#' @example inst/examples/ex_confront.R
#'
#' @export
#'
format_confrontation_report_md <- function(
  cr,
  title = "Model",
  level = 0,
  numbered = TRUE,
  bgSuccess = "green",
  txSuccess = "black",
  bgFailure = "red",
  txFailure = "white",
  bgMessage = "#FFBB33",
  txMessage = "white"
) {
  level <- round(level, digits = 0)
  stopifnot(level >= 0, level <= 4)

  ## Helpers ----
  successTag <- function(s) {
    sprintf(
      paste0(
        '<span ',
        'style="background-color:%s; color:%s; padding:2px;"',
        '>%s</span>'
      ),
      ifelse(s, bgSuccess, bgFailure),
      ifelse(s, txSuccess, txFailure),
      ifelse(s, 'SUCCESS', 'FAILURE')
    )
  }
  messageTag <- function(m) {
    sprintf(
      '<span style="background-color:%s; color:%s; padding:2px;">%s</span>',
      bgMessage,
      txMessage,
      m
    )
  }

  toRet <- c()

  ## Title ----
  toRet <- c(toRet, '', sprintf('# %s', title[1]), '')

  ## Global success ----
  toRet <- c(toRet, successTag(cr$success))

  ## Global config ----
  toRet <- c(
    toRet,
    '',
    '## Check configuration',
    '',
    sprintf('- **Optional checks**: %s', paste(cr$checks, collapse = ", ")),
    sprintf('- **Maximum number of records**: %s', cr$n_max),
    ''
  )

  ## Missing tables ----
  if (length(cr$missingTables) > 0) {
    toRet <- c(
      toRet,
      '',
      '## Missing tables',
      '',
      'The following tables are missing:',
      '',
      paste("-", cr$missingTables),
      ''
    )
  }

  ## Supplementary tables ----
  if (length(cr$suppTables) > 0) {
    toRet <- c(
      toRet,
      '',
      '## Not supported tables',
      '',
      'The following tables are not supported by the model:',
      '',
      paste("-", cr$suppTables),
      ''
    )
  }

  ## Table information ----
  ttoRet <- unlist(lapply(
    names(cr$constraints),
    function(tn) {
      tcr <- cr$constraints[[tn]]
      toRet <- c()

      ### Missing fields ----
      if (length(tcr$missingFields) > 0) {
        toRet <- c(
          toRet,
          '',
          '### Missing fields',
          '',
          'The following fields are missing:',
          '',
          paste(paste("-", tcr$missingFields)),
          ''
        )
      }

      ### Supplementary fields ----
      if (length(tcr$suppFields) > 0) {
        toRet <- c(
          toRet,
          '',
          '### Not supported fields',
          '',
          'The following fields are not supported by the model:',
          '',
          paste(paste("-", tcr$suppFields)),
          ''
        )
      }

      ### Field information ----
      ftoRet <- unlist(lapply(
        names(tcr$fields),
        function(fn) {
          s <- tcr$fields[[fn]]$success
          m <- tcr$fields[[fn]]$message
          if (!s || (!is.null(m) && !is.na(m) && m != "")) {
            return(paste0(
              '- ',
              fn,
              ': ',
              successTag(s),
              ' ',
              messageTag(m)
            ))
          } else {
            return(NULL)
          }
        }
      ))
      if (length(ftoRet) > 0) {
        toRet <- c(
          toRet,
          '',
          '### Field issues or warnings',
          '',
          ftoRet,
          ''
        )
      }

      ### Index information ----
      if (length(tcr$indexes) > 0) {
        itoRet <- unlist(lapply(
          1:length(tcr$indexes),
          function(i) {
            idx <- paste(cr$model[[tn]]$indexes[[i]]$fields, collapse = "+")
            if (cr$model[[tn]]$indexes[[i]]$unique) {
              idx <- paste(idx, '(unique)')
            }
            s <- tcr$indexes[[i]]$success
            m <- tcr$indexes[[i]]$message
            if (!s || (!is.null(m) && !is.na(m) && m != "")) {
              return(paste0(
                '- ',
                idx,
                ': ',
                successTag(s),
                ' ',
                messageTag(m)
              ))
            } else {
              return(NULL)
            }
          }
        ))
        if (length(itoRet) > 0) {
          toRet <- c(
            toRet,
            '',
            '### Index issues or warnings',
            '',
            itoRet,
            ''
          )
        }
      }

      ### Foreign key information ----
      if (length(tcr$foreignKey) > 0) {
        fktoRet <- unlist(lapply(
          1:length(tcr$foreignKey),
          function(i) {
            fk <- paste(
              cr$model[[tn]]$foreignKeys[[i]]$key$from,
              cr$model[[tn]]$foreignKeys[[i]]$key$to,
              sep = "->"
            ) |>
              paste(collapse = " + ")
            fk <- paste0(
              cr$model[[tn]]$foreignKeys[[i]]$refTable,
              ' [',
              fk,
              ']'
            )
            s <- tcr$foreignKey[[i]]$success
            m <- tcr$foreignKey[[i]]$message
            if (!s || (!is.null(m) && !is.na(m) && m != "")) {
              return(paste0(
                '- ',
                fk,
                ': ',
                successTag(s),
                ' ',
                messageTag(m)
              ))
            } else {
              return(NULL)
            }
          }
        ))
        if (length(fktoRet) > 0) {
          toRet <- c(
            toRet,
            '',
            '### Foreign keys issues or warnings',
            '',
            fktoRet,
            ''
          )
        }
      }

      ### Results if anything to show ----
      if (length(toRet) > 0 || !tcr$success) {
        toRet <- c(
          '',
          sprintf('## %s', tn),
          '',
          successTag(tcr$success),
          '',
          toRet,
          ''
        )
      }
    }
  ))

  if (length(ttoRet) > 0) {
    toRet <- c(
      toRet,
      ttoRet
    )
  }

  ## Set document level ----
  toRet <- sub(
    "^[#]",
    paste0("#", paste(rep("#", level), collapse = "")),
    toRet
  )
  ## Unnumber if requested ----
  if (!numbered) {
    h <- grep("^[#]", toRet)
    toRet[h] <- paste(toRet[h], "{.unlisted .unnumbered}")
  }

  ## Concatenate the result ----
  toRet <- paste(toRet, collapse = "\n")
  return(toRet)
}

###############################################################################@
#' View confrontation report in rstudio viewer
#'
#' @param cr the confrontation report from [confront_data]
#' @param ... additional params for
#' the [format_confrontation_report_md()] function
#'
#' @export
#'
view_confrontation_report <- function(cr, ...) {
  tf <- tempfile(fileext = ".html")
  on.exit(rm(tf))
  md <- format_confrontation_report_md(cr, ...)
  md <- markdown::renderMarkdown(text = md)
  writeLines(md, tf)
  rstudioapi::viewer(tf)
}

###############################################################################@
buildUi <- function(fromR) {
  shiny::addResourcePath(
    "www",
    system.file("www", package = utils::packageName())
  )
  shiny::addResourcePath(
    "doc",
    system.file("doc", package = utils::packageName())
  )

  shiny::shinyUI(shiny::fluidPage(
    title = "ReDaMoR",
    windowTitle = "ReDaMoR",
    id = "MainApp",

    ## Settings ----
    shinyjs::useShinyjs(),
    rintrojs::introjsUI(),

    ## HEAD ----
    shiny::tags$head(
      shiny::tags$link(
        rel = "icon",
        href = 'www/ReDaMoR.png'
      ),
      shiny::tags$link(
        rel = "stylesheet",
        type = "text/css",
        href = "www/cerulean.css"
      ),
      shiny::tags$link(
        rel = "stylesheet",
        type = "text/css",
        href = "www/defChanges.css"
      ),
      shiny::tags$link(
        rel = "stylesheet",
        type = "text/css",
        href = "www/appElements.css"
      ),
      shiny::tags$script(src = 'www/interactions.js'),
      if (fromR) NULL else shiny::tags$script(src = 'www/fromWeb.js')
    ),

    ## Main menu ----
    shiny::fluidRow(
      id = "mainMenu",
      shiny::column(
        11,
        id = "mainLColumn",
        shiny::div(
          id = "mainDone",
          if (fromR) {
            shiny::actionButton(
              "done",
              list(
                shiny::icon("check", "fa-2x", verify_fa = FALSE),
                "Done"
              )
            ) |>
              shiny::div(title = "Return the model in R session")
          } else {
            shiny::img(src = 'www/ReDaMoR.png', id = "mainLogo")
          }
        ),
        shiny::div(
          class = "mainButton",
          shiny::actionButton(
            "import",
            list(
              shiny::icon("file-import", "fa-2x", verify_fa = FALSE),
              "Import"
            )
          ),
          shiny::actionButton(
            "export",
            list(
              "Export",
              shiny::icon("file-export", "fa-2x", verify_fa = FALSE)
            )
          )
        ),
        shiny::div(
          class = "mainButton",
          shiny::actionButton(
            "undo",
            list("Undo", shiny::icon("undo", "fa-2x", verify_fa = FALSE))
          ),
          shiny::actionButton(
            "redo",
            list(shiny::icon("redo", "fa-2x", verify_fa = FALSE), "Redo")
          ),
          title = "Undo (Ctrl+Z) / Redo (Ctrl+Shift+Z)"
        ),
        shiny::div(
          class = "mainButton",
          shiny::actionButton(
            "addTable",
            "Add table",
            icon = shiny::icon("plus-square", "fa-2x", verify_fa = FALSE)
          )
        ),
        shiny::div(
          class = "mainButton",
          shiny::uiOutput("modelSummary")
        ),
        shiny::div(
          class = "mainButton",
          shiny::actionButton(
            "doc",
            "",
            icon = shiny::icon("question-circle", "fa-2x", verify_fa = FALSE)
          ) |>
            shiny::div(title = "Help tour")
        )
      ),
      shiny::column(
        1,
        id = "mainRColumn",
        shiny::tags$a(
          "About ReDaMoR",
          href = "doc/ReDaMoR.html",
          target = "_blank",
          id = "about"
        ),
        paste0(
          "(version ",
          utils::packageDescription(utils::packageName())$Version,
          ")"
        )
      )
    ),

    shiny::fluidRow(
      ## Model view ----
      shiny::column(
        7,
        shiny::fluidRow(
          id = "viewMenu",
          shiny::column(
            8,
            id = "findTableDiv",
            shiny::selectInput(
              "findTable",
              label = NULL,
              choices = NULL,
              multiple = TRUE,
              selected = NULL,
              width = "100%"
            )
          ),
          shiny::column(
            4,
            id = "viewButtons",
            shiny::div(
              class = "viewButton",
              shiny::actionButton(
                "selectAll",
                label = NULL,
                icon = shiny::icon(
                  "object-group",
                  "fa-2x",
                  verify_fa = FALSE
                ),
                class = "shrunkenButton"
              ) |>
                shiny::div(title = "Select all tables")
            ),
            shiny::div(
              class = "viewButton",
              shiny::actionButton(
                "autoLayout",
                label = NULL,
                icon = shiny::icon(
                  "pencil-ruler",
                  "fa-2x",
                  verify_fa = FALSE
                ),
                class = "shrunkenButton"
              ) |>
                shiny::div(title = "Auto layout the model")
            ),
            shiny::div(
              class = "viewButton",
              shiny::actionButton(
                "fitNet",
                label = NULL,
                icon = shiny::icon(
                  "vector-square",
                  "fa-2x",
                  verify_fa = FALSE
                ),
                class = "shrunkenButton"
              ) |>
                shiny::div(title = "Fit model")
            )
          )
        ),
        shiny::fluidRow(
          id = "modelFrame",
          visNetwork::visNetworkOutput(
            "modelNet",
            height = "75vh",
            width = "100%"
          )
        )
      ),

      shiny::column(
        5,

        ## Edit menu ----
        shiny::div(
          id = "editMenu",
          shiny::uiOutput(
            "setTableColor",
            class = "editMenuSection"
          ),
          shiny::uiOutput(
            "dupTablesInput",
            class = "editMenuSection"
          ),
          shiny::uiOutput(
            "rmTablesInput",
            class = "editMenuSection"
          ),
          shiny::uiOutput(
            "addFKInput",
            class = "editMenuSection"
          ),
          shiny::uiOutput(
            "editFKInput",
            class = "editMenuSection"
          ),
          shiny::uiOutput(
            "rmFKInput",
            class = "editMenuSection"
          )
        ),

        ## Edit table ----
        shiny::uiOutput("editTable")
      )
    )
  ))
}

###############################################################################@
buildServer <- function(
  modelInput,
  fromR,
  bcko,
  defaultColor,
  availableColors,
  example,
  forceIntro
) {
  rintrosteps <- jsonlite::fromJSON(system.file(
    "Documentation/rintrojs-steps.json",
    package = utils::packageName()
  )) |>
    lapply(
      function(x) {
        toRet <- dplyr::as_tibble(x) |>
          dplyr::select("element", "intro")
        if (!fromR) {
          toRet <- toRet |>
            dplyr::filter(is.na(.data$element) | .data$element != "#done")
        }
        return(toRet)
      }
    )

  function(input, output, session) {
    #########################################################################@
    ## Help tour ----
    #########################################################################@

    context <- shiny::reactiveValues(
      x = "main"
    )
    shiny::observe(
      if (length(selection$tables) == 0 && length(selection$fk) == 0) {
        context$x <- "main"
      } else {
        ctxt <- c()
        if (length(selection$tables) > 0) {
          ctxt <- c(ctxt, "tables")
          if (length(selection$tables) == 2) {
            ctxt <- c(ctxt, "twoTables")
          }
          if (length(selection$tables) == 1) {
            ctxt <- c(ctxt, "oneTable")
            if (nrow(model$x[[selection$tables]]$fields) > 0) {
              ctxt <- c(ctxt, "withFields")
            }
            if (length(model$x[[selection$tables]]$indexes) > 0) {
              ctxt <- c(ctxt, "withIndexes")
            }
            if (length(input$fieldTable_rows_selected) > 0) {
              ctxt <- c(ctxt, "selectedField")
            }
            if (length(input$indexTable_rows_selected) > 0) {
              ctxt <- c(ctxt, "selectedIndex")
            }
          }
        }
        if (length(selection$fk) > 0) {
          ctxt <- c(ctxt, "withFK")
          if (length(selection$fk) == 1) {
            ctxt <- c(ctxt, "oneFK")
          }
        }
        context$x <- ctxt
      }
    )
    shiny::observeEvent(input$doc, {
      docx <- do.call(rbind, rintrosteps[context$x])
      rintrojs::introjs(session, options = list(steps = docx))
    })
    shiny::observeEvent(input$docImp, {
      docx <- do.call(rbind, rintrosteps[c("Import")])
      rintrojs::introjs(session, options = list(steps = docx))
    })

    if (forceIntro) {
      rintrojs::introjs(session, options = list(steps = rintrosteps[["main"]]))
    }

    #########################################################################@
    ## Settings ----
    #########################################################################@

    settings <- shiny::reactiveValues()
    settings$defaultColor <- defaultColor
    settings$availableColors <- availableColors
    shiny::observe({
      settings$availableColors <- unique(c(
        settings$defaultColor,
        lapply(model$x, function(x) x$display$color) |>
          unlist() |>
          setdiff(NA),
        shiny::isolate(settings$availableColors)
      ))
    })

    #########################################################################@
    ## The model ----
    #########################################################################@

    model <- shiny::reactiveValues(
      x = modelInput, # The current model
      vn = modelToVn(modelInput), # VisNet representation
      new = NULL, # A new model to add in history
      history = list(modelInput), # The model history
      current = 1, # The position of current model in history
      toImport = NULL, # Model to import from file
      merged = NULL, # merge: c(x, toImport)
      table = NULL, # The table to edit
      indexTable = dplyr::tibble(
        # Indexes of the table to edit
        fields = character(),
        unique = logical()
      ),
      fieldTable = dplyr::tibble(
        # Fields of the table to edit
        name = character(),
        type = character(),
        nullable = logical(),
        unique = logical(),
        comment = character()
      )
    )
    shiny::observe(
      model$vn <- modelToVn(
        model$x,
        color = shiny::isolate(settings$defaultColor)
      )
    )

    replot <- shiny::reactiveValues(
      x = 1 # Used for triggering model re-plot
    )
    selection <- shiny::reactiveValues(
      release = 0, # Used for refreshing the visNetwork
      tables = NULL, # Selected tables
      fk = NULL, # Selected foreign keys
      fromVN = FALSE # Used for refreshing the visNetwork
    )

    #########################################################################@
    ## Notifications ----
    #########################################################################@

    warningMessage <- shiny::reactiveValues(
      n = 0,
      message = NULL
    )
    sendWarning <- function(message) {
      warningMessage$n <- shiny::isolate(warningMessage$n) + 1
      warningMessage$message <- message
    }
    shiny::observe({
      shiny::req(warningMessage$n > 0)
      shiny::showNotification(
        shiny::isolate(warningMessage$message),
        duration = 5,
        type = "warning"
      )
    })

    errorMessage <- shiny::reactiveValues(
      n = 0,
      message = NULL
    )
    sendError <- function(message) {
      errorMessage$n <- shiny::isolate(errorMessage$n) + 1
      errorMessage$message <- message
    }
    shiny::observe({
      shiny::req(errorMessage$n > 0)
      shiny::showNotification(
        shiny::isolate(errorMessage$message),
        duration = 5,
        type = "error"
      )
    })

    #########################################################################@
    ## Model view ----
    #########################################################################@

    shiny::observe({
      m <- model$x
      shiny::updateSelectInput(
        session,
        "findTable",
        choices = sort(names(m)),
        selected = intersect(shiny::isolate(selection$tables), names(m))
      )
    })
    shiny::observe({
      selTables <- selection$tables
      shiny::updateSelectInput(
        session,
        "findTable",
        selected = as.character(selTables)
      )
    })
    shiny::observe({
      selTables <- sort(input$findTable)
      shiny::req(!identical(selTables, shiny::isolate(selection$tables)))
      mn <- shiny::isolate(model$vn)
      selFK <- mn$edges |>
        dplyr::filter(
          .data$from %in% selTables | .data$to %in% selTables
        ) |>
        dplyr::pull("id")
      selection$fromVN <- FALSE
      if (length(selTables) == 0) {
        selection$tables <- NULL
      } else {
        selection$tables <- selTables
      }
      selection$release <- shiny::isolate(selection$release) + 1
    })
    shiny::observeEvent(input$selectAll, {
      m <- shiny::isolate(model$x)
      shiny::updateSelectInput(
        session,
        "findTable",
        selected = sort(names(m))
      )
    })
    shiny::observeEvent(input$autoLayout, {
      m <- auto_layout(shiny::isolate(model$x), force = TRUE)
      model$new <- m
    })

    output$modelSummary <- shiny::renderUI({
      m <- model$x
      mn <- model$vn
      nt <- length(m)
      nfk <- nrow(mn$edges)
      np <- lapply(m, function(x) nrow(x$fields)) |> unlist() |> sum()
      shiny::tagList(
        shiny::tags$strong("Tables:"),
        nt,
        "-",
        shiny::tags$strong("Foreign keys:"),
        nfk,
        "-",
        shiny::tags$strong("Fields:"),
        np
      )
    })

    #########################################################################@
    ## Model network ----
    #########################################################################@

    output$modelNet <- visNetwork::renderVisNetwork({
      replot$x
      selection$fromVN <- FALSE
      selection$tables <- NULL
      selection$fk <- NULL
      plot(
        shiny::isolate(model$x),
        color = shiny::isolate(settings$defaultColor)
      ) |>
        visNetwork::visEvents(release = "releaseVn")
    })

    shiny::observe({
      input$modelNet_release
      selection$release <- shiny::isolate(selection$release) + 1
    })
    shiny::observe({
      selection$release
      visNetwork::visNetworkProxy("modelNet") |>
        visNetwork::visGetSelectedNodes()
      visNetwork::visNetworkProxy("modelNet") |>
        visNetwork::visGetSelectedEdges()
      visNetwork::visNetworkProxy("modelNet") |>
        visNetwork::visGetNodes()
    })

    modelNet_selectedNodes <- shiny::reactive({
      input$modelNet_selectedNodes
    })
    shiny::observe({
      selTables <- intersect(
        modelNet_selectedNodes(),
        names(model$x)
      ) |>
        sort()
      selection$fromVN <- TRUE
      if (length(selTables) == 0) {
        selection$tables <- NULL
      } else {
        selection$tables <- selTables
      }
    })

    modelNet_selectedEdges <- shiny::reactive({
      input$modelNet_selectedEdges
    })
    shiny::observe({
      selFK <- intersect(
        modelNet_selectedEdges(),
        model$vn$edges$id
      ) |>
        sort()
      selection$fromVN <- TRUE
      if (length(selFK) == 0) {
        selection$fk <- NULL
      } else {
        selection$fk <- selFK
      }
    })

    shiny::observeEvent(input$fitNet, {
      visNetwork::visNetworkProxy("modelNet") |> visNetwork::visFit()
    })

    #########################################################################@
    ## Import model ----
    #########################################################################@

    shiny::observeEvent(input$import, {
      shiny::showModal(shiny::modalDialog(
        title = "Import",
        shiny::uiOutput("import"),
        size = "l",
        easyClose = TRUE
      ))
    })

    output$import <- shiny::renderUI({
      list(
        shiny::fluidRow(
          shiny::column(
            6,
            class = "leftBox",
            shiny::fileInput(
              "impModel",
              "Choose an sql or a json file",
              multiple = FALSE,
              accept = c(".sql", ".json", ".sql.gz", ".json.gz"),
              width = "100%"
            )
          ),
          shiny::column(
            2,
            class = "leftBox",
            shiny::uiOutput("exampleModel")
          ),
          shiny::column(
            4,
            class = "rightBox",
            shiny::div(
              class = "mainButton",
              shiny::actionButton(
                "docImp",
                "",
                icon = shiny::icon(
                  "question-circle",
                  "fa-2x",
                  verify_fa = FALSE
                )
              ),
              title = "Help tour"
            )
          )
        ),
        shiny::fluidRow(shiny::uiOutput("impModel"))
      )
    })

    ### Import preview ----
    output$impModel <- shiny::renderUI({
      mi <- model$toImport
      shiny::req(!is.null(mi))
      if (!is.RelDataModel(mi)) {
        list(shiny::p(mi, class = "errorMessage"))
      } else {
        list(
          shiny::div(
            visNetwork::visNetworkOutput(
              "impModelNet",
              height = "65vh",
              width = "100%"
            ),
            id = "impModelFrame"
          ),
          shiny::uiOutput("impMessage")
        )
      }
    })

    output$impModelNet <- visNetwork::renderVisNetwork({
      mi <- model$toImport
      shiny::req(mi)
      plot(mi, color = shiny::isolate(settings$defaultColor))
    })

    shiny::observe({
      mi <- model$toImport
      shiny::req(mi)
      m <- shiny::isolate(model$x)
      mm <- try(c(m, mi), silent = TRUE)
      model$merged <- mm
    })

    output$impMessage <- shiny::renderUI({
      mm <- model$merged
      if (is.RelDataModel(mm)) {
        shiny::actionButton(
          "importValidate",
          list(
            shiny::icon("file-import", "fa-2x", verify_fa = FALSE),
            "Merge with current model"
          )
        )
      } else {
        list(shiny::p(mm, class = "errorMessage"))
      }
    })

    ### From model ----
    shiny::observe({
      fi <- input$impModel
      shiny::req(fi)
      fiext <- regexpr(
        "(\\.[[:alnum:]]+)(\\.gz)?$",
        fi$name,
        ignore.case = TRUE
      )
      fiext <- substr(
        fi$name,
        fiext,
        fiext + attr(fiext, "match.length") - 1
      ) |>
        tolower()
      mi <- NULL
      if (fiext %in% c(".sql", ".sql.gz")) {
        mi <- suppressWarnings(try(
          read_SQL_data_model(fi$datapath),
          silent = TRUE
        ))
      }
      if (fiext %in% c(".json", ".json.gz")) {
        mi <- suppressWarnings(try(
          read_json_data_model(fi$datapath),
          silent = TRUE
        ))
      }
      if (is.RelDataModel(mi)) {
        model$toImport <- auto_layout(mi, lengthMultiplier = 45 * length(mi))
      } else {
        model$toImport <- mi
      }
    })

    ### From example ----
    output$exampleModel <- shiny::renderUI({
      shiny::req(file.exists(example))
      m <- try(read_json_data_model(example), silent = TRUE)
      if (!is.RelDataModel(m)) {
        m <- try(read_SQL_data_model(example), silent = TRUE)
      }
      shiny::req(m)
      shiny::actionLink("exampleLink", label = "Try an example")
    })
    shiny::observeEvent(input$exampleLink, {
      shiny::req(file.exists(example))
      mi <- try(read_json_data_model(example), silent = TRUE)
      if (!is.RelDataModel(mi)) {
        mi <- try(read_SQL_data_model(example), silent = TRUE)
      }
      shiny::req(mi)
      model$toImport <- auto_layout(mi, lengthMultiplier = 45 * length(mi))
    })

    ### Validate import ----
    shiny::observe({
      shiny::req(input$importValidate)
      mm <- shiny::isolate(model$merged)
      shiny::req(mm)

      cmn <- shiny::isolate(model$vn)
      if (
        !is.null(cmn$nodes) > 0 &&
          nrow(cmn$nodes) > 0 &&
          all(!is.na(cmn$nodes$x)) &&
          all(!is.na(cmn$nodes$y))
      ) {
        toReplot <- FALSE
        cmxrange <- c(min(cmn$nodes$x), max(cmn$nodes$x))
        cmyrange <- c(min(cmn$nodes$x), max(cmn$nodes$y))
        shiny::req(shiny::isolate(model$toImport))
        tin <- modelToVn(
          shiny::isolate(model$toImport),
          color = shiny::isolate(settings$defaultColor)
        )
        if (any(is.na(tin$nodes$x)) || any(is.na(tin$nodes$y))) {
          tin$nodes$x <- stats::runif(
            nrow(tin$nodes),
            min = cmxrange[1],
            max = cmxrange[2]
          )
          tin$nodes$y <- stats::runif(
            nrow(tin$nodes),
            min = cmyrange[1],
            max = cmyrange[2]
          )
        }
        tixrange <- c(min(tin$nodes$x), max(tin$nodes$x))
        tiyrange <- c(min(tin$nodes$x), max(tin$nodes$y))
        xshift <- cmxrange[2] - tixrange[1]
        yshift <- cmyrange[2] - tiyrange[1]
        tin$nodes$x <- tin$nodes$x + xshift
        tin$nodes$y <- tin$nodes$y + yshift
        mm <- lapply(
          mm,
          function(n) {
            i <- which(tin$nodes$id == n$tableName)
            if (length(i) == 1) {
              n$display$x = tin$nodes$x[i]
              n$display$y = tin$nodes$y[i]
            }
            return(n)
          }
        )
        class(mm) <- c("RelDataModel", "list")
      }
      model$new <- mm
      model$merged <- NULL
      model$toImport <- NULL
      shiny::removeModal()
    })

    #########################################################################@
    ## Add table ----
    #########################################################################@

    shiny::observeEvent(input$addTable, {
      shiny::showModal(shiny::modalDialog(
        title = "Add table",
        shiny::div(
          shiny::fluidRow(
            shiny::column(
              10,
              shiny::textInput(
                "newTableName",
                label = NULL,
                width = "100%",
                placeholder = "Table name"
              )
            ),
            shiny::column(
              2,
              shiny::actionButton("confirmAddTable", "Add")
            )
          ),
          shiny::fluidRow(shiny::uiOutput("newTableExists"))
        ),
        size = "s",
        easyClose = TRUE
      ))
    })

    output$newTableExists <- shiny::renderUI({
      ntn <- input$newTableName
      m <- shiny::isolate(model$x)
      if (ntn %in% names(m)) {
        shiny::p("Table name already used", class = "errorMessage")
      } else {
        list()
      }
    })

    shiny::observe({
      ntn <- input$newTableName
      m <- shiny::isolate(model$x)
      if (is.null(ntn) || ntn == "" || ntn %in% names(m)) {
        shinyjs::disable("confirmAddTable")
      } else {
        shinyjs::enable("confirmAddTable")
      }
    })

    shiny::observe({
      shiny::req(input$confirmAddTable)
      tn <- shiny::isolate(input$newTableName)
      m <- shiny::isolate(model$x)
      xs <- 100
      ys <- 100
      if (!is.null(tn) && tn != "" && !tn %in% names(m)) {
        m <- add_table(m, newTable = tn)
        pr <- stats::rbeta(1, 9, 1)
        pa <- stats::runif(1, 0, 2 * pi)
        m <- m |>
          update_table_display(
            tn,
            px = xs * pr * cos(pa),
            py = ys * pr * sin(pa)
          )
        model$new <- m
        shiny::removeModal()
      }
    })

    #########################################################################@
    ## Edit table ----
    #########################################################################@

    cpfields <- shiny::reactiveValues(
      table = NULL,
      fields = NULL
    )

    shiny::observe({
      selTables <- selection$tables
      m <- model$x
      if (!is.RelDataModel(m)) {
        model$table <- NULL
      } else {
        if (length(selTables) == 1 && selTables %in% names(m)) {
          model$table <- m[[selTables]]
        } else {
          model$table <- NULL
        }
      }
    })

    output$editTable <- shiny::renderUI({
      selt <- selection$tables
      shiny::req(length(selt) == 1)
      mt <- shiny::isolate(model$x[[selt]])
      shiny::req(mt)
      selTable <- mt$tableName
      shiny::div(
        shiny::fluidRow(
          shiny::column(8, shiny::h3(selTable)),
          shiny::column(
            4,
            class = "rightBox",
            shiny::actionButton("renameTable", "Rename") |>
              shiny::div(title = "Rename the table")
          )
        ),
        shiny::uiOutput("tableCommentUI"),
        shiny::tags$hr(class = "editSeparator"),
        shiny::uiOutput("fields"),
        shiny::tags$hr(class = "editSeparator"),
        shiny::uiOutput("primaryKey"),
        shiny::tags$hr(class = "editSeparator"),
        shiny::uiOutput("indexes")
      )
    })

    ### Rename table ----
    shiny::observeEvent(input$renameTable, {
      shiny::showModal(shiny::modalDialog(
        title = "Rename table",
        shiny::div(
          shiny::fluidRow(
            shiny::column(
              10,
              shiny::textInput(
                "tableNewName",
                label = NULL,
                width = "100%",
                placeholder = "Table new name"
              )
            ),
            shiny::column(
              2,
              shiny::actionButton("confirmRenameTable", "Rename")
            )
          ),
          shiny::fluidRow(shiny::uiOutput("newNameExists"))
        ),
        size = "s",
        easyClose = TRUE
      ))
    })
    output$newNameExists <- shiny::renderUI({
      ntn <- input$tableNewName
      m <- shiny::isolate(model$x)
      if (ntn %in% names(m)) {
        shiny::p("Table name already used", class = "errorMessage")
      } else {
        list()
      }
    })
    shiny::observe({
      ntn <- input$tableNewName
      m <- shiny::isolate(model$x)
      if (is.null(ntn) || ntn == "" || ntn %in% names(m)) {
        shinyjs::disable("confirmRenameTable")
      } else {
        shinyjs::enable("confirmRenameTable")
      }
    })
    shiny::observe({
      shiny::req(input$confirmRenameTable)
      tn <- shiny::isolate(input$tableNewName)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      m <- shiny::isolate(model$x)
      if (!is.null(tn) && tn != "" && !tn %in% names(m)) {
        m <- rename_table(m, old = selTable, new = tn)
        model$new <- m
        shiny::removeModal()
      }
    })

    ### Table comment ----
    output$tableCommentUI <- shiny::renderUI({
      selt <- selection$tables
      shiny::req(length(selt) == 1)
      mt <- shiny::isolate(model$x[[selt]])
      shiny::req(mt)
      selTable <- mt$tableName
      shiny::fluidRow(
        shiny::column(
          8,
          shiny::textAreaInput(
            "tableComment",
            label = NULL,
            value = ifelse(
              is.na(mt$display$comment),
              "",
              mt$display$comment
            ),
            width = "100%",
            placeholder = "Table description"
          )
        ),
        shiny::column(
          4,
          shiny::uiOutput("refreshCommentH", class = "updateHighlight"),
          shiny::actionButton(
            "refreshComment",
            label = NULL,
            icon = shiny::icon("check", "fa-1x", verify_fa = FALSE),
            class = "disabled"
          ) |>
            shiny::div(title = "Update table comment", class = "iblock"),
          class = "rightBox"
        )
      )
    })
    shiny::observe({
      input$refreshComment
      ntn <- input$tableComment
      shiny::req(length(ntn) > 0)
      mt <- model$table
      shiny::req(mt)
      selTable <- mt$tableName
      cc <- mt$display$comment
      ntn <- ifelse(is.na(ntn), "", ntn)
      cc <- ifelse(is.na(cc), "", cc)
      if (ntn == cc) {
        shinyjs::disable("refreshComment")
      } else {
        shinyjs::enable("refreshComment")
      }
    })
    output$refreshCommentH <- shiny::renderUI({
      input$refreshComment
      ntn <- input$tableComment
      shiny::req(length(ntn) > 0)
      mt <- model$table
      shiny::req(mt)
      selTable <- mt$tableName
      cc <- mt$display$comment
      ntn <- ifelse(is.na(ntn), "", ntn)
      cc <- ifelse(is.na(cc), "", cc)
      shiny::req(ntn != cc)
      shiny::icon("arrow-right", "fa-1x", verify_fa = FALSE)
    })
    shiny::observe({
      input$refreshComment
      nc <- shiny::isolate(input$tableComment)
      shiny::req(!is.na(nc))
      if (nc == "") {
        nc <- NA
      }
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      cc <- mt$display$comment
      m <- shiny::isolate(model$x)
      if (
        xor(is.na(nc), is.na(cc)) ||
          (!is.na(nc) && !is.na(cc) && nc != cc)
      ) {
        model$new <- m |>
          update_table_display(
            tableName = selTable,
            comment = nc
          )
      }
    })

    ### Table fields ----
    output$fields <- shiny::renderUI({
      selt <- selection$tables
      shiny::req(length(selt) == 1)
      mt <- shiny::isolate(model$x[[selt]])
      shiny::req(mt)
      selTable <- mt$tableName
      list(
        shiny::fluidRow(
          shiny::column(4, shiny::h4("Fields")),
          shiny::column(
            8,
            shiny::uiOutput("updateFieldDiv", inline = TRUE),
            if (is.MatrixModel(mt)) {
              NULL
            } else {
              shiny::tagList(
                shiny::actionButton(
                  "addField",
                  label = "",
                  icon = shiny::icon(
                    "plus-square",
                    "fa-1x",
                    verify_fa = FALSE
                  ),
                  class = "shrunkenButton"
                ) |>
                  shiny::div(
                    title = "Add a new field",
                    class = "iblock"
                  ),
                shiny::actionButton(
                  "selectAllFields",
                  label = "",
                  icon = shiny::icon(
                    "check-double",
                    "fa-1x",
                    verify_fa = FALSE
                  ),
                  class = "shrunkenButton"
                ) |>
                  shiny::div(
                    title = "Select all fields",
                    class = "iblock"
                  ),
                shiny::uiOutput("pasteFieldDiv", inline = TRUE)
              )
            },
            class = "rightBox"
          )
        ),
        shiny::fluidRow(
          shiny::column(12, DT::DTOutput("fieldTable"))
        ),
        shiny::uiOutput("fieldCommentDisplay")
      )
    })
    output$fieldTable <- DT::renderDT({
      selt <- selection$tables
      shiny::req(length(selt) == 1)
      mt <- shiny::isolate(model$x[[selt]])
      # shiny::req(mt)
      # selTable <- mt$tableName
      shiny::isolate(model$fieldTable) |>
        dplyr::select(-"comment") |>
        DT::datatable(
          rownames = FALSE,
          filter = "top",
          selection = list(mode = 'multiple', selected = c(), target = 'row'),
          options = list(
            dom = "tip",
            columnDefs = list(
              # list(targets=c(0), visible=TRUE, width='4%'),
              list(targets = c(0), visible = TRUE, width = '24%'),
              list(targets = c(1), visible = TRUE, width = '24%'),
              list(targets = c(2), visible = TRUE, width = '24%')
            ),
            ordering = FALSE
          )
        )
    })
    proxyFieldTable <- DT::dataTableProxy("fieldTable")
    shiny::observe({
      mt <- model$table
      shiny::req(mt)
      selTable <- mt$tableName
      model$fieldTable <- mt$fields
    })
    shiny::observe({
      DT::replaceData(
        proxyFieldTable,
        data = model$fieldTable |> dplyr::select(-"comment"),
        clearSelection = "none",
        resetPaging = FALSE,
        rownames = FALSE
      )
    })
    #### Display field comment
    output$fieldCommentDisplay <- shiny::renderUI({
      seli <- input$fieldTable_rows_selected
      shiny::req(length(seli) == 1)
      shiny::req(nrow(model$fieldTable) > 0)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      shiny::req(nrow(mt$fields) > 0)
      shiny::req(seli >= 1 & seli <= nrow(mt$fields))
      shiny::p(mt$fields$comment[seli])
    })
    #### Paste fields ----
    output$pasteFieldDiv <- shiny::renderUI({
      shiny::req(cpfields$table)
      shiny::req(cpfields$fields)
      mt <- model$table
      shiny::req(mt)
      shiny::req(!is.MatrixModel(mt))
      selTable <- mt$tableName
      fn <- mt$fields$name
      shiny::req(selTable != cpfields$table)
      shiny::req(!any(cpfields$fields %in% fn))
      shiny::div(
        shiny::actionButton(
          "pasteFields",
          label = "",
          icon = shiny::icon(
            "paste",
            "fa-1x",
            verify_fa = FALSE
          ),
          class = "shrunkenButton"
        ) |>
          shiny::div(
            title = "Paste fields (Ctrl+v)",
            class = "iblock"
          ),
        class = "iblock"
      )
    })
    observe({
      shiny::req(input$pasteFields > 0)
      from <- shiny::isolate(cpfields$table)
      cpf <- shiny::isolate(cpfields$fields)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      shiny::req(!is.MatrixModel(mt))
      to <- mt$tableName
      fn <- mt$fields$name
      shiny::req(to != from)
      shiny::req(!any(cpf %in% fn))
      m <- shiny::isolate(model$x)
      m <- copy_fields(m, from = from, to = to, fields = cpf)
      model$new <- m
    })
    #### Modify fields ----
    output$updateFieldDiv <- shiny::renderUI({
      seli <- input$fieldTable_rows_selected
      shiny::req(length(seli) >= 1)
      shiny::req(nrow(model$fieldTable) > 0)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      shiny::req(nrow(mt$fields) > 0)
      shiny::req(all(seli >= 1 & seli <= nrow(mt$fields)))
      shiny::div(
        if (is.MatrixModel(mt)) {
          NULL
        } else {
          shiny::actionButton(
            "copyFields",
            label = "",
            icon = shiny::icon(
              "copy",
              "fa-1x",
              verify_fa = FALSE
            ),
            class = "shrunkenButton"
          ) |>
            shiny::div(
              title = "Copy fields (Ctrl+c)",
              class = "iblock"
            )
        },
        shiny::actionButton(
          "moveFieldUp",
          label = "",
          icon = shiny::icon(
            "arrow-alt-circle-up",
            "fa-1x",
            verify_fa = FALSE
          ),
          class = "shrunkenButton"
        ) |>
          shiny::div(
            title = "Move up",
            class = "iblock"
          ),
        shiny::actionButton(
          "moveFieldDown",
          label = "",
          icon = shiny::icon(
            "arrow-alt-circle-down",
            "fa-1x",
            verify_fa = FALSE
          ),
          class = "shrunkenButton"
        ) |>
          shiny::div(
            title = "Move down",
            class = "iblock"
          ),
        if (is.MatrixModel(mt) & length(seli) > 1) {
          NULL
        } else {
          shiny::actionButton(
            "updateField",
            label = "",
            icon = shiny::icon("edit", "fa-1x", verify_fa = FALSE),
            class = "shrunkenButton"
          ) |>
            shiny::div(
              title = "Edit field properties",
              class = "iblock"
            )
        },
        if (is.MatrixModel(mt)) {
          NULL
        } else {
          shiny::actionButton(
            "removeField",
            label = "",
            icon = shiny::icon("minus-square", "fa-1x", verify_fa = FALSE),
            class = "shrunkenButton"
          ) |>
            shiny::div(title = "Remove field", class = "iblock")
        },
        class = "iblock"
      )
    })
    #### Copy fields ----
    shiny::observe({
      shiny::req(input$copyFields > 0)
      seli <- shiny::isolate(input$fieldTable_rows_selected)
      shiny::req(length(seli) >= 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      shiny::req(nrow(mt$fields) > 0)
      fn <- mt$fields$name[seli]
      cpfields$table <- selTable
      cpfields$fields <- fn
    })
    #### Select all fields ----
    shiny::observe({
      shiny::req(input$selectAllFields > 0)
      seli <- shiny::isolate(input$fieldTable_rows_selected)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      if (length(seli) < nrow(mt$fields)) {
        DT::selectRows(proxyFieldTable, selected = 1:nrow(mt$fields))
      } else {
        DT::selectRows(proxyFieldTable, selected = NULL)
      }
    })
    #### Remove field ----
    shiny::observe({
      shiny::req(input$removeField > 0)
      seli <- shiny::isolate(input$fieldTable_rows_selected)
      shiny::req(length(seli) >= 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      m <- shiny::isolate(model$x)
      shiny::req(nrow(mt$fields) > 0)
      fn <- mt$fields$name[seli]
      for (i in 1:length(fn)) {
        m <- try(
          m |>
            remove_field(
              tableName = selTable,
              fieldName = fn[i]
            ),
          silent = TRUE
        )
      }
      if (is.RelDataModel(m)) {
        model$new <- m
      } else {
        shiny::showModal(shiny::modalDialog(
          title = "Unable to remove field",
          shiny::p(
            shiny::HTML(paste(
              sprintf("<u>%s</u> is used in foreign key(s).", fn),
              "Remove the foreign key(s) before removing this fields.",
              sep = "<br>"
            )),
            class = "errorMessage"
          ),
          size = "s",
          easyClose = TRUE
        ))
      }
    })
    #### Move field ----
    shiny::observe({
      shiny::req(input$moveFieldUp > 0)
      seli <- shiny::isolate(input$fieldTable_rows_selected)
      shiny::req(length(seli) >= 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      m <- shiny::isolate(model$x)
      shiny::req(nrow(mt$fields) > 0)
      shiny::req(min(seli) > 1)
      alli <- 1:nrow(m[[selTable]]$fields)
      o <- rep(NA, length(alli))
      o[seli - 1] <- seli
      o[which(is.na(o))] <- setdiff(alli, seli)
      m <- m |>
        order_fields(
          tableName = selTable,
          order = o
        )
      DT::selectPage(proxyFieldTable, page = ((min(seli) - 2) %/% 10) + 1)
      DT::selectRows(proxyFieldTable, selected = seli - 1)
      model$new <- m
    })
    shiny::observe({
      shiny::req(input$moveFieldDown > 0)
      seli <- shiny::isolate(input$fieldTable_rows_selected)
      shiny::req(length(seli) >= 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      m <- shiny::isolate(model$x)
      shiny::req(nrow(mt$fields) > 0)
      alli <- 1:nrow(m[[selTable]]$fields)
      shiny::req(max(seli) < length(alli))
      o <- rep(NA, length(alli))
      o[seli + 1] <- seli
      o[which(is.na(o))] <- setdiff(alli, seli)
      m <- m |>
        order_fields(
          tableName = selTable,
          order = o
        )
      DT::selectPage(proxyFieldTable, page = (max(seli) %/% 10) + 1)
      DT::selectRows(proxyFieldTable, selected = seli + 1)
      model$new <- m
    })
    #### Add field ----
    shiny::observeEvent(input$addField, {
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      fields <- mt$fields
      shiny::showModal(shiny::modalDialog(
        title = "Add field",
        shiny::fluidRow(
          shiny::column(
            10,
            shiny::textInput(
              "newFieldName",
              label = "Field",
              placeholder = "Field name",
              width = "100%"
            ),
            shiny::uiOutput("existingNewField"),
            shiny::selectInput(
              "newFieldType",
              label = "Type",
              choices = if (nrow(fields) == 0) {
                c(SUPPTYPES, "row", "column")
              } else {
                SUPPTYPES
              },
              selected = NULL,
              multiple = FALSE
            ),
            shiny::checkboxInput(
              "newFieldNullable",
              label = "Nullable?",
              value = FALSE,
            ),
            shiny::checkboxInput(
              "newFieldUnique",
              label = "Unique?",
              value = FALSE,
            ),
            shiny::textAreaInput(
              "newFieldComment",
              label = "Comment",
              placeholder = paste(
                'Field description.',
                'If base64, you should start with the extension of the',
                'file between {} (e.g. "{png}", "{html}", "{zip}"...)'
              ),
              width = "100%"
            )
          ),
          shiny::column(
            2,
            shiny::actionButton("confirmAddField", "Add")
          )
        ),
        size = "s",
        easyClose = TRUE
      ))
    })
    shiny::observe({
      nfn <- input$newFieldName
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      fields <- mt$fields
      if (
        length(nfn) == 0 ||
          is.na(nfn) ||
          nfn == "" ||
          nfn %in% fields$name
      ) {
        shinyjs::disable("confirmAddField")
      } else {
        shinyjs::enable("confirmAddField")
      }
    })
    output$existingNewField <- shiny::renderUI({
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      fields <- mt$fields
      shiny::req(input$newFieldName %in% fields$name)
      shiny::p("Field name already used", class = "errorMessage")
    })
    shiny::observeEvent(input$confirmAddField, {
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      fields <- mt$fields
      nfn <- shiny::isolate(input$newFieldName)
      shiny::req(
        !is.null(nfn) &&
          nfn != "" &&
          !nfn %in% fields$name
      )
      nm <- shiny::isolate(model$x) |>
        add_field(
          tableName = selTable,
          name = nfn,
          type = shiny::isolate(input$newFieldType),
          nullable = shiny::isolate(input$newFieldNullable),
          unique = shiny::isolate(input$newFieldUnique),
          comment = as.character(shiny::isolate(input$newFieldComment))
        )
      if (!identical(nm, shiny::isolate(model$x))) {
        model$new <- nm
      }
      shiny::removeModal()
    })

    #### Update field ----
    shiny::observeEvent(input$updateField, {
      seli <- shiny::isolate(input$fieldTable_rows_selected)
      shiny::req(length(seli) >= 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      fields <- mt$fields
      shiny::req(nrow(fields) > 0)
      shiny::showModal(shiny::modalDialog(
        title = paste0("Update field", ifelse(length(seli) > 1, "s", "")),
        shiny::fluidRow(
          shiny::column(
            10,
            if (length(seli) > 1) {
              NULL
            } else {
              shiny::textInput(
                "fieldName",
                label = "Field",
                value = fields$name[seli],
                placeholder = "Field name",
                width = "100%"
              )
            },
            shiny::uiOutput("existingField"),
            shiny::selectInput(
              "fieldType",
              label = "Type",
              choices = if (any(fields$type[seli] %in% c("row", "column"))) {
                c("row", "column")
              } else {
                if (is.MatrixModel(mt)) {
                  setdiff(SUPPTYPES, "base64")
                } else {
                  if (length(seli) == 1) {
                    SUPPTYPES
                  } else {
                    c("", SUPPTYPES)
                  }
                }
              },
              selected = if (length(unique(fields$type[seli])) > 1) {
                NULL
              } else {
                fields$type[seli][1]
              },
              multiple = FALSE
            ),
            if (any(fields$type[seli] %in% c("row", "column"))) {
              NULL
            } else {
              shiny::checkboxInput(
                "fieldNullable",
                label = "Nullable?",
                value = any(fields$nullable[seli]),
              )
            },
            if (any(fields$type[seli] %in% c("row", "column"))) {
              NULL
            } else {
              shiny::checkboxInput(
                "fieldUnique",
                label = "Unique?",
                value = all(fields$unique[seli]),
              )
            },
            if (length(seli) > 1) {
              NULL
            } else {
              shiny::textAreaInput(
                "fieldComment",
                label = "Comment",
                value = fields$comment[seli],
                placeholder = paste(
                  'Field description.',
                  'If base64, you should start',
                  'with the extension of the file',
                  'between {} (e.g. "{png}", "{html}", "{zip}"...)'
                ),
                width = "100%"
              )
            },
            shiny::uiOutput("updateFieldError")
          ),
          shiny::column(
            2,
            shiny::actionButton("confirmUpdateField", "Update")
          )
        ),
        size = "s",
        easyClose = TRUE
      ))
    })
    shiny::observe({
      nfn <- input$fieldName
      seli <- shiny::isolate(input$fieldTable_rows_selected)
      shiny::req(length(seli) == 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      fields <- mt$fields
      if (
        length(nfn) == 0 ||
          is.na(nfn) ||
          nfn == "" ||
          nfn %in% fields$name[-seli]
      ) {
        shinyjs::disable("confirmUpdateField")
      } else {
        shinyjs::enable("confirmUpdateField")
      }
    })
    output$existingField <- shiny::renderUI({
      nfn <- input$fieldName
      seli <- shiny::isolate(input$fieldTable_rows_selected)
      shiny::req(length(seli) == 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      fields <- mt$fields
      shiny::req(nfn %in% fields$name[-seli])
      shiny::p("Field name already used", class = "errorMessage")
    })
    updateField <- shiny::reactiveValues(error = NULL)
    shiny::observeEvent(input$confirmUpdateField, {
      seli <- shiny::isolate(input$fieldTable_rows_selected)
      shiny::req(length(seli) >= 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      fields <- mt$fields
      nm <- shiny::isolate(model$x)
      if (length(seli) == 1) {
        nfn <- shiny::isolate(input$fieldName)
        shiny::req(
          !is.null(nfn) &&
            nfn != "" &&
            !nfn %in% fields$name[-seli]
        )
        if (nfn != fields$name[seli]) {
          nm <- nm |>
            rename_field(
              tableName = selTable,
              current = fields$name[seli],
              new = nfn
            )
        }
        nm <- try(
          nm |>
            update_field(
              tableName = selTable,
              fieldName = nfn,
              type = shiny::isolate(input$fieldType),
              nullable = shiny::isolate(input$fieldNullable),
              unique = shiny::isolate(input$fieldUnique),
              comment = as.character(shiny::isolate(input$fieldComment))
            ),
          silent = TRUE
        )
      } else {
        for (i in seli) {
          mtype <- shiny::isolate(input$fieldType)
          if (mtype == "") {
            mtype <- NULL
          }
          nm <- try(
            nm |>
              update_field(
                tableName = selTable,
                fieldName = fields$name[i],
                type = mtype,
                nullable = shiny::isolate(input$fieldNullable),
                unique = shiny::isolate(input$fieldUnique)
              ),
            silent = TRUE
          )
        }
      }
      if (is.RelDataModel(nm)) {
        updateField$error <- NULL
        if (!identical(nm, shiny::isolate(model$x))) {
          model$new <- nm
          sendWarning(paste(
            "Uniqueness or mandatory contraints may not have been",
            "changed if they are required for",
            "existing indexes or foreign keys."
          ))
        } else {
          sendError(paste(
            "The field has not been modified because",
            "uniqueness or mandatory constraints are required for",
            "existing indexes or foreign keys."
          ))
        }
        shiny::removeModal()
      } else {
        updateField$error <- nm
      }
    })
    output$updateFieldError <- shiny::renderUI({
      e <- updateField$error
      shiny::req(!is.null(e))
      shiny::p(e, class = "errorMessage")
    })

    ### Table primary key ----
    output$primaryKey <- shiny::renderUI({
      selt <- selection$tables
      shiny::req(length(selt) == 1)
      mt <- model$x[[selt]]
      shiny::req(mt)
      selTable <- mt$tableName
      fnames <- mt$fields$name
      shiny::req(fnames)
      shiny::fluidRow(
        shiny::column(3, shiny::h4("Primary key")),
        shiny::column(
          5,
          shiny::selectInput(
            "primaryKey",
            label = NULL,
            choices = fnames,
            selected = mt$primaryKey,
            multiple = TRUE,
            width = "100%"
          )
        ),
        shiny::column(
          4,
          shiny::uiOutput("refreshPKH", class = "updateHighlight"),
          shiny::actionButton(
            "refreshPrimaryKey",
            label = NULL,
            icon = shiny::icon("check", "fa-1x", verify_fa = FALSE),
            class = "disabled"
          ) |>
            shiny::div(
              title = "Update table primary key",
              class = "iblock"
            ),
          class = "rightBox"
        )
      )
    })
    shiny::observe({
      input$refreshPrimaryKey
      npk <- input$primaryKey
      mt <- model$table
      shiny::req(mt)
      selTable <- mt$tableName
      cpk <- mt$primaryKey
      if (length(cpk) != length(npk) || any(sort(cpk) != sort(npk))) {
        shinyjs::enable("refreshPrimaryKey")
      } else {
        shinyjs::disable("refreshPrimaryKey")
      }
      if (is.MatrixModel(mt)) {
        shinyjs::disable("primaryKey")
      } else {
        shinyjs::enable("primaryKey")
      }
    })
    output$refreshPKH <- shiny::renderUI({
      input$refreshPrimaryKey
      npk <- input$primaryKey
      mt <- model$table
      shiny::req(mt)
      selTable <- mt$tableName
      cpk <- mt$primaryKey
      shiny::req(
        length(cpk) != length(npk) || any(sort(cpk) != sort(npk))
      )
      shiny::icon("arrow-right", "fa-1x", verify_fa = FALSE)
    })
    shiny::observe({
      shiny::req(input$refreshPrimaryKey > 0)
      npk <- shiny::isolate(input$primaryKey)
      # npk <- input$primaryKey
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      m <- shiny::isolate(model$x)
      cpk <- mt$primaryKey
      if (length(cpk) != length(npk) || any(sort(cpk) != sort(npk))) {
        model$new <- m |>
          set_primary_key(tableName = selTable, fieldNames = npk)
        sendWarning(paste(
          "Some indexes, uniqueness and mandatory constraints",
          "may have been added to support the primary key."
        ))
      }
    })

    ### Table indexes ----
    output$indexes <- shiny::renderUI({
      selt <- selection$tables
      shiny::req(length(selt) == 1)
      mt <- model$x[[selt]]
      shiny::req(mt)
      selTable <- mt$tableName
      fnames <- mt$fields$name
      shiny::req(fnames)
      list(
        shiny::fluidRow(
          shiny::column(6, shiny::h4("Indexes")),
          shiny::column(
            6,
            shiny::uiOutput("updateIndexDiv", inline = TRUE),
            shiny::actionButton(
              "addIndex",
              label = "",
              icon = shiny::icon("plus-square", "fa-1x", verify_fa = FALSE),
              class = "shrunkenButton"
            ) |>
              shiny::div(title = "Add an index", class = "iblock"),
            class = "rightBox"
          )
        ),
        shiny::fluidRow(
          shiny::column(12, DT::DTOutput("indexTable"))
        )
      )
    })
    output$indexTable <- DT::renderDT({
      selt <- selection$tables
      shiny::req(length(selt) == 1)
      shiny::isolate(model$indexTable) |>
        DT::datatable(
          rownames = TRUE,
          filter = "top",
          selection = list(mode = 'multiple', selected = c(), target = 'row'),
          options = list(
            dom = "tip",
            columnDefs = list(
              list(targets = c(0), visible = TRUE, width = '10%'),
              list(targets = c(1), visible = TRUE, width = '70%'),
              list(targets = c(2), visible = TRUE, width = '20%')
            )
          )
        )
    })
    proxyIndexTable <- DT::dataTableProxy("indexTable")
    shiny::observe({
      mt <- model$table
      shiny::req(mt)
      selTable <- mt$tableName
      if (length(mt$indexes) > 0) {
        indexTable <- mt$indexes |>
          lapply(function(x) {
            dplyr::tibble(
              Fields = sprintf("[%s]", paste(x$fields, collapse = "], [")),
              Unique = x$unique
            )
          })
        model$indexTable <- do.call(rbind, indexTable)
      } else {
        model$indexTable <- dplyr::tibble(
          Fields = character(),
          Unique = logical()
        )
      }
    })
    shiny::observe({
      DT::replaceData(
        proxyIndexTable,
        data = model$indexTable,
        clearSelection = "all"
      )
    })

    output$updateIndexDiv <- shiny::renderUI({
      seli <- input$indexTable_rows_selected
      shiny::req(length(seli) >= 1)
      shiny::req(nrow(model$indexTable) > 0)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      shiny::req(length(mt$indexes) > 0)
      shiny::req(all(seli >= 1 & seli <= length(mt$indexes)))
      # ui <- mt$indexes[[seli]]$unique
      list(
        shiny::actionButton(
          "moveIndexdUp",
          label = "",
          icon = shiny::icon(
            "arrow-alt-circle-up",
            "fa-1x",
            verify_fa = FALSE
          ),
          class = "shrunkenButton"
        ) |>
          shiny::div(
            title = "Move up",
            class = "iblock"
          ),
        shiny::actionButton(
          "moveIndexDown",
          label = "",
          icon = shiny::icon(
            "arrow-alt-circle-down",
            "fa-1x",
            verify_fa = FALSE
          ),
          class = "shrunkenButton"
        ) |>
          shiny::div(
            title = "Move down",
            class = "iblock"
          ),
        if (length(seli) == 1) {
          shiny::actionButton(
            "updateIndex",
            label = "",
            icon = shiny::icon("edit", "fa-1x", verify_fa = FALSE),
            class = "shrunkenButton"
          ) |>
            shiny::div(
              title = "Update index properties",
              class = "iblock"
            )
        },
        if (length(seli) == 1) {
          shiny::actionButton(
            "removeIndex",
            label = "",
            icon = shiny::icon("minus-square", "fa-1x", verify_fa = FALSE),
            class = "shrunkenButton"
          ) |>
            shiny::div(
              title = "Remove index",
              class = "iblock"
            )
        }
      )
    })

    #### Move Index ----
    shiny::observe({
      shiny::req(input$moveIndexdUp > 0)
      seli <- shiny::isolate(input$indexTable_rows_selected)
      shiny::req(length(seli) >= 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      m <- shiny::isolate(model$x)
      shiny::req(length(mt$indexes) > 0)
      shiny::req(min(seli) > 1)
      alli <- 1:length(m[[selTable]]$indexes)
      o <- rep(NA, length(alli))
      o[seli - 1] <- seli
      o[which(is.na(o))] <- setdiff(alli, seli)
      m <- m |>
        order_indexes(
          tableName = selTable,
          order = o
        )
      DT::selectPage(proxyIndexTable, page = ((min(seli) - 2) %/% 10) + 1)
      DT::selectRows(proxyIndexTable, selected = seli - 1)
      model$new <- m
    })
    shiny::observe({
      shiny::req(input$moveIndexDown > 0)
      seli <- shiny::isolate(input$indexTable_rows_selected)
      shiny::req(length(seli) >= 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      m <- shiny::isolate(model$x)
      shiny::req(length(mt$indexes) > 0)
      alli <- 1:length(m[[selTable]]$indexes)
      shiny::req(max(seli) < length(alli))
      o <- rep(NA, length(alli))
      o[seli + 1] <- seli
      o[which(is.na(o))] <- setdiff(alli, seli)
      m <- m |>
        order_indexes(
          tableName = selTable,
          order = o
        )
      DT::selectPage(proxyIndexTable, page = (max(seli) %/% 10) + 1)
      DT::selectRows(proxyIndexTable, selected = seli + 1)
      model$new <- m
    })

    #### Update index ----
    shiny::observeEvent(input$updateIndex, {
      seli <- shiny::isolate(input$indexTable_rows_selected)
      shiny::req(length(seli) == 1)
      shiny::req(nrow(model$indexTable) > 0)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      shiny::req(length(mt$indexes) > 0)
      shiny::req(seli >= 1 & seli <= length(mt$indexes))
      ui <- mt$indexes[[seli]]$unique
      shiny::showModal(shiny::modalDialog(
        title = "Update field",
        shiny::checkboxInput("setUniqueIndex", "Unique?", value = ui),
        size = "s",
        easyClose = TRUE
      ))
    })
    shiny::observe({
      ui <- input$setUniqueIndex
      shiny::req(!is.null(ui) && !is.na(ui))
      seli <- shiny::isolate(input$indexTable_rows_selected)
      shiny::req(length(seli) == 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      shiny::req(length(mt$indexes) > 0)
      m <- shiny::isolate(model$x)
      if (mt$indexes[[seli]]$unique != ui) {
        m <- m |>
          set_unique_index(
            tableName = selTable,
            fieldNames = mt$indexes[[seli]]$fields,
            unique = ui
          )
        if (!identical(m, shiny::isolate(model$x))) {
          model$new <- m
        } else {
          sendError(paste(
            "The index could not be modified because it is required",
            "for primary or foreign keys."
          ))
        }
        shiny::removeModal()
      }
    })
    #### Remove index ----
    shiny::observe({
      shiny::req(input$removeIndex > 0)
      seli <- shiny::isolate(input$indexTable_rows_selected)
      shiny::req(length(seli) == 1)
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      shiny::req(length(mt$indexes) > 0)
      m <- shiny::isolate(model$x)
      nm <- m |>
        remove_index(
          tableName = selTable,
          fieldNames = mt$indexes[[seli]]$fields
        )
      model$new <- nm
      if (!identical(nm, m)) {
        model$new <- nm
      } else {
        sendError(paste(
          "The index could not be removed because it is required",
          "for primary or foreign keys."
        ))
      }
    })
    #### Add index ----
    shiny::observeEvent(input$addIndex, {
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      fnames <- mt$fields$name
      shiny::showModal(shiny::modalDialog(
        title = "Add index",
        shiny::div(
          shiny::fluidRow(
            shiny::column(
              8,
              shiny::selectInput(
                "newIndexFields",
                label = NULL,
                choices = fnames,
                multiple = TRUE
              )
            ),
            shiny::column(
              2,
              shiny::checkboxInput(
                "uniqueNewIndex",
                "Unique?",
                value = FALSE
              )
            ),
            shiny::column(
              2,
              shiny::actionButton("confirmAddIndex", "Add")
            )
          )
        ),
        size = "s",
        easyClose = TRUE
      ))
    })
    shiny::observe({
      if (length(input$newIndexFields) == 0) {
        shinyjs::disable("confirmAddIndex")
      } else {
        shinyjs::enable("confirmAddIndex")
      }
    })
    shiny::observeEvent(input$confirmAddIndex, {
      mt <- shiny::isolate(model$table)
      shiny::req(mt)
      selTable <- mt$tableName
      shiny::req(length(input$newIndexFields) > 0)
      nm <- shiny::isolate(model$x) |>
        add_index(
          tableName = selTable,
          fieldNames = input$newIndexFields,
          unique = input$uniqueNewIndex
        )
      if (!identical(nm, shiny::isolate(model$x))) {
        model$new <- nm
      } else {
        sendError(paste(
          "The index could not be added:",
          "it may already exist or it may not fit other constraints."
        ))
      }
      shiny::removeModal()
    })

    #########################################################################@
    ## Edit menu ----
    #########################################################################@

    output$addFKInput <- shiny::renderUI({
      selTable <- selection$tables
      shiny::req(length(selTable) > 0 & length(selTable) <= 2)
      shiny::actionButton(
        "addForeignKey",
        "Key",
        icon = shiny::icon("plus", "fa-2x", verify_fa = FALSE),
        # icon=shiny::icon("external-link-alt", "fa-2x", verify_fa=FALSE),
        class = "shrunkenButton"
      ) |>
        shiny::div(
          title = "Add a foreign key"
        )
    })
    output$editFKInput <- shiny::renderUI({
      selFK <- selection$fk
      shiny::req(length(selFK) == 1)
      shiny::actionButton(
        "editFK",
        label = shiny::HTML(paste(
          '<i class="far fa-edit fa-2x"></i>',
          'keys'
        )),
        class = "shrunkenButton"
      ) |>
        shiny::div(
          title = "Edit cardinalities of the selected foreign key (F2)"
        )
    })
    output$rmFKInput <- shiny::renderUI({
      selFK <- selection$fk
      shiny::req(length(selFK) > 0)
      shiny::actionButton(
        "removeFK",
        label = shiny::HTML(paste(
          '<i class="far fa-trash-alt fa-2x"></i>',
          'keys'
        )),
        class = "shrunkenButton"
      ) |>
        shiny::div(
          title = "Remove selected foreign keys (del)"
        )
    })
    output$rmTablesInput <- shiny::renderUI({
      selTable <- selection$tables
      shiny::req(length(selTable) > 0)
      shiny::actionButton(
        "removeTables",
        label = shiny::HTML(paste(
          '<i class="fas fa-trash fa-2x"></i>',
          'tables'
        )),
        class = "shrunkenButton"
      ) |>
        shiny::div(
          title = "Remove selected tables (del)"
        )
    })
    output$dupTablesInput <- shiny::renderUI({
      selTable <- selection$tables
      shiny::req(length(selTable) > 0)
      shiny::actionButton(
        "duplicateTables",
        label = shiny::HTML(paste(
          '<i class="fas fa-copy fa-2x"></i>',
          'tables'
        )),
        class = "shrunkenButton"
      ) |>
        shiny::div(
          title = "Duplicate selected tables"
        )
    })

    #########################################################################@
    ## Table color ----
    #########################################################################@

    output$setTableColor <- shiny::renderUI({
      selTables <- selection$tables
      shiny::req(length(selTables) > 0)
      tval <- lapply(
        shiny::isolate(model$x),
        function(x) x$display$color
      ) |>
        unlist()
      tval <- tval[selTables] |> unique()
      if (length(tval) > 1 || is.na(tval)) {
        tval = ""
      }
      colourpicker::colourInput(
        "tableColor",
        label = NULL,
        value = tval,
        showColour = "background",
        palette = "limited",
        allowedCols = c(
          "",
          shiny::isolate(settings$availableColors)
        ),
        allowTransparent = TRUE
      ) |>
        shiny::div(
          id = "tabColPick",
          title = "Select table color"
        )
    })
    shiny::observe({
      newCol <- input$tableColor
      shiny::req(newCol != "")
      selTables <- shiny::isolate(selection$tables)
      shiny::req(length(selTables) > 0)
      m <- shiny::isolate(model$x)
      tval <- lapply(
        m,
        function(x) x$display$color
      ) |>
        unlist()
      tval <- tval[selTables] |> unique()
      if (length(tval) > 1 || is.na(tval) || tval != newCol) {
        for (tn in selTables) {
          m <- m |>
            update_table_display(
              tableName = tn,
              color = newCol
            )
        }
        model$new <- m
      }
    })

    #########################################################################@
    ## Remove tables ----
    #########################################################################@

    shiny::observe({
      takeAction <- ((!is.null(input$removeTables) && input$removeTables > 0))
      shiny::req(takeAction)
      tns <- shiny::isolate(selection$tables)
      m <- shiny::isolate(model$x)
      if (length(tns) > 0 && all(tns != "") && all(tns %in% names(m))) {
        for (tn in tns) {
          m <- try(remove_table(m, tableName = tn), silent = TRUE)
          if (!is.RelDataModel(m)) {
            break()
          }
        }
        if (is.RelDataModel(m)) {
          model$new <- m
        } else {
          shiny::showModal(shiny::modalDialog(
            title = "Unable to remove table",
            shiny::p(
              shiny::HTML(paste(
                sprintf("<u>%s</u> is referenced by other tables.", tn),
                "Remove foreign key(s) before removing this table.",
                sep = "<br>"
              )),
              class = "errorMessage"
            ),
            size = "s",
            easyClose = TRUE
          ))
        }
      }
    })

    #########################################################################@
    ## Duplicate tables ----
    #########################################################################@

    shiny::observe({
      takeAction <- ((!is.null(input$duplicateTables) &&
        input$duplicateTables > 0))
      shiny::req(takeAction)
      tns <- shiny::isolate(selection$tables)
      m <- shiny::isolate(model$x)
      vn <- shiny::isolate(model$vn)
      xs <- 100
      ys <- 100
      nm <- unclass(m)
      if (length(tns) > 0 && all(tns != "") && all(tns %in% names(m))) {
        for (tn in tns) {
          i <- 1
          ntn <- paste(tn, "COPY", i, sep = "_")
          while (ntn %in% names(m)) {
            i <- i + 1
            ntn <- paste(tn, "COPY", i, sep = "_")
          }
          toAdd <- nm[tn]
          toAdd[[1]]$tableName <- ntn
          pr <- stats::rbeta(1, 9, 1)
          pa <- stats::runif(1, 0, 2 * pi)
          toAdd[[1]]$display$x <- toAdd[[1]]$display$x +
            xs * pr * cos(pa)
          toAdd[[1]]$display$y <- toAdd[[1]]$display$y +
            ys * pr * sin(pa)
          names(toAdd) <- ntn
          nm <- c(nm, toAdd)
        }
      }
      model$new <- RelDataModel(nm)
    })

    #########################################################################@
    ## Add foreign keys ----
    #########################################################################@

    foreignKey <- shiny::reactiveValues(
      triggered = 0,
      fromTable = NULL,
      toTable = NULL,
      fromFields = NULL,
      toFields = NULL,
      fmin = NULL,
      fmax = NULL,
      tmin = NULL,
      tmax = NULL
    )

    shiny::observe({
      shiny::req(input$addForeignKey > 0)
      tns <- shiny::isolate(selection$tables)
      m <- shiny::isolate(model$x)
      foreignKey$fromTable <- foreignKey$toTable <-
        foreignKey$fromFields <- foreignKey$toFields <- NULL
      foreignKey$triggered = shiny::isolate(foreignKey$triggered) + 1
      if (length(tns) > 0 && all(tns != "") && all(tns %in% names(m))) {
        shiny::showModal(shiny::modalDialog(
          title = "Add foreign key",
          shiny::uiOutput("addForeignKey"),
          size = "m",
          easyClose = TRUE
        ))
      }
    })

    output$addForeignKey <- shiny::renderUI({
      tns <- selection$tables
      shiny::req(length(tns) > 0)
      shiny::req(foreignKey$triggered > 0)
      foreignKey$fromTable <- tns[1]
      foreignKey$toTable <- tns[length(tns)]
      shiny::div(
        shiny::fluidRow(
          shiny::column(
            5,
            shiny::fluidRow(shiny::h4(tns[1]), class = "centerBox"),
            shiny::fluidRow(shiny::uiOutput("ilcard"))
          ),
          if (length(tns) == 1) {
            shiny::column(
              2,
              shiny::actionButton("confirmAddFK", "Add", disabled = TRUE),
              shiny::tags$br(),
              shiny::icon(
                "long-arrow-alt-right",
                "fa-2x",
                verify_fa = FALSE
              ),
              class = "centerBox"
            )
          } else {
            shiny::column(
              2,
              shiny::actionButton("confirmAddFK", "Add", disabled = TRUE),
              shiny::tags$br(),
              shiny::actionButton(
                "fkDirection",
                "",
                icon = shiny::icon(
                  "long-arrow-alt-right",
                  "fa-2x",
                  verify_fa = FALSE
                )
              ) |>
                shiny::div(title = "Change foreign key direction"),
              class = "centerBox"
            )
          },
          shiny::column(
            5,
            shiny::fluidRow(
              shiny::h4(tns[length(tns)]),
              class = "centerBox"
            ),
            shiny::fluidRow(shiny::uiOutput("ircard"))
          )
        ),
        shiny::tags$hr(class = "editSeparator"),
        ##
        shiny::fluidRow(shiny::uiOutput("fkFields")),
        shiny::tags$hr(class = "editSeparator"),
        ##
        shiny::fluidRow(shiny::uiOutput("possibleFkFields"))
      )
    })

    shiny::observe({
      if (
        length(foreignKey$fromTable) == 0 ||
          length(foreignKey$toTable) == 0 ||
          length(foreignKey$fromFields) == 0 ||
          length(foreignKey$toFields) == 0
      ) {
        shinyjs::disable("confirmAddFK")
      } else {
        shinyjs::enable("confirmAddFK")
      }
    })

    shiny::observe({
      shiny::req(input$fkDirection > 0)
      ft <- shiny::isolate(foreignKey$toTable)
      tt <- shiny::isolate(foreignKey$fromTable)
      shiny::req(ft)
      shiny::req(tt)
      shiny::req(ft != tt)
      tns <- sort(c(ft, tt))
      foreignKey$fromTable <- ft
      foreignKey$toTable <- tt
      foreignKey$fromFields <- foreignKey$toFields <- NULL
      if (ft == tns[1]) {
        shiny::updateActionButton(
          session,
          "fkDirection",
          icon = shiny::icon(
            "long-arrow-alt-right",
            "fa-2x",
            verify_fa = FALSE
          )
        )
      } else {
        shiny::updateActionButton(
          session,
          "fkDirection",
          icon = shiny::icon(
            "long-arrow-alt-left",
            "fa-2x",
            verify_fa = FALSE
          )
        )
      }
    })

    output$possibleFkFields <- shiny::renderUI({
      ft <- foreignKey$fromTable
      tt <- foreignKey$toTable
      shiny::req(ft)
      shiny::req(tt)
      tns <- sort(c(ft, tt))
      m <- shiny::isolate(model$x)
      if (is.MatrixModel(m[[ft]])) {
        ftfields <- m[[ft]]$fields$name[which(
          m[[ft]]$fields$type %in% c("row", "column")
        )]
      } else {
        ftfields <- m[[ft]]$fields$name
      }
      if (is.MatrixModel(m[[tt]])) {
        ttfields <- m[[tt]]$fields$name[which(
          m[[tt]]$fields$type %in% c("row", "column")
        )]
      } else {
        ttfields <- m[[tt]]$fields$name
      }
      toRet <- list(
        shiny::column(
          5,
          shiny::div(
            shiny::selectInput(
              "fkFromField",
              "",
              ftfields,
              multiple = FALSE,
              width = "100%",
              selectize = FALSE
            ),
            class = "fkFieldSel"
          )
        ),
        shiny::column(2, shiny::uiOutput("addFkFields"), class = "centerBox"),
        shiny::column(
          5,
          shiny::div(
            shiny::selectInput(
              "fkToField",
              "",
              ttfields,
              multiple = FALSE,
              width = "100%",
              selectize = FALSE
            ),
            class = "fkFieldSel"
          )
        )
      )
      if (tns[1] != ft) {
        toRet <- toRet[c(3, 2, 1)]
      }
      return(toRet)
    })

    output$addFkFields <- shiny::renderUI({
      ft <- foreignKey$fromTable
      tt <- foreignKey$toTable
      shiny::req(ft)
      shiny::req(tt)
      m <- shiny::isolate(model$x)
      ftfields <- m[[ft]]$fields
      ttfields <- m[[tt]]$fields
      from <- input$fkFromField
      to <- input$fkToField
      shiny::req(from %in% ftfields$name)
      shiny::req(to %in% ttfields$name)
      fft <- ftfields[which(ftfields$name == from), ]$type
      fft <- ifelse(fft %in% c("row", "column"), "character", fft)
      tft <- ttfields[which(ttfields$name == to), ]$type
      tft <- ifelse(tft %in% c("row", "column"), "character", tft)
      if (fft != tft) {
        return(shiny::tagList(
          shiny::tags$br(),
          shiny::p("Incompatible types", class = "errorMessage")
        ))
      } else {
        if (fft == "base64") {
          return(shiny::tagList(
            shiny::tags$br(),
            shiny::p(
              "base64 fields cannot be used in foreign keys",
              class = "errorMessage"
            )
          ))
        } else {
          selFrom <- foreignKey$fromFields
          selTo <- foreignKey$toFields
          alreadyIn <- length(which(selFrom == from & selTo == to)) > 0
          if (alreadyIn) {
            return(shiny::tagList(
              shiny::tags$br(),
              shiny::p("Already in key", class = "errorMessage")
            ))
          } else {
            return(shiny::tagList(
              shiny::tags$br(),
              shiny::actionButton(
                "addFkField",
                label = "",
                icon = shiny::icon(
                  "plus-square",
                  "fa-1x",
                  verify_fa = FALSE
                )
              ) |>
                shiny::div(title = "Add key field")
            ))
          }
        }
      }
    })

    shiny::observe({
      shiny::req(input$addFkField > 0)
      foreignKey$fromFields <- c(
        shiny::isolate(foreignKey$fromFields),
        shiny::isolate(input$fkFromField)
      )
      foreignKey$toFields <- c(
        shiny::isolate(foreignKey$toFields),
        shiny::isolate(input$fkToField)
      )
    })

    output$fkFields <- shiny::renderUI({
      from <- foreignKey$fromFields
      to <- foreignKey$toFields
      shiny::req(from)
      shiny::req(to)
      shiny::fluidRow(
        shiny::column(1, ""),
        shiny::column(10, DT::DTOutput("fkFieldTable")),
        shiny::column(1, shiny::uiOutput("rmFkField"), class = "rightBox")
      )
    })
    output$fkFieldTable <- DT::renderDT({
      from <- foreignKey$fromFields
      to <- foreignKey$toFields
      shiny::req(from)
      shiny::req(to)
      ft <- shiny::isolate(foreignKey$fromTable)
      tt <- shiny::isolate(foreignKey$toTable)
      tns <- sort(c(ft, tt))
      if (tns[1] == ft) {
        left <- from
        right <- to
      } else {
        left <- to
        right <- from
      }
      DT::datatable(
        dplyr::tibble(l = left, r = right),
        rownames = FALSE,
        colnames = c("", ""),
        options = list(
          dom = ifelse(length(left) > 10, "tip", "t"),
          columnDefs = list(
            list(targets = c(0), visible = TRUE, width = '50%'),
            list(targets = c(1), visible = TRUE, width = '50%')
          )
        )
      ) |>
        DT::formatStyle(1, "text-align" = "left") |>
        DT::formatStyle(2, "text-align" = "right")
    })
    output$rmFkField <- shiny::renderUI({
      sel <- input$fkFieldTable_rows_selected
      shiny::req(sel)
      return(
        shiny::actionButton(
          "confirmRmFkField",
          label = "",
          icon = shiny::icon("minus-square", "fa-1x", verify_fa = FALSE)
        ) |>
          shiny::div(title = "Remove key field")
      )
    })
    shiny::observe({
      shiny::req(input$confirmRmFkField)
      sel <- shiny::isolate(input$fkFieldTable_rows_selected)
      shiny::req(length(sel) > 0)
      foreignKey$fromFields <- shiny::isolate(foreignKey$fromFields)[-sel]
      foreignKey$toFields <- shiny::isolate(foreignKey$toFields)[-sel]
    })

    shiny::observe({
      shiny::req(input$confirmAddFK > 0)
      m <- shiny::isolate(model$x)
      suppressWarnings(
        nm <- try(
          {
            m |>
              add_foreign_key(
                fromTable = shiny::isolate(foreignKey$fromTable),
                toTable = shiny::isolate(foreignKey$toTable),
                fromFields = shiny::isolate(foreignKey$fromFields),
                toFields = shiny::isolate(foreignKey$toFields),
                fmin = shiny::isolate(foreignKey$fmin),
                fmax = shiny::isolate(foreignKey$fmax),
                tmin = shiny::isolate(foreignKey$tmin),
                tmax = shiny::isolate(foreignKey$tmax)
              )
          },
          silent = TRUE
        )
      )
      if (inherits(nm, "try-error")) {
        sendError(as.character(nm))
      } else {
        shiny::removeModal()
        if (identical(nm, m)) {
          sendError(paste(
            "The foreign key could not be added:",
            "it may already exist or it may not fit other constraints."
          ))
        } else {
          sendWarning(paste(
            "Some indexes, uniqueness and mandatory constraints may",
            "have been added to fields to support the foreign key",
            "cardinalities."
          ))
        }
        model$new <- nm
        foreignKey$fromTable <- NULL
        foreignKey$toTable <- NULL
        foreignKey$fromFields <- NULL
        foreignKey$toFields <- NULL
        foreignKey$fmin <- NULL
        foreignKey$fmax <- NULL
        foreignKey$tmin <- NULL
        foreignKey$tmax <- NULL
      }
    })

    ### Cardinality ----
    output$ilcard <- shiny::renderUI({
      ft <- foreignKey$fromTable
      tt <- foreignKey$toTable
      shiny::req(ft)
      shiny::req(tt)
      tns <- sort(c(ft, tt))
      m <- shiny::isolate(model$x)
      cmin <- if (tns[1] == ft) {
        shiny::isolate(foreignKey$fmin)
      } else {
        shiny::isolate(foreignKey$tmin)
      }
      cmax <- if (tns[1] == ft) {
        shiny::isolate(foreignKey$fmax)
      } else {
        shiny::isolate(foreignKey$tmax)
      }
      cmin <- as.character(cmin)
      cmax <- ifelse(cmax == -1, "n", as.character(cmax))
      toRet <- list(
        shiny::column(
          6,
          shiny::selectInput(
            "leftcardmin",
            "Min. card.",
            choices = c("0", "1"),
            selected = ifelse(
              length(cmin) == 0,
              ifelse(tns[1] == ft, "0", "1"),
              cmin
            ),
            selectize = FALSE
          )
        ),
        shiny::column(
          6,
          shiny::selectInput(
            "leftcardmax",
            "Max. card.",
            choices = c("1", "n"),
            selected = ifelse(
              length(cmax) == 0,
              ifelse(tns[1] == ft, "n", "1"),
              cmax
            ),
            selectize = FALSE
          )
        )
      )
      return(toRet)
    })
    output$ircard <- shiny::renderUI({
      ft <- foreignKey$fromTable
      tt <- foreignKey$toTable
      shiny::req(ft)
      shiny::req(tt)
      tns <- sort(c(ft, tt))
      m <- shiny::isolate(model$x)
      cmin <- if (tns[1] != ft) {
        shiny::isolate(foreignKey$fmin)
      } else {
        shiny::isolate(foreignKey$tmin)
      }
      cmax <- if (tns[1] != ft) {
        shiny::isolate(foreignKey$fmax)
      } else {
        shiny::isolate(foreignKey$tmax)
      }
      cmin <- as.character(cmin)
      cmax <- ifelse(cmax == -1, "n", as.character(cmax))
      toRet <- list(
        shiny::column(
          6,
          shiny::selectInput(
            "rightcardmin",
            "Min. card.",
            choices = c("0", "1"),
            selected = ifelse(
              length(cmin) == 0,
              ifelse(tns[1] != ft, "0", "1"),
              cmin
            ),
            selectize = FALSE
          )
        ),
        shiny::column(
          6,
          shiny::selectInput(
            "rightcardmax",
            "Max. card.",
            choices = c("1", "n"),
            selected = ifelse(
              length(cmax) == 0,
              ifelse(tns[1] != ft, "n", "1"),
              cmax
            ),
            selectize = FALSE
          )
        )
      )
      return(toRet)
    })

    shiny::observe({
      cval <- c("0" = 0L, "1" = 1L, "n" = -1L)
      ft <- foreignKey$fromTable
      tt <- foreignKey$toTable
      shiny::req(ft)
      shiny::req(tt)
      tns <- sort(c(ft, tt))
      lftmin <- input$leftcardmin
      lftmax <- input$leftcardmax
      rgtmin <- input$rightcardmin
      rgtmax <- input$rightcardmax
      shiny::req(lftmin)
      shiny::req(lftmax)
      shiny::req(rgtmin)
      shiny::req(rgtmax)
      foreignKey$fmin <- as.integer(cval[ifelse(
        tns[1] == ft,
        lftmin,
        rgtmin
      )])
      foreignKey$fmax <- as.integer(cval[ifelse(
        tns[1] == ft,
        lftmax,
        rgtmax
      )])
      foreignKey$tmin <- as.integer(cval[ifelse(
        tns[1] != ft,
        lftmin,
        rgtmin
      )])
      foreignKey$tmax <- as.integer(cval[ifelse(
        tns[1] != ft,
        lftmax,
        rgtmax
      )])
    })

    #########################################################################@
    ## Update foreign keys ----
    #########################################################################@

    shiny::observe({
      shiny::req(input$editFK > 0)
      selFK <- shiny::isolate(selection$fk)
      shiny::req(length(selFK) == 1)
      m <- shiny::isolate(model$x)
      mne <- shiny::isolate(model$vn)$edges
      shiny::req(selFK %in% mne$id)
      i <- which(mne$id == selFK)
      foreignKey$fromTable <- mne$from[i]
      foreignKey$toTable <- mne$to[i]
      foreignKey$fromFields <- mne$ff[[i]]
      foreignKey$toFields <- mne$tf[[i]]
      foreignKey$triggered = shiny::isolate(foreignKey$triggered) + 1
      shiny::showModal(shiny::modalDialog(
        title = "Edit foreign key",
        shiny::uiOutput("editForeignKey"),
        size = "m",
        easyClose = TRUE
      ))
    })

    output$editForeignKey <- shiny::renderUI({
      shiny::req(foreignKey$triggered > 0)
      ft <- shiny::isolate(foreignKey$fromTable)
      tt <- shiny::isolate(foreignKey$toTable)
      ff <- shiny::isolate(foreignKey$fromFields)
      tf <- shiny::isolate(foreignKey$toFields)
      shiny::req(ft)
      shiny::req(tt)
      shiny::req(ff)
      shiny::req(tf)
      tns <- sort(c(ft, tt))
      shiny::div(
        shiny::fluidRow(
          shiny::column(
            5,
            shiny::fluidRow(shiny::h4(tns[1]), class = "centerBox"),
            shiny::fluidRow(shiny::uiOutput("ilcard"))
          ),
          if (tns[1] == ft) {
            shiny::column(
              2,
              shiny::actionButton("confirmUpdateFK", "Update"),
              shiny::tags$br(),
              shiny::icon(
                "long-arrow-alt-right",
                "fa-2x",
                verify_fa = FALSE
              ),
              class = "centerBox"
            )
          } else {
            shiny::column(
              2,
              shiny::actionButton("confirmUpdateFK", "Update"),
              shiny::tags$br(),
              shiny::icon(
                "long-arrow-alt-left",
                "fa-2x",
                verify_fa = FALSE
              ),
              class = "centerBox"
            )
          },
          shiny::column(
            5,
            shiny::fluidRow(
              shiny::h4(tns[length(tns)]),
              class = "centerBox"
            ),
            shiny::fluidRow(shiny::uiOutput("ircard"))
          )
        )
      )
    })

    shiny::observe({
      shiny::req(input$confirmUpdateFK > 0)
      model$new <- shiny::isolate(model$x) |>
        update_foreign_key(
          fromTable = shiny::isolate(foreignKey$fromTable),
          toTable = shiny::isolate(foreignKey$toTable),
          fromFields = shiny::isolate(foreignKey$fromFields),
          toFields = shiny::isolate(foreignKey$toFields),
          fmin = shiny::isolate(foreignKey$fmin),
          fmax = shiny::isolate(foreignKey$fmax),
          tmin = shiny::isolate(foreignKey$tmin),
          tmax = shiny::isolate(foreignKey$tmax)
        )
      shiny::removeModal()
      foreignKey$fromTable <- NULL
      foreignKey$toTable <- NULL
      foreignKey$fromFields <- NULL
      foreignKey$toFields <- NULL
      foreignKey$fmin <- NULL
      foreignKey$fmax <- NULL
      foreignKey$tmin <- NULL
      foreignKey$tmax <- NULL
    })

    #########################################################################@
    ## Remove foreign keys ----
    #########################################################################@

    shiny::observe({
      takeAction <- ((!is.null(input$removeFK) && input$removeFK > 0))
      shiny::req(takeAction)
      fks <- shiny::isolate(selection$fk)
      m <- shiny::isolate(model$x)
      mne <- shiny::isolate(model$vn)$edges
      if (length(fks) > 0 && all(fks != "") && all(fks %in% mne$id)) {
        for (fk in fks) {
          i <- which(mne$id == fk)
          m <- m |>
            remove_foreign_key(
              fromTable = mne$from[i],
              fromFields = mne$ff[[i]],
              toTable = mne$to[i],
              toFields = mne$tf[[i]]
            )
        }
        model$new <- m
      }
    })

    #########################################################################@
    ## Node positions ----
    #########################################################################@

    shiny::observe({
      dispNodes <- input$modelNet_nodes
      shiny::req(dispNodes)
      m <- shiny::isolate(model$x)
      cp <- do.call(
        rbind,
        lapply(m, function(n) dplyr::tibble(x = n$display$x, y = n$display$y))
      )
      np <- do.call(
        rbind,
        lapply(dispNodes[names(m)], function(n) dplyr::tibble(x = n$x, y = n$y))
      )
      if (!all(np$x == cp$x & np$y == cp$y)) {
        m <- lapply(
          m,
          function(n) {
            if (n$tableName %in% names(dispNodes)) {
              n$display$x = dispNodes[[n$tableName]]$x
              n$display$y = dispNodes[[n$tableName]]$y
            }
            return(n)
          }
        )
        class(m) <- c("RelDataModel", "list")
        attr(m, "updateVis") <- FALSE
        model$new <- m
      }
    })

    #########################################################################@
    ## Update model ----
    #########################################################################@

    shiny::observe({
      nm <- model$new
      shiny::req(nm)
      updateVis <- attr(nm, "updateVis")
      if (is.null(updateVis)) {
        updateVis <- TRUE
      }
      attr(nm, "updateVis") <- NULL
      dm <- shiny::isolate(model$x)
      shiny::req(!identical(nm, dm))

      ##
      tdm <- nm
      if (length(dm) == 0) {
        toReplot <- TRUE
      } else {
        toReplot <- FALSE
        if (updateVis) {
          ndm <- shiny::isolate(model$vn)
          ntdm <- modelToVn(
            tdm,
            color = shiny::isolate(settings$defaultColor)
          )
          edgeToDel <- setdiff(ndm$edges$id, ntdm$edges$id)
          if (length(edgeToDel) > 0) {
            visNetwork::visNetworkProxy("modelNet") |>
              visNetwork::visRemoveEdges(edgeToDel)
          }
          nodeToDel <- setdiff(names(dm), names(tdm))
          if (length(nodeToDel) > 0) {
            visNetwork::visNetworkProxy("modelNet") |>
              visNetwork::visRemoveNodes(nodeToDel)
          }
          visNetwork::visNetworkProxy("modelNet") |>
            visNetwork::visUpdateNodes(ntdm$nodes) |>
            visNetwork::visUpdateEdges(ntdm$edges)
        }
      }
      ##

      ch <- shiny::isolate(model$history)
      cm <- shiny::isolate(model$current)
      ch <- ch[-((cm:length(ch)) + 1)]
      ch <- c(ch, list(nm))
      cm <- cm + 1
      model$history <- ch
      model$current <- cm
      model$x <- nm
      model$new <- NULL
      if (toReplot) {
        replot$x <- shiny::isolate(replot$x) + 1
      }
    })

    shiny::observe({
      selTables <- selection$tables
      shiny::req(!shiny::isolate(selection$fromVN))
      visNetwork::visNetworkProxy("modelNet") |>
        visNetwork::visSelectNodes(selTables)
    })

    shiny::observe({
      selFK <- selection$fk
      selTables <- intersect(names(model$x), selection$tables)
      shiny::req(!shiny::isolate(selection$fromVN))
      if (length(selTables) == 0) {
        visNetwork::visNetworkProxy("modelNet") |>
          visNetwork::visSelectEdges(selFK)
      }
    })

    #########################################################################@
    ## Manage history ----
    #########################################################################@

    shiny::observe({
      m <- model$x
      mn <- model$vn
      selection$tables <- sort(intersect(
        shiny::isolate(selection$tables),
        names(m)
      ))
      selection$fk <- sort(intersect(
        shiny::isolate(selection$fk),
        mn$edges$id
      ))
    })

    shiny::observe({
      shiny::req(input$undo)
      ch <- shiny::isolate(model$history)
      cm <- shiny::isolate(model$current)
      cm <- cm - 1
      shiny::req(cm > 0)

      ###########################@
      ## The commented code below was first used to
      ## replot the network from scratch
      # model$x <- ch[[cm]]
      # model$current <- cm
      # replot$x <- shiny::isolate(replot$x)+1
      ###########################@

      dm <- shiny::isolate(model$x)
      tdm <- ch[[cm]]
      ndm <- shiny::isolate(model$vn)
      ntdm <- modelToVn(tdm, color = shiny::isolate(settings$defaultColor))
      edgeToDel <- setdiff(ndm$edges$id, ntdm$edges$id)
      if (length(edgeToDel) > 0) {
        visNetwork::visNetworkProxy("modelNet") |>
          visNetwork::visRemoveEdges(edgeToDel)
      }
      nodeToDel <- setdiff(names(dm), names(tdm))
      if (length(nodeToDel) > 0) {
        visNetwork::visNetworkProxy("modelNet") |>
          visNetwork::visRemoveNodes(nodeToDel)
      }
      visNetwork::visNetworkProxy("modelNet") |>
        visNetwork::visUpdateNodes(ntdm$nodes) |>
        visNetwork::visUpdateEdges(ntdm$edges)
      model$x <- tdm
      model$current <- cm
    })

    shiny::observe({
      shiny::req(input$redo)
      ch <- shiny::isolate(model$history)
      cm <- shiny::isolate(model$current)
      cm <- cm + 1
      shiny::req(cm <= length(ch))

      ###########################@
      ## The commented code below was first used to
      ## replot the network from scratch
      # model$x <- ch[[cm]]
      # model$current <- cm
      # replot$x <- shiny::isolate(replot$x)+1
      ###########################@

      dm <- shiny::isolate(model$x)
      tdm <- ch[[cm]]
      ndm <- shiny::isolate(model$vn)
      ntdm <- modelToVn(tdm, color = shiny::isolate(settings$defaultColor))
      edgeToDel <- setdiff(ndm$edges$id, ntdm$edges$id)
      if (length(edgeToDel) > 0) {
        visNetwork::visNetworkProxy("modelNet") |>
          visNetwork::visRemoveEdges(edgeToDel)
      }
      nodeToDel <- setdiff(names(dm), names(tdm))
      if (length(nodeToDel) > 0) {
        visNetwork::visNetworkProxy("modelNet") |>
          visNetwork::visRemoveNodes(nodeToDel)
      }
      visNetwork::visNetworkProxy("modelNet") |>
        visNetwork::visUpdateNodes(ntdm$nodes) |>
        visNetwork::visUpdateEdges(ntdm$edges)
      model$x <- tdm
      model$current <- cm
    })

    shiny::observe({
      if (model$current == 1) {
        shinyjs::disable("undo")
      }
      if (model$current > 1) {
        shinyjs::enable("undo")
      }
      if (model$current >= length(model$history)) {
        shinyjs::disable("redo")
      }
      if (model$current < length(model$history)) {
        shinyjs::enable("redo")
      }
    })

    #########################################################################@
    ## Export model ----
    #########################################################################@

    shiny::observeEvent(input$export, {
      shiny::showModal(shiny::modalDialog(
        title = "Export",
        shiny::uiOutput("export"),
        size = "s",
        easyClose = TRUE
      ))
    })

    output$export <- shiny::renderUI({
      shiny::fluidRow(
        shiny::column(
          6,
          shiny::downloadButton(
            "exportJson",
            list(
              shiny::icon("file-code", "fa-2x", verify_fa = FALSE),
              "JSON"
            ),
            class = "exportButtons"
          )
        ),
        shiny::column(
          6,
          shiny::downloadButton(
            "exportHtml",
            list(
              shiny::icon("map", "fa-2x", verify_fa = FALSE),
              "HTML"
            ),
            class = "exportButtons"
          )
        )
      )
    })

    ### JSON ----
    output$exportJson <- shiny::downloadHandler(
      filename = function() {
        paste0("Data-model", ".json")
      },
      content = function(file) {
        m <- shiny::isolate(model$x)
        shiny::req(m)
        write_json_data_model(m, file)
      }
    )

    ### HTML ----
    output$exportHtml <- shiny::downloadHandler(
      filename = function() {
        paste0("Data-model", ".html")
      },
      content = function(file) {
        m <- shiny::isolate(model$x)
        shiny::req(m)
        plot(m, color = shiny::isolate(settings$defaultColor)) |>
          visNetwork::visSave(file)
      }
    )

    #########################################################################@
    ## From R ----
    #########################################################################@

    if (fromR) {
      ### Autosaved object ----
      shiny::observe({
        assign(bcko, model$x, envir = modelEnv)
      })
      ### Done button ----
      shiny::observeEvent(input$done, {
        shiny::stopApp(invisible(model$x))
      })
    }
  }
}

###############################################################################@
#' Relational data modeler GUI
#'
#' @param modelInput the [RelDataModel] to start from
#' @param fromR a logical indicating if the application is launched from R
#' @param defaultColor a single color indicating the default table color
#' @param availableColors a character of possible colors for tables
#' @param example a file path to an sql or json model
#' @param forceIntro if TRUE the help tour start when the application
#' is launched (default: FALSE)
#'
#' @return The [RelDataModel] designed with the GUI.
#'
#' @export
#'
model_relational_data <- function(
  modelInput = RelDataModel(list()),
  fromR = interactive(),
  defaultColor = "#D9D9D9",
  # availableColors=c(
  #    "#7CE65F", "#D9B661", "#75E4AE", "#8FE6E0",
  #    "#DFFB86", "#FEFE8F", "#A9ECC9", "#F8DEC3",
  #    "#E0B8A0", "#FAC6DC", "#CC8490", "#F67FC4",
  #    "#C6BDF1", "#D8B8E3", "#9BC8FE", "#C6D1DC",
  #    "#9638E5", "#7D76D9", "#6E9DCE"
  # ),
  availableColors = c(
    "#9BC8FE",
    "#F67FC4", #"#ACF1BC",
    "#C6BDF1",
    "#DFFB86",
    "#F8DEC3",
    "#8FE6E0",
    "#FEFE8F",
    "#FAC6DC",
    "#A9ECC9"
  ),
  # availableColors=c(
  #    "#D9B661", "#E14D7D", "#75E4AE", "#9638E5", "#C6D1DC", "#D8B8E3",
  #    "#C663CB", "#7CE65F", "#DF7442", "#E0B8A0", "#D4E355", "#7D76D9",
  #    "#6E9DCE", "#E747D7", "#77DFDE", "#CC8490", "#D991D6", "#D0E9CF",
  #    "#CAE095", "#6E8D65"
  # ),
  example = system.file(
    "examples/HPO-model.json",
    package = utils::packageName()
  ),
  forceIntro = FALSE
) {
  bcko <- NA
  if (fromR) {
    bcko <- paste0(
      "Model_",
      format(Sys.time(), format = "%Y_%m_%d_%H_%M_%S")
    )
    while (exists(bcko, where = modelEnv)) {
      Sys.sleep(1)
      bcko <- paste0(
        "Model_",
        format(Sys.time(), format = "%Y_%m_%d_%H_%M_%S")
      )
    }
  }

  modelInput <- auto_layout(modelInput, lengthMultiplier = 300)

  ui <- buildUi(fromR = fromR)
  server <- buildServer(
    modelInput = modelInput,
    fromR = fromR,
    bcko = bcko,
    defaultColor = defaultColor,
    availableColors = availableColors,
    example = example,
    forceIntro = forceIntro
  )

  if (fromR) {
    ## From R ----
    on.exit(
      cat(
        sprintf(
          "The returned model has also been autosaved as %s",
          crayon::green(bcko)
        ),
        sprintf(
          "\n  - Use %s to get it back.",
          crayon::yellow(
            sprintf('recover_RelDataModel("%s")', bcko)
          )
        ),
        sprintf(
          "\n  - Use %s",
          crayon::yellow("list_autosaved_RelDataModel()")
        ),
        sprintf(
          " and %s",
          crayon::yellow("clean_autosaved_RelDataModels()")
        ),
        " to respectively list and remove all autosaved models.",
        "\n"
      )
    )
    shiny::runApp(shiny::shinyApp(ui, server))
  } else {
    ## Remote app ----
    shiny::shinyApp(ui, server)
  }
}

###############################################################################@
#' Recover an autosaved [RelDataModel]
#'
#' @param name The name of the autosaved [RelDataModel] to bring back.
#' Available autosaved [RelDataModel] can be listed using
#' the [list_autosaved_RelDataModel()]. If NA (default) the latest model is
#' returned.
#'
#' @export
#'
recover_RelDataModel <- function(name = NA) {
  if (length(name) > 1) {
    warning("Several names provided: taking only the first one into account.")
  }
  if (length(name) == 0) {
    stop("No name provided")
  }
  if (is.na(name)) {
    name <- list_autosaved_RelDataModel()
    name <- name[length(name)]
    if (length(name) == 0) {
      stop("There is not any model to recover")
    }
  }
  get(name, envir = modelEnv)
}

###############################################################################@
#' List autosaved [RelDataModel]
#'
#' @seealso [clean_autosaved_RelDataModels()] to clean this list.
#'
#' @export
#'
list_autosaved_RelDataModel <- function() {
  ls(envir = modelEnv, all.names = TRUE)
}

###############################################################################@
#' Remove all autosaved [RelDataModel]
#'
#' @export
#'
clean_autosaved_RelDataModels <- function() {
  rm(list = list_autosaved_RelDataModel(), envir = modelEnv)
}

###############################################################################@
modelEnv <- new.env(hash = TRUE, parent = emptyenv())

###############################################################################@
#' Plot a [RelDataModel] object
#'
#' This function draw a visNetwork of the [RelDataModel].
#'
#' @param x a [RelDataModel]
#' @param ... additional parameters:
#' - **color** default table background color
#' - **border** border color (single character)
#' - **highlightBorder** color of highlighted borders
#'
#' @example inst/examples/ex_plot_model.R
#'
#' @export
#'
plot.RelDataModel <- function(
  x,
  ...
) {
  toPlot <- modelToVn(x, ...)

  visNetwork::visNetwork(nodes = toPlot$nodes, edges = toPlot$edges) |>
    visNetwork::visNodes(
      labelHighlightBold = FALSE,
      borderWidth = 2
    ) |>
    visNetwork::visEdges(
      # color=list(
      #    color=border,
      #    highlight=highlightBorder
      # ),
      width = 2,
      selectionWidth = 2
    ) |>
    visNetwork::visInteraction(multiselect = TRUE) |>
    ################################################@
    ## The code below is useless when edge smooth is
    ## define by edge
    # visNetwork::visPhysics(
    #    solver="repulsion",
    #    repulsion=list(
    #       nodeDistance=100,
    #       springLength=100,
    #       springConstant=0.001,
    #       damping=1,
    #       avoidOverlap=1
    #    )
    # ) |>
    ################################################@
    visNetwork::visLayout(randomSeed = 2) |>
    visNetwork::visPhysics(enabled = FALSE)
}

###############################################################################@
#' VisNetwork representation of a [RelDataModel] object
#'
#' @param model a [RelDataModel]
#' @param color default table background color
#' @param border border color (single character)
#' @param highlightBorder color of highlighted borders
#'
#'
#' Internal function
#'
modelToVn <- function(
  model,
  color = "lightgrey",
  border = "black",
  highlightBorder = "orange"
) {
  nodes <- do.call(
    rbind,
    lapply(
      model,
      function(m) {
        f <- m$fields
        pk <- m$primaryKey
        it <- index_table(m)
        ind <- NULL
        # uq <- NULL
        if (!is.null(it) && any(it$index > 0)) {
          it <- it |>
            dplyr::filter(.data$index != 0) |>
            dplyr::group_by(.data$index) |>
            dplyr::mutate(
              n = dplyr::n(),
              nn = 1:dplyr::n()
            ) |>
            dplyr::ungroup() |>
            dplyr::mutate(
              index = ifelse(
                .data$n > 1,
                paste(.data$index, .data$nn, sep = "-"),
                as.character(.data$index)
              )
            )
          ind <- unique(it$field)
          uind <- it |>
            dplyr::filter(.data$uniqueIndex) |>
            dplyr::pull("field") |>
            unique()
          # uq <- unique(it$field[which(it$unique)])
        }
        f$i <- unlist(lapply(
          f$name,
          function(n) {
            paste(sort(it$index[which(it$field == n)]), collapse = ",")
          }
        ))
        flab <- paste(
          sprintf(
            '    - %s%s%s%s%s {%s%s}%s',
            ifelse(f$nullable, "(", ""),
            ifelse(f$name %in% pk, "<b>", ""),
            # ifelse(f$unique & !f$name %in% pk, "*", ""),
            ifelse(f$unique, "*", ""),
            f$name,
            ifelse(f$name %in% pk, "</b>", ""),
            f$type,
            ifelse(
              f$name %in% ind,
              sprintf(
                "<b> - %s</b>",
                ifelse(
                  f$name %in% uind,
                  paste0("uidx.", f$i),
                  paste0("idx.", f$i)
                )
              ),
              ""
            ),
            ifelse(f$nullable, ")", "")
          ),
          collapse = "\n"
        )
        label <- paste(
          sprintf(
            '<b>%s%s</b>',
            m$tableName,
            ifelse(is.MatrixModel(m), " []", "")
          ),
          flab,
          sep = "\n"
        )
        fcomment <- gsub(
          "\\{([^}]*)\\}",
          "<b>{\\1}</b>",
          f$comment
        )
        tcomment <- gsub(
          "\\{([^}]*)\\}",
          "<b>{\\1}</b>",
          m$display$comment
        )
        ftit <- paste(
          sprintf(
            '<li><strong>%s</strong>%s%s</li>',
            f$name,
            ifelse(is.na(f$comment) | f$comment == "", "", ": "),
            ifelse(
              is.na(f$comment) | f$comment == "",
              "",
              fcomment
            )
          ),
          collapse = " "
        )
        title <- paste(
          sprintf(
            paste0(
              '<p><strong style="text-decoration:underline;">%s',
              '</strong>%s</p>'
            ),
            m$tableName,
            ifelse(
              is.na(m$display$comment),
              "",
              sprintf(" (%s)", tcomment)
            )
          ),
          "<ul>",
          ftit,
          "</ul>",
          sep = " "
        )
        title <- sprintf(
          paste0(
            '<div ',
            'style="',
            'max-width:400px; max-height:300px;',
            'overflow-y:auto; overflow-wrap:break-word;',
            'word-break:break-word; white-space:normal;',
            '">%s</div>'
          ),
          title
        )
        return(dplyr::tibble(
          tableName = m$tableName,
          label = label,
          title = title,
          shape = "box",
          font.multi = TRUE,
          font.align = "left",
          x = m$display$x,
          y = m$display$y,
          color.background = m$display$color
        ))
      }
    )
  )
  if (!is.null(nodes) && nrow(nodes) > 0) {
    nodes <- nodes |>
      dplyr::mutate(
        color.border = !!border,
        color.highlight.border = !!highlightBorder,
        color.background = ifelse(
          is.na(.data$color.background),
          !!color,
          .data$color.background
        )
      ) |>
      dplyr::mutate(
        color.highlight.background = .data$color.background
      )
    nodes$id <- names(model)
  }

  edges <- do.call(
    rbind,
    lapply(
      model,
      function(m) {
        mt <- m$tableName
        fk <- m$foreignKeys
        if (length(fk) == 0) {
          return(NULL)
        }
        toRet <- do.call(
          rbind,
          lapply(
            fk,
            function(k) {
              to <- k$refTable
              kt <- k$key |> dplyr::arrange(.data$from, .data$to)
              kcard <- ifelse(k$cardinality == -1, "n", k$cardinality)
              fcard <- paste(kcard["fmin"], kcard["fmax"], sep = "..")
              tcard <- paste(kcard["tmin"], kcard["tmax"], sep = "..")
              title <- paste0(
                '<tr style="border: 1px solid black; padding:1px;">',
                '<td style="border: 1px solid black; padding:1px;">',
                c(kt$from),
                '</td>',
                '<td style="border: 1px solid black; padding:1px;">',
                c(kt$to),
                '</td>',
                '</tr>'
              )
              title <- paste0(
                '<table style="border: 1px solid black; padding:1px;">',
                '<tr style="border: 1px solid black; padding:1px;">',
                sprintf(
                  paste0(
                    '<th style=',
                    '"border: 1px solid black; padding:1px;',
                    'text-align:center;"',
                    '>',
                    '%s<br>(%s)',
                    '</th>'
                  ),
                  mt,
                  fcard
                ),
                sprintf(
                  paste0(
                    '<th style=',
                    '"border: 1px solid black; padding:1px;',
                    'text-align:center;"',
                    '>',
                    '%s<br>(%s)',
                    '</th>'
                  ),
                  k$refTable,
                  tcard
                ),
                '</tr>',
                paste(title, collapse = ""),
                '</table>'
              )
              id <- paste(kt$from, kt$to, sep = "->")
              id <- paste(id, collapse = " && ")
              id <- paste(to, id, sep = ": ")
              return(dplyr::tibble(
                id = id,
                to = to,
                title = title,
                ff = list(kt$from),
                tf = list(kt$to)
              ))
            }
          )
        )
        toRet$from <- mt
        toRet$id <- paste(mt, toRet$id, sep = "->")
        toRet$arrows <- "to"
        toRet$font.align <- "bottom"
        return(toRet)
      }
    )
  )
  if (is.null(edges)) {
    edges <- dplyr::tibble(
      id = character(),
      from = character(),
      to = character()
    )
  } else {
    edges$smooth.type <- "curvedCCW"
    edges$smooth.roundness <- 0
    edges$selfReferenceSize <- 30
    edges <- dplyr::bind_cols(
      edges,
      edges |>
        dplyr::select("from", "to") |>
        apply(1, function(x) c(sort(x), paste(sort(x), collapse = "<->"))) |>
        t() |>
        (function(x) {
          colnames(x) <- c("uef", "uet", "ue")
          x
        })() |>
        dplyr::as_tibble()
    )
    edges <- edges |>
      dplyr::group_by(.data$ue) |>
      dplyr::mutate(
        smooth.roundness = {
          mr <- min(1, 0.2 * (length(.data$ue) %/% 2))
          seq(-mr, mr, length.out = length(.data$ue))
        },
        selfReferenceSize = {
          seq(30, 50, length.out = length(.data$ue))
        }
      ) |>
      dplyr::ungroup() |>
      dplyr::mutate(
        smooth.roundness = ifelse(
          .data$uef == .data$from,
          .data$smooth.roundness,
          -.data$smooth.roundness
        )
      ) |>
      dplyr::mutate(
        color.color = border,
        color.highlight = highlightBorder
      )
  }

  return(list(nodes = nodes, edges = edges))
}

###############################################################################@
#' Create a [RelDataModel] object from column names of data frames
#'
#' @param ... the data frame objects, as names (unquoted) or character strings
#' (quoted)
#' @param list a character vector naming data frame objects
#' @param pos where to get the objects.
#' By default, uses the current environment.
#' See ‘details’ for other possibilities.
#' @param envir the environment to use. See ‘details’.
#'
#' @details The pos argument can specify the environment from which to get
#' the objects in any of several ways:
#' as an integer (the position in the search list);
#' as the character string name of an element in the search list;
#' or as an environment.
#' The envir argument is an alternative way to specify an environment,
#' but is primarily there for back compatibility.
#'
#' @return A [RelDataModel] object.
#'
#' @example inst/examples/ex_from_df.R
#'
#' @export
#'
df_to_model <- function(
  ...,
  list = character(),
  pos = -1,
  envir = as.environment(pos)
) {
  ## Get object names ----
  ## This part has been taken from the `rm` function
  dots <- match.call(expand.dots = FALSE)$...
  if (
    length(dots) &&
      !all(vapply(
        dots,
        function(x) {
          is.symbol(x) || is.character(x)
        },
        NA,
        USE.NAMES = FALSE
      ))
  ) {
    stop("... must contain names or character strings")
  }
  names <- vapply(dots, as.character, "")
  if (length(names) == 0) {
    names <- character()
  }
  list <- c(list, names) |> unique()
  ## Get and check object values ----
  values <- lapply(list, get, envir = envir, mode = "any", inherits = TRUE)
  names(values) <- list
  isdf <- lapply(
    values,
    inherits,
    c("data.frame", "matrix", "Matrix")
  ) |>
    unlist()
  if (any(!isdf)) {
    stop(
      "The following objects are neither of class ",
      "data.frame, matrix nor Matrix: ",
      paste(names(isdf)[!isdf], collapse = ", ")
    )
  }
  ## Build the data model ----
  toRet <- list()
  for (tn in names(values)) {
    df <- values[[tn]]
    if (inherits(df, c("matrix", "Matrix"))) {
      toRet[[tn]] <- RelTableModel(
        tableName = tn,
        fields = dplyr::tibble(
          name = c("row", "column", "value"),
          type = c("row", "column", class(df[1])),
          nullable = c(FALSE, FALSE, TRUE),
          unique = c(FALSE, FALSE, FALSE),
          comment = as.character(NA)
        ),
        primaryKey = c("row", "column"),
        foreignKeys = NULL,
        indexes = NULL,
        display = list(
          x = as.numeric(NA),
          y = as.numeric(NA),
          color = as.character(NA),
          comment = as.character(NA)
        )
      )
    } else {
      types <- character()
      for (cn in colnames(df)) {
        ct <- (df |> dplyr::pull(!!cn) |> class())[1]
        if (!ct %in% SUPPTYPES) {
          stop(
            sprintf(
              'Type "%s" of column "%s" is not supported. ',
              ct,
              cn
            ),
            'Supported types are provided in "SUPPTYPES".'
          )
        }
        types <- c(types, ct)
      }
      toRet[[tn]] <- RelTableModel(
        tableName = tn,
        fields = dplyr::tibble(
          name = colnames(df),
          type = types,
          nullable = TRUE,
          unique = FALSE,
          comment = as.character(NA)
        ),
        primaryKey = NULL,
        foreignKeys = NULL,
        indexes = NULL,
        display = list(
          x = as.numeric(NA),
          y = as.numeric(NA),
          color = as.character(NA),
          comment = as.character(NA)
        )
      )
    }
  }
  return(RelDataModel(toRet))
}

###############################################################################@
#' Guess [RelDataModel] constraints based on the provided or existing tables
#'
#' @param x a [RelDataModel]
#' @param data a named list of tables. All names of x should exist in data. If
#' NULL, the data are taken from env.
#' @param env the R environment in which to find the tables
#' @param constraints the type of constraints to guess
#'
#' @return A [RelDataModel]
#'
#' @details
#' The guessed constraints should be carefully review, especially the foreign
#' keys.
#'
#' Complex foreign keys involving multiple fields are not guessed.
#'
#' @example inst/examples/ex_from_df.R
#'
#' @export
#'
guess_constraints <- function(
  x,
  data = NULL,
  env = parent.frame(n = 1),
  constraints = c("unique", "not nullable", "foreign keys")
) {
  dml <- unclass(x)
  if (is.null(data)) {
    data <- lapply(names(x), get, envir = env) |>
      stats::setNames(names(x))
  }
  stopifnot(all(names(x) %in% names(data)))
  stopifnot(all(unlist(
    lapply(data, inherits, what = c("matrix", "Matrix", "data.frame"))
  )))
  for (n in names(dml)) {
    d <- data[[n]]
    if (is.data.frame(d)) {
      if ("not nullable" %in% constraints) {
        nullable <- lapply(
          d,
          function(x) any(is.na(x))
        ) |>
          unlist()
        dml[[n]]$fields$nullable <- nullable
      }
      if ("unique" %in% constraints) {
        uni <- lapply(
          d,
          function(x) !is.numeric(x) && !any(duplicated(x))
        ) |>
          unlist()
        dml[[n]]$fields$unique <- uni
      }
    } else {
      if ("not nullable" %in% constraints) {
        nullable <- any(is.na(d))
        dml[[n]]$fields$nullable[which(
          !dml[[n]]$fields$type %in% c("row", "column")
        )] <- nullable
      }
    }
  }
  dm <- RelDataModel(dml)
  if ("foreign keys" %in% constraints) {
    dbm <- toDBM(dm)
    pfk <- dplyr::inner_join(
      dbm$fields |>
        dplyr::filter(!.data$nullable & .data$unique) |>
        select(-"nullable", -"unique", -"comment", -"fieldOrder"),
      dbm$fields |>
        select(-"comment", -"fieldOrder"),
      by = c("name", "type"),
      suffix = c(".p", ".f"),
      relationship = "many-to-many"
    ) |>
      dplyr::filter(.data$table.p != .data$table.f)
    if (nrow(pfk) > 0) {
      drows <- lapply(data, nrow) |> unlist()
      pfk <- pfk |>
        dplyr::mutate(
          nrow.p = drows[.data$table.p],
          nrow.f = drows[.data$table.f]
        ) |>
        dplyr::filter(
          !.data$unique | .data$nrow.f <= .data$nrow.p
        )
      if (nrow(pfk) > 0) {
        for (i in 1:nrow(pfk)) {
          if (
            all(
              setdiff(data[[pfk$table.f[i]]][[pfk$name[i]]], NA) %in%
                data[[pfk$table.p[i]]][[pfk$name[i]]]
            )
          ) {
            dm <- add_foreign_key(
              dm,
              fromTable = pfk$table.f[i],
              fromFields = pfk$name[i],
              toTable = pfk$table.p[i],
              toFields = pfk$name[i],
              tmin = ifelse(pfk$nullable[i], 0L, 1L)
            )
          }
        }
      }
    }
  }
  return(dm)
}

###############################################################################@
#' Read a data model from JSON
#'
#' @param txt a JSON string, URL or file
#'
#' @example inst/examples/ex_read_json.R
#'
#' @export
#'
read_json_data_model <- function(txt) {
  mfj <- jsonlite::fromJSON(txt, simplifyVector = FALSE)
  lapply(
    mfj,
    function(x) {
      ## table name ----
      x$tableName <- unlist(x$tableName)
      ## fields ----
      if (length(x$fields) > 0) {
        x$fields <- do.call(rbind, lapply(x$fields, dplyr::as_tibble)) |>
          dplyr::mutate(
            name = as.character(.data$name),
            type = as.character(.data$type),
            nullable = as.logical(.data$nullable),
            unique = as.logical(.data$unique),
            comment = as.character(.data$comment)
          )
      } else {
        x$fields <- dplyr::tibble(
          name = character(),
          type = character(),
          nullable = logical(),
          unique = logical(),
          comment = character()
        )
      }
      ## primary key ----
      x$primaryKey <- as.character(x$primaryKey)
      ## foreign keys ----
      x$foreignKeys <- lapply(
        x$foreignKeys,
        function(y) {
          y$refTable <- unlist(y$refTable)
          y$key <- do.call(rbind, lapply(y$key, dplyr::as_tibble))
          y$cardinality <- as.integer(unlist(y$cardinality))
          names(y$cardinality) <- c("fmin", "fmax", "tmin", "tmax")
          return(y)
        }
      )
      ## indexes ----
      x$indexes <- lapply(
        x$indexes,
        function(y) {
          y$fields <- unlist(y$fields)
          y$unique <- unlist(y$unique)
          return(y)
        }
      )
      ## display ----
      x$display <- lapply(x$display, function(y) {
        y <- unlist(y)
        ifelse(y == "NA", NA, y)
      })
      x$display$x <- as.numeric(x$display$x)
      x$display$y <- as.numeric(x$display$y)
      x$display$color <- as.character(x$display$color)
      x$display$comment <- as.character(x$display$comment)
      do.call(RelTableModel, x)
    }
  ) |>
    RelDataModel()
}

###############################################################################@
#' Write a data model in a JSON file
#'
#' @param x the model to be written
#' @param path file on disk
#'
#' @export
#'
write_json_data_model <- function(x, path) {
  jsonlite::write_json(x, path = path, na = "string", pretty = TRUE)
}

###############################################################################@
#' Create a RelTableModel object
#'
#' @param l DEPRECATED. A named list with the function parameters.
#' If `NULL` (default) the function parameters are used. If not `NULL`, the
#' function parameters are ignored and taken from l.
#' @param tableName a character vector of length one
#' @param fields a tibble with the following columns:
#'    - *name*: character
#'    - *type*: character
#'    - *nullable*: logical (optional, defaults to TRUE)
#'    - *unique*: logical (optional, defaults = FALSE)
#'    - *comment*:  character (optional, defaults to NA_character_)
#' @param primaryKey a character vector of any length. All
#' values should be in fields$name
#' @param foreignKeys a list of foreign keys. Each foreign key is defined
#' as a list with the following elements:
#'    - *refTable*: a character vector of length one (the referenced table)
#'    - *key*: a tibble with a "from" and a "to" columns
#'    - (*cardinality*): an optional integer vector with 4 values:
#'       + fmin: from minimum cardinality
#'       + fmax: from maximum cardinality
#'       + tmin: to minimum cardinality
#'       + tmax: to maximum cardinality
#' @param indexes a list of indexes. Each index is defined by
#' 3 columns:
#'    - *field*: character (all in fields$name)
#'    - *order*: character
#'    - *unique*: logical
#' @param display a list gathering:
#'    - *x*: single numeric value for the x position of the table
#'    - *y*: single numeric value for the y position of the table
#'    - *color*: single character value corresponding to the color of the table
#'    - *comment*: single character value with some description of the table
#'
#' @details When defining a matrix, 3 and only 3 fields must be defined:
#' 2 of types 'row' and 'column' and the 3rd of your choice. In this case
#' primaryKey is defined automatically as the combination of row and column.
#'
#' @return A RelTableModel object.
#'
#' @export
#'
RelTableModel <- function(
  l = NULL,
  tableName,
  fields,
  primaryKey = NULL,
  foreignKeys = NULL,
  indexes = NULL,
  display = list(
    x = as.numeric(NA),
    y = as.numeric(NA),
    color = as.character(NA),
    comment = as.character(NA)
  )
) {
  ############################################################################@
  ## Management of params ----
  if (!is.null(l)) {
    warning("l param is deprecated and will probably removed in the future")
  } else {
    l <- list(
      tableName = tableName,
      fields = fields,
      primaryKey = primaryKey,
      foreignKeys = foreignKeys,
      indexes = indexes,
      display = display
    )
  }

  ############################################################################@
  ## Checks ----

  ### Simple checks ----
  tableInfo <- c(
    "tableName",
    "fields",
    "primaryKey",
    "foreignKeys",
    "indexes",
    "display"
  )
  mandatoryFieldInfo <- c("name", "type")
  fieldInfo <- c("name", "type", "nullable", "unique", "comment")
  stopifnot(
    is.data.frame(l$fields),
    all(mandatoryFieldInfo %in% colnames(l$fields))
  )
  if (!('nullable' %in% colnames(l$fields))) {
    l$fields$nullable <- TRUE
  }
  if (!('unique' %in% colnames(l$fields))) {
    l$fields$unique <- FALSE
  }
  if (!('comment' %in% colnames(l$fields))) {
    l$fields$comment <- NA_character_
  }
  stopifnot(
    all(tableInfo %in% names(l)),
    all(names(l) %in% tableInfo),

    is.character(l$tableName),
    length(l$tableName) == 1,
    !is.na(l$tableName),

    all(colnames(l$fields) %in% fieldInfo),
    # nrow(l$fields) > 0,
    is.character(l$fields$name),
    all(!is.na(l$fields$name)),
    !any(duplicated(l$fields$name)),
    is.character(l$fields$type),
    all(!is.na(l$fields$type)),
    is.logical(l$fields$nullable),
    all(!is.na(l$fields$nullable)),
    is.logical(l$fields$unique),
    all(!is.na(l$fields$unique)),
    is.character(l$fields$comment),

    is.null(l$primaryKey) || is.character(l$primaryKey),
    all(!is.na(l$primaryKey)),
    all(l$primaryKey %in% l$fields$name)
  )
  for (att in setdiff(
    names(attributes(l$fields)),
    c("row.names", "names", "class")
  )) {
    attr(l$fields, att) <- NULL
  }
  l$fields <- dplyr::as_tibble(l$fields) |>
    dplyr::select(c("name", "type", "nullable", "unique", "comment")) |>
    dplyr::mutate(
      "comment" = as.character(
        ifelse(is.na(.data$comment), "", .data$comment)
      )
    )

  ### Primary key ----
  l$primaryKey <- sort(l$primaryKey)
  if (length(l$primaryKey) == 1) {
    l$fields[
      which(l$fields$name == l$primaryKey),
      "unique"
    ] <- TRUE
  }

  ### Matrix and Field types ----
  if (any(l$fields$type %in% c("row", "column"))) {
    if (!all(c("row", "column") %in% l$fields$type)) {
      stop("Both row and column should be provided when modelling a matrix")
    }
    if (nrow(l$fields) != 3) {
      stop("A matrix model should have 3 and only 3 fields")
    }
    if (length(unique(l$fields$type)) != 3) {
      stop(paste(
        "A matrix model should have 3 fields":"2 of types 'row' and 'column' and the 3rd of your choice"
      ))
    }
    rcfields <- l$fields |> dplyr::filter(.data$type %in% c("row", "column"))
    if (any(rcfields$nullable)) {
      stop("Matrix row and column cannot be nullable")
    }
    if (any(rcfields$unique)) {
      stop(paste(
        "The combination of row and column is unique (primary key)",
        " but not row names and column names independently"
      ))
    }
    l$primaryKey <- rcfields$name
    if (setdiff(l$fields$type, c("row", "column")) == "base64") {
      stop("A matrix cannot store base64 documents")
    }
    check_types(setdiff(l$fields$type, c("row", "column")))
  } else {
    check_types(l$fields$type)
  }

  ### Foreign keys ----
  if (!is.null(l$foreignKeys)) {
    stopifnot(is.list(l$foreignKeys))
    fkn <- c("refTable", "key")
    l$foreignKeys <- lapply(
      l$foreignKeys,
      function(fko) {
        fk <- fko[fkn]
        stopifnot(
          all(names(fk) %in% fkn),
          all(fkn %in% names(fk)),
          !is.na(fk$refTable),
          is.character(fk$refTable),
          length(fk$refTable) == 1,
          !is.na(fk$refTable),
          is.data.frame(fk$key),
          all(c("from", "to") %in% names(fk$key)),
          all(names(fk$key) %in% c("from", "to")),
          is.character(fk$key$from),
          all(!is.na(fk$key$from)),
          is.character(fk$key$to),
          all(!is.na(fk$key$to)),
          all(fk$key$from %in% l$fields$name)
        )
        for (att in setdiff(
          names(attributes(fk$key)),
          c("row.names", "names", "class")
        )) {
          attr(fk$key, att) <- NULL
        }
        fk$key <- dplyr::as_tibble(fk$key)
        cnames <- c("fmin", "fmax", "tmin", "tmax")
        if ("cardinality" %in% names(fko)) {
          stopifnot(
            is.integer(fko$cardinality),
            length(fko$cardinality) == 4,
            all(
              cnames %in%
                names(fko$cardinality)
            ),
            all(!is.na(fko$cardinality)),
            all(fko$cardinality >= -1),
            fko$cardinality["fmin"] > -1,
            fko$cardinality["tmin"] > -1,
            fko$cardinality["fmax"] == -1 ||
              (fko$cardinality["fmax"] > 0 &&
                fko$cardinality["fmax"] >= fko$cardinality["fmin"]),
            fko$cardinality["tmax"] == -1 ||
              (fko$cardinality["tmax"] > 0 &&
                fko$cardinality["tmax"] >= fko$cardinality["tmin"])
          )
          fk$cardinality <- fko$cardinality[cnames]
        } else {
          fk$cardinality <- c(0, -1, 1, 1)
          names(fk$cardinality) <- cnames
        }
        return(fk)
      }
    )
    names(l$foreignKeys) <- NULL
  }

  ### Indexes ----
  if (!is.null(l$indexes)) {
    stopifnot(is.list(l$indexes))
    # idn <- c("field", "unique")
    idn <- c("fields", "unique")
    l$indexes <- lapply(
      l$indexes,
      function(ind) {
        stopifnot(
          is.list(ind),
          all(names(ind) %in% idn),
          all(idn %in% names(ind)),
          is.character(ind$fields),
          all(!is.na(ind$fields)),
          is.logical(ind$unique),
          all(!is.na(ind$unique)),
          length(ind$unique) == 1,
          all(ind$fields %in% l$fields$name)
        )
        ind$fields <- unique(ind$fields)
        return(ind)
      }
    )
  }

  ### Display ----
  if (is.null(l$display)) {
    l$display <- list(
      x = as.numeric(NA),
      y = as.numeric(NA),
      color = as.character(NA),
      comment = as.character(NA)
    )
  }
  stopifnot(is.list(l$display))
  dn <- c("x", "y", "color", "comment")
  stopifnot(
    all(names(l$display) %in% dn),
    all(dn %in% names(l$display)),
    is.numeric(l$display$x),
    is.numeric(l$display$y),
    is.character(l$display$color),
    is.character(l$display$comment),
    all(unlist(lapply(l$display, length)) == 1)
  )

  ############################################################################@
  ## Creating the object ----
  toRet <- l
  class(toRet) <- c("RelTableModel", class(toRet))
  toRet <- correct_constraints(toRet)
  return(toRet)
}

###############################################################################@
#' Check if the object is  a [RelTableModel] object
#'
#' @param x any object
#'
#' @return A single logical: TRUE if x is a [RelTableModel] object
#'
#' @export
#'
is.RelTableModel <- function(x) {
  inherits(x, "RelTableModel")
}

###############################################################################@
#' Check if the object is a [RelTableModel] matrix object
#'
#' A matrix model is a special [RelTableModel] object with 3 and only 3 fields:
#' 2 of types 'row' and 'column' and the 3rd of your choice.
#'
#' @param x any object
#'
#' @return A single logical: TRUE if x is a [RelTableModel] matrix object
#'
#' @export
#'
is.MatrixModel <- function(x) {
  inherits(x, "RelTableModel") &&
    "row" %in% x$fields$type
}

###############################################################################@
#' Format a [RelTableModel] object for printing
#'
#' @param x a [RelTableModel] object
#' @param ... for generics compatibility (not used)
#'
#' @return A single character
#'
#' @export
#'
format.RelTableModel <- function(x, ...) {
  f <- x$fields
  pk <- x$primaryKey
  it <- index_table(x)
  ind <- NULL
  # uq <- NULL
  if (!is.null(it)) {
    it <- it |> dplyr::filter(.data$index != 0)
    ind <- unique(it$field)
    # uq <- unique(it$field[which(it$unique)])
  }
  f$i <- unlist(lapply(
    f$name,
    function(n) {
      paste(sort(it$index[which(it$field == n)]), collapse = ",")
    }
  ))
  toRet <- paste0(
    sprintf(
      "Table name: %s%s",
      x$tableName,
      ifelse(is.MatrixModel(x), " (matrix)", "")
    ),
    "\n",
    paste(
      apply(
        f,
        1,
        function(y) {
          toRet <- "   "
          toRet <- paste0(
            toRet,
            ifelse(y[1] %in% pk, "*", " "),
            # ifelse(y[1] %in% uq, "+ ", "  "),
            ifelse(y[4] == "TRUE" & !y[1] %in% pk, "+ ", "  "),
            y[1]
          )
          toRet <- paste0(
            toRet,
            " (",
            y[2],
            ifelse(
              y[1] %in% ind,
              # paste0(", idx.", y[5]),
              paste0(", idx.", y[6]),
              ""
            ),
            ifelse(y[3] == "FALSE", ", not", ","),
            " nullable",
            ")"
          )
          return(toRet)
        }
      ),
      collapse = "\n"
    )
  )
  if (length(x$foreignKeys) > 0) {
    toRet <- paste0(
      toRet,
      "\n",
      "Referenced tables:\n",
      paste(
        unlist(lapply(
          x$foreignKeys,
          function(y) {
            y$refTable
          }
        )),
        collapse = "\n"
      )
    )
  }
  return(toRet)
}

###############################################################################@
#' @export
#'
print.RelTableModel <- function(x, ...) {
  cat(format(x, ...), "\n")
}


###############################################################################@
#' @export
#'
length.RelTableModel <- function(x) {
  nrow(x$fields)
}


###############################################################################@
#' @export
#'
lengths.RelTableModel <- function(x, use.names = TRUE) {
  lengths(unclass(x), use.names = use.names)
}


###############################################################################@
#' List indexes of a [RelTableModel] object
#'
#' @param x a [RelTableModel] object
#'
#' @return A tibble with the following columns:
#' - **index**: an integer corresponding to the index number
#' - **field**: a character corresponding to field belonging to the index
#' - **unique**: a logical indicating the uniqueness of the field
#'
#' @export
#'
index_table <- function(x) {
  stopifnot(is.RelTableModel(x))
  pk <- x$primaryKey
  ind <- x$indexes
  toRet <- NULL
  i <- 0
  if (length(pk) > 0) {
    toRet <- dplyr::tibble(
      index = i,
      field = pk,
      # unique=length(pk)==1
      uniqueIndex = TRUE
    )
  }
  for (ci in ind) {
    i <- i + 1
    toRet <- dplyr::bind_rows(
      toRet,
      dplyr::tibble(
        index = i,
        field = ci$fields,
        uniqueIndex = ci$unique
      )
    )
  }
  return(toRet)
}

###############################################################################@
#' Get the types of the columns of a [RelTableModel] object
#'
#' @param x a [RelTableModel] object
#'
#' @return A col_spec object with the type of each column
#'
#' @export
#'
col_types <- function(x) {
  stopifnot(is.RelTableModel(x))
  do.call(
    readr::cols,
    structure(
      lapply(
        x$fields$type,
        function(y) {
          switch(
            y,
            "row" = readr::col_character(),
            "column" = readr::col_character(),
            "integer" = readr::col_integer(),
            "numeric" = readr::col_double(),
            "logical" = readr::col_logical(),
            "character" = readr::col_character(),
            "Date" = readr::col_date(),
            "POSIXct" = readr::col_datetime(),
            "base64" = readr::col_character()
          )
        }
      ),
      .Names = x$fields$name
    )
  )
}

###############################################################################@
#' Correct the constraints of a table to make them consistent
#'
#' @param x a [RelTableModel] object
#' @param createPKIndex should an index be create for the primary key
#' (default: FALSE)
#'
#' @export
#'
correct_constraints <- function(x, createPKIndex = FALSE) {
  stopifnot(is.RelTableModel(x))
  ## Primary key uniqueness
  if (length(x$primaryKey) == 1) {
    x$fields[which(x$fields$name == x$primaryKey), "unique"] <- TRUE
    x$fields[which(x$fields$name == x$primaryKey), "nullable"] <- FALSE
  }
  if (length(x$primaryKey) > 0) {
    ei <- lapply(
      x$indexes,
      function(y) {
        identical(sort(y$fields), sort(x$primaryKey))
      }
    ) |>
      unlist() |>
      as.logical() |>
      which()
    if (length(ei) > 1) {
      stop("Check this part of code")
    }
    if (length(ei) == 0) {
      if (createPKIndex) {
        x$indexes <- unique(c(
          x$indexes,
          list(
            list(
              fields = sort(x$primaryKey),
              unique = TRUE
            )
          )
        ))
      }
    } else {
      x$indexes[[ei]]$unique <- TRUE
    }
  }
  ## Index uniqueness
  if (length(x$indexes) > 0) {
    for (i in 1:length(x$indexes)) {
      if (x$indexes[[i]]$unique && length(x$indexes[[i]]$fields) == 1) {
        if (
          x$fields[
            which(x$fields$name == x$indexes[[i]]$fields),
            "type"
          ] %in%
            c("row", "column")
        ) {
          stop("Matrix row and column cannot be unique individually")
        }
        x$fields[
          which(x$fields$name == x$indexes[[i]]$fields),
          "unique"
        ] <- TRUE
      }
    }
  }
  ## Field uniqueness
  uniqueFields <- x$fields |>
    dplyr::filter(.data$unique) |>
    dplyr::pull("name")
  if (length(x$indexes) > 0 && length(uniqueFields) > 0) {
    for (i in 1:length(x$indexes)) {
      if (any(x$indexes[[i]]$fields %in% uniqueFields)) {
        x$indexes[[i]]$unique <- TRUE
      }
    }
  }
  ##
  return(x)
}

###############################################################################@
#' Confront a [RelTableModel] to actual data
#'
#' @param x a [RelTableModel]
#' @param d a data frame or a matrix for matrix model
#' @param checks a character vector with the name of optional checks to be done
#' (Default: all of them c("unique", "not nullable"))
#'
#' @return A report as a list
#'
#' @export
#'
confront_table_data <- function(
  x,
  d,
  checks = c("unique", "not nullable")
) {
  ## Optional checks ----
  if (length(checks) > 0) {
    checks <- match.arg(
      checks,
      c("unique", "not nullable"),
      several.ok = TRUE
    )
  }
  stopifnot(is.RelTableModel(x))
  ## Matrix model ----
  if (is.MatrixModel(x)) {
    if (!inherits(d, c("matrix", "Matrix"))) {
      stop(x$tableName, " is neither a matrix nor a Matrix")
    }
    vf <- x$fields |> dplyr::filter(!.data$type %in% c("row", "column"))
    toRet <- list(
      missingFields = character(0),
      suppFields = character(0),
      availableFields = vf$name,
      fields = list(),
      success = TRUE
    )
    toRet$fields[[vf$name]] <- list(success = TRUE, message = NULL)
    if (!inherits(d[1], vf$type)) {
      toRet$fields[[vf$name]]$success <- FALSE
      toRet$success <- FALSE
      toRet$fields[[vf$name]]$message <- paste(
        c(
          toRet$fields[[vf$name]]$message,
          sprintf(
            'Unexpected "%s"',
            paste(class(d[1]), collapse = ", ")
          )
        ),
        collapse = " "
      )
    }
    if ("not nullable" %in% checks) {
      mis <- sum(is.na(d))
      if (mis != 0) {
        toRet$fields[[vf$name]]$message <- paste(
          c(
            toRet$fields[[vf$name]]$message,
            sprintf(
              'Missing values %s/%s = %s%s',
              mis,
              length(d),
              round(mis * 100 / length(d)),
              "%"
            )
          ),
          collapse = " "
        )
        if (!vf$nullable) {
          toRet$fields[[vf$name]]$success <- FALSE
          toRet$success <- FALSE
        }
      }
    }
    if ("unique" %in% checks && vf$unique && any(duplicated(d))) {
      toRet$fields[[vf$name]]$success <- FALSE
      toRet$success <- FALSE
      toRet$fields[[vf$name]]$message <- paste(
        c(
          toRet$fields[[vf$name]]$message,
          "Some values are duplicated"
        ),
        collapse = " "
      )
    }
    return(toRet)
  }

  if (!inherits(d, "data.frame")) {
    stop(x$tableName, " is not a data.frame")
  }
  ## Available fields ----
  missingFields <- setdiff(x$fields$name, colnames(d))
  suppFields <- setdiff(colnames(d), x$fields$name)
  availableFields <- intersect(colnames(d), x$fields$name)
  toRet <- list(
    missingFields = missingFields,
    suppFields = suppFields,
    availableFields = availableFields,
    fields = list(),
    success = length(missingFields) == 0 && length(suppFields) == 0
  )
  ## Fields ----
  for (fn in availableFields) {
    ft <- x$fields$type[which(x$fields$name == fn)]
    ft <- ifelse(ft %in% c("row", "column"), "character", ft)
    fe <- x$fields$nullable[which(x$fields$name == fn)]
    fu <- x$fields$unique[which(x$fields$name == fn)]
    toRet$fields[[fn]] <- list(success = TRUE, message = NULL)
    if (!inherits(dplyr::pull(d, !!fn), ft)) {
      if (ft != "base64" || !inherits(dplyr::pull(d, !!fn), "character")) {
        toRet$fields[[fn]]$success <- FALSE
        toRet$success <- FALSE
        toRet$fields[[fn]]$message <- paste(
          c(
            toRet$fields[[fn]]$message,
            sprintf(
              'Unexpected "%s"',
              paste(class(dplyr::pull(d, !!fn)), collapse = ", ")
            )
          ),
          collapse = " "
        )
      }
    }
    if ("not nullable" %in% checks) {
      mis <- sum(is.na(dplyr::pull(d, !!fn)))
      if (mis != 0) {
        toRet$fields[[fn]]$message <- paste(
          c(
            toRet$fields[[fn]]$message,
            sprintf(
              'Missing values %s/%s = %s%s',
              mis,
              nrow(d),
              round(mis * 100 / nrow(d)),
              "%"
            )
          ),
          collapse = " "
        )
        if (!fe) {
          toRet$fields[[fn]]$success <- FALSE
          toRet$success <- FALSE
        }
      }
    }
    toTest <- dplyr::pull(d, !!fn)
    toTest <- toTest[which(!is.na(toTest))]
    if ("unique" %in% checks && fu && any(duplicated(toTest))) {
      toRet$fields[[fn]]$success <- FALSE
      toRet$success <- FALSE
      toRet$fields[[fn]]$message <- paste(
        c(
          toRet$fields[[fn]]$message,
          "Some values are duplicated"
        ),
        collapse = " "
      )
    }
  }
  ## Indexes ----
  if ("unique" %in% checks && length(x$indexes) > 0) {
    toRet$indexes <- list()
    for (i in 1:length(x$indexes)) {
      toRet$indexes[[i]] <- list()
      idx <- x$indexes[[i]]
      if (any(idx$fields %in% missingFields)) {
        toRet$indexes[[i]]$success <- FALSE
        toRet$indexes[[i]]$message <- paste(
          c(
            toRet$indexes[[i]]$message,
            "Missing field"
          ),
          collapse = " "
        )
        toRet$success <- FALSE
      } else {
        if (!idx$unique) {
          toRet$indexes[[i]]$success <- TRUE
        } else {
          toTest <- dplyr::filter(
            d[, idx$fields],
            stats::complete.cases(d[, idx$fields])
          )
          if (nrow(dplyr::distinct(toTest)) != nrow(toTest)) {
            toRet$indexes[[i]]$success <- FALSE
            toRet$indexes[[i]]$message <- paste(
              c(
                toRet$indexes[[i]]$message,
                "Some values are duplicated"
              ),
              collapse = " "
            )
            toRet$success <- FALSE
          } else {
            toRet$indexes[[i]]$success <- TRUE
          }
        }
      }
    }
  }
  if ("not nullable" %in% checks && is.MatrixModel(x)) {
    if (!"indexes" %in% names(toRet)) {
      toRet$indexes <- list()
    }
    i <- length(toRet$indexes) + 1
    rf <- x$fields |>
      dplyr::filter(.data$type == "row") |>
      dplyr::pull("name")
    cf <- x$fields |>
      dplyr::filter(.data$type == "column") |>
      dplyr::pull("name")
    fe <- x$fields |>
      dplyr::filter(!.data$type %in% c("row", "column")) |>
      dplyr::pull("nullable")
    ncells <- length(unique(d[, rf])) * length(unique(d[, cf]))
    mis <- ncells - nrow(d)
    if (mis > 0) {
      toRet$indexes[[i]]$message <- paste(
        c(
          toRet$indexes[[i]]$message,
          sprintf(
            'Missing cells %s/%s = %s%s',
            mis,
            ncells,
            round(mis * 100 / ncells),
            "%"
          )
        ),
        collapse = " "
      )
      if (!fe) {
        toRet$indexes[[i]]$success <- FALSE
        toRet$success <- FALSE
      }
    }
  }

  ## Return the results ----
  return(toRet)
}


###############################################################################@
#' Check if two [RelTableModel] are identical
#'
#' @param x a [RelTableModel]
#' @param y a [RelTableModel]
#' @param includeDisplay a single logical (default: TRUE) indicating if
#' the display should be included in the comparison
#'
#' @return A logical: TRUE if the 2 models are identical
#'
#' @export
#'
identical_RelTableModel <- function(x, y, includeDisplay = TRUE) {
  stopifnot(is.RelTableModel(x), is.RelTableModel(y))

  ## Name ----
  toRet <- x$tableName == y$tableName

  ## Fields ----
  toRet <- toRet &&
    identical(
      x$fields |> dplyr::arrange(.data$name),
      y$fields |> dplyr::arrange(.data$name)
    )

  ## Primary key ----
  toRet <- toRet &&
    length(x$primaryKey) == length(y$primaryKey) &&
    all(x$primaryKey == y$primaryKey)

  ## Indexes ----
  toRet <- toRet && length(x$indexes) == length(y$indexes)
  if (toRet && length(x$indexes) > 0) {
    # xidx <- lapply(
    #   x$indexes,
    #   function(z) {
    #     z$fields <- sort(z$fields)
    #     return(z)
    #   }
    # )
    xidx <- x$indexes
    xidx <- xidx[order(unlist(lapply(
      xidx,
      function(z) paste(z$fields, collapse = ", ")
    )))]
    # yidx <- lapply(
    #   y$indexes,
    #   function(z) {
    #     z$fields <- sort(z$fields)
    #     return(z)
    #   }
    # )
    yidx <- y$indexes
    yidx <- yidx[order(unlist(lapply(
      yidx,
      function(z) paste(z$fields, collapse = ", ")
    )))]
    toRet <- identical(xidx, yidx)
  }

  ## Foreign keys ----
  toRet <- toRet && length(x$foreignKeys) == length(y$foreignKeys)
  if (toRet && length(x$foreignKeys) > 0) {
    xfk <- lapply(
      x$foreignKeys,
      function(z) {
        z$key <- z$key |> dplyr::arrange(.data$from, .data$to)
        return(z)
      }
    )
    xfk <- xfk[order(unlist(lapply(
      xfk,
      function(z) {
        paste0(
          z$refTable,
          " [",
          paste(
            paste(z$key$from, z$key$to, sep = "->"),
            collapse = ", "
          ),
          "]"
        )
      }
    )))]
    yfk <- lapply(
      y$foreignKeys,
      function(z) {
        z$key <- z$key |> dplyr::arrange(.data$from, .data$to)
        return(z)
      }
    )
    yfk <- yfk[order(unlist(lapply(
      yfk,
      function(z) {
        paste0(
          z$refTable,
          " [",
          paste(
            paste(z$key$from, z$key$to, sep = "->"),
            collapse = ", "
          ),
          "]"
        )
      }
    )))]
    toRet <- identical(xfk, yfk)
  }

  ## Display ----
  if (includeDisplay) {
    toRet <- toRet &&
      length(x$display) == length(y$display) &&
      (length(x$display) == 0 ||
        identical(
          x$display[order(names(x$display))],
          y$display[order(names(y$display))]
        ))
  }
  return(toRet)
}

###############################################################################@
#' Get foreign keys from [RelTableModel]
#'
#' @param x a [RelTableModel]
#'
#' @return A tibble with the following fields:
#' - from: the origin of the key
#' - ff: the key fields in from
#' - to: the target of the key
#' - tf: the key fields in to
#' - fmin: minimum cardinality of from
#' - fmax: maximum cardinality of from
#' - tmin: minimum cardinality of to
#' - tmax: maximum cardinality of to
#'
#' @export
#'
get_foreign_keys.RelTableModel <- function(x) {
  tn <- x$tableName
  fk <- x$foreignKeys
  if (length(fk) == 0) {
    return(NULL)
  }
  toRet <- do.call(
    rbind,
    lapply(
      fk,
      function(k) {
        to <- k$refTable
        kt <- k$key |> dplyr::arrange(.data$from, .data$to)
        dplyr::tibble(
          to = to,
          ff = list(kt$from),
          tf = list(kt$to)
        ) |>
          dplyr::bind_cols(dplyr::as_tibble(t(k$cardinality)))
      }
    )
  )
  toRet <- toRet |>
    dplyr::mutate(
      from = tn
    ) |>
    dplyr::select(
      "from",
      "ff",
      "to",
      "tf",
      "fmin",
      "fmax",
      "tmin",
      "tmax"
    )
  return(toRet)
}

TYPETABLE <- list(
  MySQLWB = c(
    "^TINYINT([()].*[)])?$",
    "TINYINT",
    "logical",
    "^BIGINT([()].*[)])?$",
    "BIGINT",
    "integer",
    "^INT([()].*[)])?$",
    "INT",
    "integer",
    "^SMALLINT([()].*[)])?$",
    "SMALLINT",
    "integer",
    "^DOUBLE([()].*[)])?$",
    "DOUBLE",
    "numeric",
    "^FLOAT([()].*[)])?$",
    "FLOAT",
    "numeric",
    "^DECIMAL([()].*[)])?$",
    "DECIMAL",
    "numeric",
    "^VARCHAR([()].*[)])?$",
    "VARCHAR",
    "character",
    "^TEXT([()].*[)])?$",
    "TEXT",
    "character",
    "^TINYTEXT([()].*[)])?$",
    "TINYTEXT",
    "character",
    "^LONGTEXT([()].*[)])?$",
    "LONGTEXT",
    "character",
    "^MEDIUMTEXT([()].*[)])?$",
    "MEDIUMTEXT",
    "character",
    "^BLOB([()].*[)])?$",
    "BLOB",
    "base64",
    "^LONGBLOB([()].*[)])?$",
    "LONGBLOB",
    "base64",
    "^MEDIUMBLOB([()].*[)])?$",
    "MEDIUMBLOB",
    "base64",
    "^TINYBLOB([()].*[)])?$",
    "TINYBLOB",
    "base64",
    "^DATETIME([()].*[)])?$",
    "DATETIME",
    "POSIXct",
    "^DATE([()].*[)])?$",
    "DATE",
    "Date",
    "^ENUM([()].*[)])?$",
    "ENUM",
    "character",
    "^SET([()].*[)])?$",
    "SET",
    "character"
  ) |>
    matrix(
      ncol = 3,
      byrow = TRUE,
      dimnames = list(NULL, c("match", "inst", "R"))
    ) |>
    dplyr::as_tibble(),
  ClickHouse = c(
    "Int32",
    "Int32",
    "integer",
    "Float64",
    "Float64",
    "numeric",
    "UInt8",
    "UInt8",
    "logical",
    "String",
    "String",
    "character",
    "Date",
    "Date",
    "Date",
    "DateTime",
    "DateTime",
    "POSIXct",
    "Array(String)",
    "Array(String)",
    "base64"
  ) |>
    matrix(
      ncol = 3,
      byrow = TRUE,
      dimnames = list(NULL, c("match", "inst", "R"))
    ) |>
    dplyr::as_tibble()
)

###############################################################################@
#' Supported R types
#'
#' @export
#'
SUPPTYPES <- c(
  "integer",
  "numeric",
  "logical",
  "character",
  "Date",
  "POSIXct",
  "base64"
)

###############################################################################@
#' List supported types references
#'
#' @export
#'
list_type_ref <- function() {
  names(TYPETABLE)
}

###############################################################################@
#' Normalize type names
#'
#' @param x a character vector to normalize
#' @param typeRef a character vector of length one: the type
#' reference ([list_type_ref])
#' @param ignore.case should case be ignored (default: TRUE)
#'
#' @export
#'
norm_type_ref <- function(x, typeRef, ignore.case = TRUE) {
  stopifnot(is.character(x))
  typeRef <- match.arg(typeRef, list_type_ref())
  ntypes <- unique(TYPETABLE[[typeRef]] |> dplyr::pull("match"))
  toRet <- x
  for (nt in ntypes) {
    toRet <- sub(
      nt,
      nt,
      toRet,
      ignore.case = ignore.case
    )
  }
  return(toRet)
}

###############################################################################@
#' Convert a set of types from or to R supported types
#'
#' @param x a character vector of types to be converted. If from is not null,
#' x should be a set of valid types in the from reference. If to is not null,
#' x should be a set of supported R types (SUPPTYPES).
#' @param from a character vector of length one: the type reference
#' ([list_type_ref]) of x
#' @param to a character vector of length one: the targeted type reference
#' ([list_type_ref])
#' @param ignore.case should case be ignored when converting `from``
#' type reference (default: TRUE)
#'
#' @details Only `from` XOR `to` should be set
#'
#' @export
#'
conv_type_ref <- function(x, from = NULL, to = NULL, ignore.case = TRUE) {
  stopifnot(
    !is.null(from) | !is.null(to),
    is.null(from) | is.null(to)
  )
  if (!is.null(from)) {
    r <- match.arg(from, list_type_ref())
    ct <- TYPETABLE[[r]]
    x <- norm_type_ref(x, from, ignore.case = ignore.case)
    notSupported <- setdiff(x, ct |> dplyr::pull("match"))
    if (length(notSupported) > 0) {
      stop(paste(
        sprintf("The following types are not supported %s types:", r),
        paste(notSupported, collapse = ", ")
      ))
    }
    toRet <- ct |> dplyr::slice(match(x, .data$match)) |> dplyr::pull("R")
  }
  if (!is.null(to)) {
    r <- match.arg(to, list_type_ref())
    ct <- TYPETABLE[[r]]
    check_types(x)
    toRet <- ct |> dplyr::slice(match(x, .data$R)) |> dplyr::pull("inst")
  }
  return(toRet)
}

###############################################################################@
#' Check if a set of types is supported
#'
#' @param x a character vector of types to be checked
#'
#' @export
#'
check_types <- function(x) {
  stopifnot(is.character(x))
  notSupported <- setdiff(x, SUPPTYPES)
  if (length(notSupported) > 0) {
    stop(paste(
      "The following types are not supported:",
      paste(notSupported, collapse = ", ")
    ))
  }
}

###############################################################################@
#' Convert an object into a specific type
#'
#' @param x an object to convert
#' @param type the targeted type
#'
#' @export
#'
as_type <- function(x, type) {
  type <- match.arg(type, SUPPTYPES)
  return(switch(
    type,
    "integer" = as.integer(x),
    "numeric" = as.numeric(x),
    "logical" = as.logical(x),
    "character" = as.character(x),
    "Date" = as.Date(x),
    "POSIXct" = as.POSIXct(x),
    "base64" = if (length(x) == 0) {
      character(0)
    } else {
      unlist(lapply(x, function(y) {
        if (length(y) == 0 || (length(y) == 1 && is.na(y))) {
          return(NA)
        } else {
          paste(y, collapse = "")
        }
      }))
    }
  ))
}

###############################################################################@
#' Create a RelDataModel object
#'
#' @param l the list of table models ([RelTableModel] objects)
#' @param checkFK a logical indicating if foreign keys should be checked
#' (default: TRUE)
#' @param createFKIndex should an index be create for the foreign keys
#' (default: FALSE)
#'
#' @return A RelDataModel object.
#'
#' @export
#'
RelDataModel <- function(l, checkFK = TRUE, createFKIndex = FALSE) {
  ## Checks ----
  stopifnot(
    all(unlist(lapply(l, is.RelTableModel)))
  )

  ## Creating the object ----
  toRet <- l
  names(toRet) <- NULL
  tn <- unlist(lapply(
    toRet,
    # function(x) paste(x$dbName, x$tableName, sep=".")
    function(x) x$tableName
  ))
  dtn <- unique(tn[which(duplicated(tn))])
  if (length(dtn) > 0) {
    stop(paste(
      "Duplicated tables:",
      paste(dtn, collapse = ", ")
    ))
  }
  if (any(tn %in% PCKRESERVED)) {
    stop(sprintf("%s are reserved words", paste(PCKRESERVED, collapse = ", ")))
  }
  names(toRet) <- tn

  ## Correct constraints ----
  for (ft in names(toRet)) {
    x <- toRet[[ft]]
    if (length(x$foreignKeys) > 0) {
      for (i in 1:length(x$foreignKeys)) {
        fk <- x$foreignKeys[[i]]
        if (
          fk$cardinality["fmin"] > 0 &&
            nrow(fk$key) == 1 &&
            fk$refTable %in% names(toRet)
        ) {
          toRet[[fk$refTable]]$fields[
            which(toRet[[fk$refTable]]$fields$name == fk$key$from),
            "nullable"
          ] <- FALSE
        }
        if (fk$cardinality["tmin"] > 0 && nrow(fk$key) == 1) {
          toRet[[ft]]$fields[
            which(toRet[[ft]]$fields$name == fk$key$from),
            "nullable"
          ] <- FALSE
        }

        if (fk$cardinality["fmax"] == 1) {
          if (nrow(fk$key) == 1) {
            if (
              toRet[[ft]]$fields[
                which(toRet[[ft]]$fields$name == fk$key$from),
                "type"
              ] %in%
                c("row", "column")
            ) {
              stop("Matrix row and column cannot be unique individually")
            }
            toRet[[ft]]$fields[
              which(toRet[[ft]]$fields$name == fk$key$from),
              "unique"
            ] <- TRUE
          }

          ei <- lapply(
            toRet[[ft]]$indexes,
            function(y) {
              identical(sort(y$fields), sort(fk$key$from))
            }
          ) |>
            unlist() |>
            as.logical() |>
            which()
          if (length(ei) > 1) {
            stop("Check this part of code")
          }
          if (length(ei) == 0) {
            if (createFKIndex) {
              toRet[[ft]]$indexes <- unique(c(
                toRet[[ft]]$indexes,
                list(
                  list(
                    fields = sort(fk$key$from),
                    unique = TRUE
                  )
                )
              ))
            }
          } else {
            toRet[[ft]]$indexes[[ei]]$unique <- TRUE
          }
        }

        if (fk$cardinality["tmax"] == 1 && fk$refTable %in% names(toRet)) {
          if (nrow(fk$key) == 1) {
            if (
              toRet[[fk$refTable]]$fields[
                which(toRet[[fk$refTable]]$fields$name == fk$key$to),
                "type"
              ] %in%
                c("row", "column")
            ) {
              stop("Matrix row and column cannot be unique individually")
            }
            toRet[[fk$refTable]]$fields[
              which(toRet[[fk$refTable]]$fields$name == fk$key$to),
              "unique"
            ] <- TRUE
          }

          ei <- lapply(
            toRet[[fk$refTable]]$indexes,
            function(y) {
              identical(sort(y$fields), sort(fk$key$to))
            }
          ) |>
            unlist() |>
            as.logical() |>
            which()
          if (length(ei) > 1) {
            stop("Check this part of code")
          }
          if (length(ei) == 0) {
            if (createFKIndex) {
              toRet[[fk$refTable]]$indexes <- unique(c(
                toRet[[fk$refTable]]$indexes,
                list(
                  list(
                    fields = sort(fk$key$to),
                    unique = TRUE
                  )
                )
              ))
            }
          } else {
            toRet[[fk$refTable]]$indexes[[ei]]$unique <- TRUE
          }
        }
      }
    }
  }
  toRet <- lapply(toRet, correct_constraints)

  ## Return the results ----
  class(toRet) <- c("RelDataModel", class(toRet))
  if (checkFK) {
    check_foreign_keys(toRet)
  }
  return(toRet)
}

###############################################################################@
#' Check if the object is  a [RelDataModel] object
#'
#' @param x any object
#'
#' @return A single logical: TRUE if x is a [RelDataModel] object
#'
#' @export
#'
is.RelDataModel <- function(x) {
  inherits(x, "RelDataModel")
}

###############################################################################@
#' Check the availability of foreign keys
#'
#' @param x a [RelDataModel] object
#'
#' @return Nothing. The function throws an error if there is an issue with
#' foreign keys.
#'
#' @export
#'
check_foreign_keys <- function(x) {
  stopifnot(is.RelDataModel(x))
  lapply(
    x,
    function(tm) {
      fks <- tm$foreignKeys
      if (is.null(fks)) {
        return()
      }
      lapply(
        fks,
        function(fk) {
          ft <- which(unlist(lapply(
            x,
            function(y) {
              y$tableName == fk$refTable
            }
          )))
          if (length(ft) != 1) {
            stop(sprintf(
              "Cannot find the foreign table: %s",
              fk$refTable
            ))
          }
          ft <- x[[ft]]
          for (i in 1:nrow(fk$key)) {
            from <- fk$key[i, ]$from
            to <- fk$key[i, ]$to
            if (!to %in% ft$fields$name) {
              stop(sprintf(
                "Cannot find the %s field in the %s table",
                to,
                fk$refTable
              ))
            }
            fromt <- tm$fields$type[which(tm$fields$name == from)]
            fromt <- ifelse(
              fromt %in% c("row", "column"),
              "character",
              fromt
            )
            tot <- ft$fields$type[which(ft$fields$name == to)]
            tot <- ifelse(
              tot %in% c("row", "column"),
              "character",
              tot
            )
            if (fromt != tot) {
              stop(sprintf(
                paste(
                  "The %s field in the %s table and",
                  "the %s field in the %s table",
                  "are not of the same type"
                ),
                from,
                tm$tableName,
                to,
                ft$tableName
              ))
            }
          }
        }
      )
    }
  )
  invisible(NULL)
}

###############################################################################@
#' @export
#'
length.RelDataModel <- function(x) {
  length(unclass(x))
}

###############################################################################@
#' Subset a [RelDataModel]
#'
#' @param x the [RelDataModel] objcet
#' @param i the index or the names of the elements to extract
#' @param rmForeignKeys if TRUE, remove foreign keys which are not
#' available after extraction. If FALSE (default) the function will throw an
#' error if any foreign keys does not exist in the extracted RelDataModel.
#' @param ... additional arguments for the \code{\link{RelDataModel}} function.
#'
#' @export
#'
'[.RelDataModel' <- function(x, i, rmForeignKeys = FALSE, ...) {
  l <- unclass(x)
  l <- l[i]
  if (rmForeignKeys) {
    l <- lapply(
      l,
      function(tm) {
        if (length(tm$foreignKeys) > 0) {
          tm$foreignKeys <- tm$foreignKeys[which(unlist(lapply(
            tm$foreignKeys,
            function(fk) {
              fk$refTable %in% names(l)
            }
          )))]
        }
        return(tm)
      }
    )
    return(RelDataModel(l = l, ...))
  } else {
    toRet <- try(RelDataModel(l = l, ...), silent = TRUE)
    if (inherits(toRet, "try-error")) {
      if (length(grep("Cannot find the foreign table", toRet)) > 0) {
        toRet <- paste0(
          toRet,
          "Set rmForeignKeys to TRUE to remove unavailable foreign keys\n"
        )
      }
      stop(toRet)
    } else {
      return(toRet)
    }
  }
}

###############################################################################@
#' @export
#'
'[<-.RelDataModel' <- function(x, i, value) {
  stop("'[<-' is not supported for RelDataModel: use 'c' instead")
}

###############################################################################@
#' @export
#'
'[[<-.RelDataModel' <- function(x, i, value) {
  stop("'[[<-' is not supported for RelDataModel: use 'c' instead")
}

###############################################################################@
#' @export
#'
names.RelDataModel <- function(x) {
  names(unclass(x))
}

###############################################################################@
#' @export
#'
'names<-.RelDataModel' <- function(x, value) {
  stopifnot(
    is.character(value),
    length(value) == length(x),
    sum(duplicated(value)) == 0
  )
  ovalues <- names(x)
  l <- unclass(x)
  for (i in 1:length(l)) {
    l[[i]]$tableName <- value[i]
    fk <- l[[i]]$foreignKeys
    if (length(fk) > 0) {
      for (j in 1:length(fk)) {
        nn <- value[which(
          ovalues == fk[[j]]$refTable
        )]
        fk[[j]]$refTable <- nn
      }
    }
    l[[i]]$foreignKeys <- fk
  }
  return(RelDataModel(l = l))
}


###############################################################################@
#' Merge [RelDataModel] objects
#'
#' @param ... [RelDataModel] objects
#' @param checkFK a logical indicating if foreign keys should be checked
#' (default: TRUE)
#'
#' @return A [RelDataModel] objects
#'
#' @export
#'
c.RelDataModel <- function(..., checkFK = TRUE) {
  inputs <- list(...)
  stopifnot(
    all(unlist(lapply(inputs, is.RelDataModel)))
  )
  il <- lapply(
    inputs,
    function(x) {
      return(unclass(x))
    }
  )
  l <- do.call(c, il)
  return(RelDataModel(
    l = l,
    checkFK = checkFK
  ))
}

###############################################################################@
#' Convert a [RelDataModel] object in a list of 5 normalized tibbles
#'
#' @param rdm a [RelDataModel] object
#'
#' @return A list with the following tibbles:
#'
#' - **tables**: The tables in the model with the following information
#'    + **name**: the name of the table
#'    + **x**: the x coordinate of the table in the model drawing
#'    (NA ==> position undefined)
#'    + **y**: the y coordinate of the table in the model drawing
#'    (NA ==> position undefined)
#'    + **color**: the color of the table in the model drawing
#'    (NA ==> undefined)
#'    + **comment**: comment about the table
#'
#' - **fields**: The fields in the model with the following information
#'    + **name**: the name of the field
#'    + **type**: the type of the field
#'    + **nullable**: a logical indicating if the field can be null
#'    + **comment**: comment about the field
#'    + **table**: the name of the table to which the field belongs
#'
#' - **primaryKeys**: The primary keys in the model with the
#' following information
#'    + **table**: the name of the relevant table
#'    + **field**: the name of the field participating to the primary key
#'
#' - **foreignKeys**: The foreign keys in the model with the
#' following information
#'    + **table**: the name of the referring table
#'    + **fki**: the identifier of the foreign key (by referring table)
#'    + **field**: the name of the referring field
#'    + **refTable**: the name of the referred table
#'    + **refField**: the name of the referred field
#'
#' - **indexes**: The indexes in the model with the following information
#'    + **table**: the name of the relevant table
#'    + **idx**: the identifier of the index (by table)
#'    + **field**: the name of the field participating to the index
#'    + **unique**: a logical indicating if the field is unique
#'
#' @export
toDBM <- function(rdm) {
  stopifnot(is.RelDataModel(rdm))
  toRet <- lapply(
    rdm,
    function(tm) {
      ##
      tables <- dplyr::tibble(
        name = tm$tableName,
        x = as.numeric(NA),
        y = as.numeric(NA),
        color = as.character(NA),
        comment = as.character(NA)
      )
      if (!is.null(tm$display)) {
        for (n in c("x", "y", "color", "comment")) {
          tables[, n] <- tm$display[[n]]
        }
      }
      ##
      fields <- tm$fields |>
        dplyr::mutate(
          table = tm$tableName,
          fieldOrder = 1:nrow(tm$fields)
        )
      ##
      if (length(tm$primaryKey) == 0) {
        primaryKeys <- dplyr::tibble(
          table = character(),
          field = character()
        )
      } else {
        primaryKeys <- dplyr::tibble(
          table = tm$tableName,
          field = tm$primaryKey
        )
      }
      ##
      if (length(tm$foreignKeys) == 0) {
        foreignKeys <- dplyr::tibble(
          table = character(),
          fki = integer(),
          field = character(),
          refTable = character(),
          refField = character()
        )
      } else {
        foreignKeys <- do.call(
          dplyr::bind_rows,
          lapply(
            1:length(tm$foreignKeys),
            function(fki) {
              fmin <- tm$foreignKeys[[fki]]$cardinality["fmin"]
              fmax <- tm$foreignKeys[[fki]]$cardinality["fmax"]
              tmin <- tm$foreignKeys[[fki]]$cardinality["tmin"]
              tmax <- tm$foreignKeys[[fki]]$cardinality["tmax"]
              toRet <- tm$foreignKeys[[fki]]$key |>
                dplyr::mutate(
                  table = tm$tableName,
                  refTable = tm$foreignKeys[[fki]]$refTable,
                  fki = fki,
                  fmin = fmin,
                  fmax = fmax,
                  tmin = tmin,
                  tmax = tmax
                )
              toRet |>
                dplyr::rename("field" = "from", "refField" = "to") |>
                dplyr::select(
                  "table",
                  "fki",
                  "field",
                  "refTable",
                  "refField",
                  "fmin",
                  "fmax",
                  "tmin",
                  "tmax"
                )
            }
          )
        )
      }
      ##
      if (length(tm$indexes) == 0) {
        indexes <- dplyr::tibble(
          table = character(),
          idx = integer(),
          field = character(),
          unique = logical()
        )
      } else {
        indexes <- do.call(
          dplyr::bind_rows,
          lapply(
            1:length(tm$indexes),
            function(idx) {
              if (length(idx) == 0) {
                return(NULL)
              }
              toRet <- dplyr::tibble(
                table = tm$tableName,
                idx = idx,
                unique = tm$indexes[[idx]]$unique,
                field = tm$indexes[[idx]]$fields
              ) |>
                dplyr::select("table", "idx", "field", "unique")
              return(toRet)
            }
          )
        )
      }
      ##
      return(list(
        tables = tables,
        fields = fields,
        primaryKeys = primaryKeys,
        foreignKeys = foreignKeys,
        indexes = indexes
      ))
    }
  )
  return(list(
    tables = do.call(dplyr::bind_rows, lapply(toRet, function(x) x$tables)),
    fields = do.call(dplyr::bind_rows, lapply(toRet, function(x) x$fields)),
    primaryKeys = do.call(
      dplyr::bind_rows,
      lapply(toRet, function(x) x$primaryKeys)
    ),
    foreignKeys = do.call(
      dplyr::bind_rows,
      lapply(toRet, function(x) x$foreignKeys)
    ),
    indexes = do.call(dplyr::bind_rows, lapply(toRet, function(x) x$indexes))
  ))
}

###############################################################################@
#' Convert a list of 5 normalized tibbles in a [RelDataModel] object
#'
#' @param dbm a list with the following tibbles:
#'
#' - **tables**: The tables in the model with the following information
#'    + **name**: the name of the table
#'    + **x**: the x coordinate of the table in the model drawing
#'    (NA ==> position undefined)
#'    + **y**: the y coordinate of the table in the model drawing
#'    (NA ==> position undefined)
#'    + **color**: the color of the table in the model drawing
#'    (NA ==> undefined)
#'    + **comment**: comment about the table
#'
#' - **fields**: The fields in the model with the following information
#'    + **name**: the name of the field
#'    + **type**: the type of the field
#'    + **nullable**: a logical indicating if the field can be null
#'    + **comment**: comment about the field
#'    + **table**: the name of the table to which the field belongs
#'
#' - **primaryKeys**: The primary keys in the model with the
#' following information
#'    + **table**: the name of the relevant table
#'    + **field**: the name of the field participating to the primary key
#'
#' - **foreignKeys**: The foreign keys in the model with the
#' following information
#'    + **table**: the name of the referring table
#'    + **fki**: the identifier of the foreign key (by referring table)
#'    + **field**: the name of the referring field
#'    + **refTable**: the name of the referred table
#'    + **refField**: the name of the referred field
#'
#' - **indexes**: The indexes in the model with the following information
#'    + **table**: the name of the relevant table
#'    + **idx**: the identifier of the index (by table)
#'    + **field**: the name of the field participating to the index
#'    + **unique**: a logical indicating if the field is unique
#'
#' @return A [RelDataModel] object
#'
#' @export
#'
fromDBM <- function(dbm) {
  tableName <- split(dbm$tables$name, dbm$tables$name)

  dbm$fields <- dplyr::arrange(dbm$fields, .data$fieldOrder)
  fields <- split(
    dplyr::select(dbm$fields, -"table", -"fieldOrder"),
    dbm$fields$table
  )

  primaryKey <- split(dbm$primaryKeys$field, dbm$primaryKeys$table)

  foreignKeys <- split(
    dplyr::select(dbm$foreignKeys, -"table"),
    dbm$foreignKeys$table
  ) |>
    lapply(function(tfk) {
      split(dplyr::select(tfk, -"fki"), tfk$fki) |>
        lapply(function(fk) {
          list(
            refTable = fk$refTable[1],
            key = dplyr::select(fk, "from" = "field", "to" = "refField"),
            cardinality = c(
              fk$fmin[1],
              fk$fmax[1],
              fk$tmin[1],
              fk$tmax[1]
            ) |>
              stats::setNames(c("fmin", "fmax", "tmin", "tmax"))
          )
        }) |>
        stats::setNames(NULL)
    })

  indexes <- split(dplyr::select(dbm$indexes, -"table"), dbm$indexes$table) |>
    lapply(function(ti) {
      split(dplyr::select(ti, -"idx"), ti$idx) |>
        lapply(function(idx) {
          return(list(fields = idx$field, unique = idx$unique[1]))
        }) |>
        stats::setNames(NULL)
    })

  display <- split(dplyr::select(dbm$tables, -"name"), dbm$tables$name) |>
    lapply(function(di) {
      di |>
        (function(x) {
          class(x) <- "list"
          x
        })() |>
        `attr<-`("row.names", NULL)
    })

  toRet <- lapply(
    tableName,
    function(tn) {
      return(RelTableModel(
        tableName = tn,
        fields = fields[[tn]],
        primaryKey = primaryKey[[tn]],
        foreignKeys = foreignKeys[[tn]],
        indexes = indexes[[tn]],
        display = display[[tn]]
      ))
    }
  ) |>
    RelDataModel()

  return(toRet)
}

###############################################################################@
#' Rename a table in a [RelDataModel]
#'
#' @param x a [RelDataModel] object
#' @param old a single character corresponding to the table name to change
#' @param new the new table name
#'
#' @return A [RelDataModel]
#'
#' @export
#'
rename_table <- function(x, old, new) {
  stopifnot(is.RelDataModel(x))
  stopifnot(
    is.RelDataModel(x),
    length(old) == 1,
    length(new) == 1,
    old %in% names(x)
  )
  newNames <- names(x)
  newNames[which(newNames == old)] <- new
  names(x) <- newNames
  return(x)
}

###############################################################################@
#' Add a table to a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param newTable the name of the new table or a [RelTableModel]
#'
#' @return A [RelDataModel]
#'
#' @export
#'
#'
add_table <- function(x, newTable) {
  stopifnot(is.RelDataModel(x))
  stopifnot(is.RelTableModel(newTable) || is.character(newTable))
  if (is.character(newTable)) {
    stopifnot(
      length(newTable) == 1
    )
    newTable <- RelTableModel(
      tableName = newTable,
      fields = dplyr::tibble(
        name = character(),
        type = character(),
        nullable = logical(),
        unique = logical(),
        comment = character()
      ),
      primaryKey = NULL,
      foreignKeys = NULL,
      indexes = NULL,
      display = list(
        x = as.numeric(NA),
        y = as.numeric(NA),
        color = as.character(NA),
        comment = as.character(NA)
      )
    )
  }
  stopifnot(!newTable$tableName %in% names(x))
  x <- unclass(x)
  x[[newTable$tableName]] <- newTable
  return(RelDataModel(x))
}

###############################################################################@
#' Remove a table from a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to remove
#' @param rmForeignKeys if TRUE, remove foreign keys which are not
#' available after extraction. If FALSE (default) the function will throw an
#' error if any foreign keys does not exist in the extracted RelDataModel.
#'
#' @return A [RelDataModel]
#'
#' @export
#'
remove_table <- function(x, tableName, rmForeignKeys = FALSE) {
  stopifnot(is.RelDataModel(x))
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    tableName %in% names(x)
  )
  return(x[-which(names(x) == tableName), rmForeignKeys = rmForeignKeys])
}

###############################################################################@
#' Add a field to a table in a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to modify (a single character)
#' @param name the name of the field to add (a single character)
#' @param type the type of the field (a single character)
#' @param nullable if the field is nullable (a single logical)
#' @param unique if the values are unique (a single logical)
#' @param comment a description (a single character)
#'
#' @return A [RelDataModel]
#'
#' @export
#'
add_field <- function(
  x,
  tableName,
  name,
  type,
  nullable,
  unique,
  comment
) {
  stopifnot(is.RelDataModel(x))
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    !is.MatrixModel(x[[tableName]]),
    !is.na(tableName),
    tableName %in% names(x),
    is.character(name),
    length(name) == 1,
    !is.na(name),
    !name %in% x[[tableName]]$fields$name,
    is.character(type),
    length(type) == 1,
    !is.na(type),
    type %in% c(SUPPTYPES, "row", "column"),
    is.logical(nullable),
    length(nullable) == 1,
    !is.na(nullable),
    is.logical(unique),
    length(unique) == 1,
    !is.na(unique),
    is.character(comment),
    length(comment) == 1
  )
  x <- unclass(x)
  if (type %in% c("row", "column")) {
    stopifnot(
      length(x[[tableName]]) == 0 #, !nullable, unique
    )
    x[[tableName]]$fields <- dplyr::tibble(
      name = c(
        name,
        paste0(
          toupper(setdiff(c("row", "column"), type)),
          "_TO_UPDATE"
        ),
        "VALUE_TO_UPDATE"
      ),
      type = c(type, setdiff(c("row", "column"), type), "character"),
      nullable = c(FALSE, FALSE, TRUE),
      unique = c(FALSE, FALSE, FALSE),
      comment = c(comment, NA, NA)
    )
    x[[tableName]]$primaryKey <- x[[tableName]]$fields$name[1:2]
  } else {
    x[[tableName]]$fields <- dplyr::bind_rows(
      x[[tableName]]$fields,
      dplyr::tibble(
        name = name,
        type = type,
        nullable = nullable,
        unique = unique,
        comment = comment
      )
    )
  }
  return(RelDataModel(x))
}

###############################################################################@
#' Rename an existing field in a [RelDataModel] table
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to modify (a single character)
#' @param current the current name of the field to modify (a single character)
#' @param new the new name of the field (a single character)
#'
#' @return A [RelDataModel]
#'
#' @export
#'
rename_field <- function(x, tableName, current, new) {
  stopifnot(is.RelDataModel(x))
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    tableName %in% names(x),
    is.character(current),
    length(current) == 1,
    current %in% x[[tableName]]$fields$name,
    current != new,
    is.character(new),
    length(new) == 1,
    !new %in% x[[tableName]]$fields$name
  )
  x <- unclass(x)
  tm <- x[[tableName]]
  ## Renaming the field ----
  tm$fields[
    which(tm$fields$name == current),
    "name"
  ] <- new
  ## Adapting primary key ----
  tm$primaryKey[which(tm$primaryKey == current)] <- new
  tm$primaryKey <- sort(tm$primaryKey)
  ## Adapting indexes ----
  if (length(tm$indexes) > 0) {
    for (i in 1:length(tm$indexes)) {
      tm$indexes[[i]]$fields[
        which(tm$indexes[[i]]$fields == current)
      ] <- new
      tm$indexes[[i]]$fields <- sort(tm$indexes[[i]]$fields)
    }
  }
  ## Adapting foreign keys ----
  if (length(tm$foreignKeys) > 0) {
    for (i in 1:length(tm$foreignKeys)) {
      tm$foreignKeys[[i]]$key[
        which(tm$foreignKeys[[i]]$key$from == current),
        "from"
      ] <- new
    }
  }
  x[[tableName]] <- tm
  ## Changing foreign Keys in other tables ----
  for (i in 1:length(x)) {
    tm <- x[[i]]
    keys <- tm$foreignKeys
    if (length(keys) > 0) {
      for (j in 1:length(keys)) {
        if (keys[[j]]$refTable == tableName) {
          keys[[j]]$key[which(keys[[j]]$key$to == current), "to"] <- new
        }
      }
    }
    tm$foreignKeys <- keys
    x[[i]] <- tm
  }
  return(RelDataModel(x))
}

###############################################################################@
fk_match <- function(
  x,
  fromTable,
  fromFields,
  toTable,
  toFields
) {
  tocheck <- paste(fromFields, toFields, sep = "-->") |> sort()
  tfk <- x[[fromTable]]$foreignKeys
  lapply(
    tfk,
    function(y) {
      y$refTable == toTable &&
        nrow(y$key) == length(tocheck) &&
        all(
          sort(paste(y$key$from, y$key$to, sep = "-->")) == tocheck
        )
    }
  ) |>
    unlist() |>
    as.logical() |>
    which()
}

###############################################################################@
#' Add a foreign key between two tables
#'
#' @param x a [RelDataModel]
#' @param fromTable the name of the referencing table
#' @param fromFields the name of the referencing fields
#' @param toTable the name of the referenced table
#' @param toFields the names of the referenced fields
#' @param fmin from minimum cardinality (default: 0L)
#' @param fmax from maximum cardinality (default: -1L ==> Infinite)
#' @param tmin to minimum cardinality (default: 1L)
#' @param tmax to maximum cardinality (default: 1L)
#'
#' @return A [RelDataModel]
#'
#' @export
#'
add_foreign_key <- function(
  x,
  fromTable,
  fromFields,
  toTable,
  toFields,
  fmin = 0L,
  fmax = -1L,
  tmin = 1L,
  tmax = 1L
) {
  stopifnot(is.RelDataModel(x))
  ft <- x[[fromTable]]$fields$type[
    match(fromFields, x[[fromTable]]$fields$name)
  ]
  ft <- ifelse(ft %in% c("row", "column"), "character", ft)
  tt <- x[[toTable]]$fields$type[
    match(toFields, x[[toTable]]$fields$name)
  ]
  tt <- ifelse(tt %in% c("row", "column"), "character", tt)
  stopifnot(
    is.character(fromTable),
    length(fromTable) == 1,
    fromTable %in% names(x),
    is.character(toTable),
    length(toTable) == 1,
    toTable %in% names(x),
    is.character(fromFields),
    length(fromFields) >= 1,
    is.character(toFields),
    length(toFields) == length(fromFields),
    all(fromFields %in% x[[fromTable]]$fields$name),
    all(toFields %in% x[[toTable]]$fields$name),
    all(ft == tt),
    all(ft != "base64"),
    all(!is.na(c(fmin, fmax, tmin, tmax))),
    is.integer(fmin),
    is.integer(fmax),
    is.integer(tmin),
    is.integer(tmax),
    fmin > -1,
    tmin > -1,
    fmax == -1 || (fmax > 0 && fmax >= fmin),
    tmax == -1 || (tmax > 0 && tmax >= tmin)
  )
  if (is.MatrixModel(x[[fromTable]])) {
    fi <- x[[fromTable]]$fields |> dplyr::filter(.data$name %in% fromFields)
    if (any(!fi$type %in% c("row", "column"))) {
      stop(paste(
        "Only row and column of matrices can be used in foreign keys"
      ))
    }
    if (fmax == 1 && nrow(fi) == 1) {
      stop("Matix row and column cannot be mapped individually unambiguisly")
    }
  }
  if (is.MatrixModel(x[[toTable]])) {
    fi <- x[[toTable]]$fields |> dplyr::filter(.data$name %in% toFields)
    if (any(!fi$type %in% c("row", "column"))) {
      stop(paste(
        "Only row and column of matrices can be used in foreign keys"
      ))
    }
    if (tmax == 1 && nrow(fi) == 1) {
      stop("Matix row and column cannot be mapped individually unambiguisly")
    }
  }
  efk <- fk_match(x, fromTable, fromFields, toTable, toFields)
  if (length(efk) != 0) {
    warning("The foreign key already exists ==> no change")
    return(x)
  }
  x <- unclass(x)
  x[[fromTable]]$foreignKeys <- c(
    x[[fromTable]]$foreignKeys,
    list(list(
      refTable = toTable,
      key = dplyr::tibble(
        from = fromFields,
        to = toFields
      ),
      cardinality = c(
        "fmin" = as.integer(fmin),
        "fmax" = as.integer(fmax),
        "tmin" = as.integer(tmin),
        "tmax" = as.integer(tmax)
      )
    ))
  )
  return(RelDataModel(x))
}

###############################################################################@
#' Remove a foreign key between two tables
#'
#' @param x a [RelDataModel]
#' @param fromTable the name of the referencing table
#' @param fromFields the name of the referencing fields
#' @param toTable the name of the referenced table
#' @param toFields the names of the referenced fields
#'
#' @return A [RelDataModel]
#'
#' @export
#'
remove_foreign_key <- function(
  x,
  fromTable,
  fromFields,
  toTable,
  toFields
) {
  stopifnot(is.RelDataModel(x))
  stopifnot(
    is.character(fromTable),
    length(fromTable) == 1,
    fromTable %in% names(x),
    is.character(toTable),
    length(toTable) == 1,
    toTable %in% names(x),
    is.character(fromFields),
    length(fromFields) >= 1,
    is.character(toFields),
    length(toFields) == length(fromFields),
    all(fromFields %in% x[[fromTable]]$fields$name),
    all(toFields %in% x[[toTable]]$fields$name)
  )
  efk <- fk_match(x, fromTable, fromFields, toTable, toFields)
  if (length(efk) == 0) {
    warning("The foreign key does not exist ==> no change")
    return(x)
  }
  x <- unclass(x)
  x[[fromTable]]$foreignKeys <- x[[fromTable]]$foreignKeys[-efk]
  return(RelDataModel(x))
}

###############################################################################@
#' Update a the cardinalities of a foreign key between two tables
#'
#' @param x a [RelDataModel]
#' @param fromTable the name of the referencing table
#' @param fromFields the name of the referencing fields
#' @param toTable the name of the referenced table
#' @param toFields the names of the referenced fields
#' @param fmin from minimum cardinality
#' @param fmax from maximum cardinality
#' @param tmin to minimum cardinality
#' @param tmax to maximum cardinality
#'
#' @return A [RelDataModel]
#'
#' @export
#'
update_foreign_key <- function(
  x,
  fromTable,
  fromFields,
  toTable,
  toFields,
  fmin,
  fmax,
  tmin,
  tmax
) {
  stopifnot(is.RelDataModel(x))
  toRet <- x |>
    remove_foreign_key(
      fromTable,
      fromFields,
      toTable,
      toFields
    ) |>
    add_foreign_key(
      fromTable,
      fromFields,
      toTable,
      toFields,
      fmin,
      fmax,
      tmin,
      tmax
    )
  return(toRet)
}

###############################################################################@
#' Remove a field from a table in a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to modify (a single character)
#' @param fieldName the name of the field to remove (a single character)
#' @param rmForeignKeys a single logical indicating if the corresponding foreign
#' keys should be removed. If FALSE (default), the function will throw an
#' error if it encounter a foreign key using the field.
#'
#' @return A [RelDataModel]
#'
#' @export
#'
remove_field <- function(
  x,
  tableName,
  fieldName,
  rmForeignKeys = FALSE
) {
  stopifnot(is.RelDataModel(x))
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    !is.MatrixModel(x[[tableName]]),
    tableName %in% names(x),
    is.character(fieldName),
    length(fieldName) == 1,
    fieldName %in% x[[tableName]]$fields$name
  )
  ## Foreign keys ----
  refk <- lapply(
    x[[tableName]]$foreignKeys,
    function(y) {
      fieldName %in% y$key$from
    }
  ) |>
    unlist() |>
    as.logical() |>
    which()
  tefk <- lapply(
    x,
    function(y) {
      lapply(
        y$foreignKeys,
        function(z) {
          tableName == z$refTable &
            fieldName %in% z$key$to
        }
      ) |>
        unlist() |>
        as.logical() |>
        which()
    }
  )
  x <- unclass(x)
  if (length(refk) > 0 || any(unlist(lapply(tefk, length)) > 0)) {
    if (!rmForeignKeys) {
      stop(
        "This field is used as a foreign key.\n",
        "Set rmForeignKeys to TRUE to remove them"
      )
    } else {
      if (length(refk) > 0) {
        x[[tableName]]$foreignKeys <- x[[tableName]]$foreignKeys[-refk]
      }
      for (tn in names(tefk)) {
        if (length(tefk[[tn]] > 0)) {
          x[[tn]]$foreignKeys <- x[[tn]]$foreignKeys[-tefk[[tn]]]
        }
      }
    }
  }
  ## Adapting primary key ----
  if (fieldName %in% x[[tableName]]$primaryKey) {
    x[[tableName]]$primaryKey <- character()
  }
  ## Adapting indexes ----
  if (length(x[[tableName]]$indexes) > 0) {
    toRm <- c()
    for (i in 1:length(x[[tableName]]$indexes)) {
      if (fieldName %in% x[[tableName]]$indexes[[i]]$fields) {
        toRm <- c(toRm, i)
      }
    }
    if (length(toRm) > 0) {
      x[[tableName]]$indexes <- x[[tableName]]$indexes[-toRm]
    }
  }
  ## Removing the field ----
  x[[tableName]]$fields <- x[[tableName]]$fields |>
    dplyr::filter(.data$name != fieldName)
  ## Returning the results ----
  return(RelDataModel(x))
}

###############################################################################@
#' Order fields in a table in a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to modify (a single character)
#' @param order a vector of integers all in (1:number_of_fields)
#'
#' @return A [RelDataModel]
#'
#' @export
#'
order_fields <- function(
  x,
  tableName,
  order
) {
  stopifnot(is.RelDataModel(x))
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    tableName %in% names(x),
    is.numeric(order),
    length(order) == nrow(x[[tableName]]$fields),
    all(order %in% 1:nrow(x[[tableName]]$fields))
  )
  x <- unclass(x)
  x[[tableName]]$fields <- x[[tableName]]$fields[order, ]
  return(RelDataModel(x))
}

###############################################################################@
#' Set the primary key a table in a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to modify (a single character)
#' @param fieldNames the names of the fields to include in the primary key
#'
#' @return A [RelDataModel]
#'
#' @export
#'
set_primary_key <- function(x, tableName, fieldNames) {
  stopifnot(is.RelDataModel(x))
  fieldNames <- as.character(fieldNames)
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    !is.MatrixModel(x[[tableName]]),
    tableName %in% names(x),
    all(fieldNames %in% x[[tableName]]$fields$name)
  )
  x <- unclass(x)
  x[[tableName]]$primaryKey <- sort(fieldNames)
  return(RelDataModel(x))
}


###############################################################################@
#' Add an index to a table in a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to modify (a single character)
#' @param fieldNames the names of the fields to include in the index
#' @param unique a logical indicating if the indexed values are unique
#'
#' @return A [RelDataModel]
#'
#' @export
#'
add_index <- function(x, tableName, fieldNames, unique) {
  stopifnot(is.RelDataModel(x))
  fieldNames <- as.character(fieldNames) |> unique()
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    tableName %in% names(x),
    all(fieldNames %in% x[[tableName]]$fields$name),
    is.logical(unique),
    length(unique) == 1
  )
  ei <- lapply(
    x[[tableName]]$indexes,
    function(y) {
      identical(sort(y$fields), sort(fieldNames))
    }
  ) |>
    unlist() |>
    as.logical() |>
    any()
  if (ei) {
    warning("An index with the same fields already exists ==> no change")
    return(x)
  }
  x <- unclass(x)
  x[[tableName]]$indexes <- c(
    x[[tableName]]$indexes,
    list(list(fields = fieldNames, unique = unique))
  )
  return(RelDataModel(x))
}

###############################################################################@
#' Remove an index from a table in a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to modify (a single character)
#' @param fieldNames the names of the fields composing the index
#'
#' @return A [RelDataModel]
#'
#' @export
#'
remove_index <- function(x, tableName, fieldNames) {
  stopifnot(is.RelDataModel(x))
  fieldNames <- as.character(fieldNames)
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    tableName %in% names(x),
    all(fieldNames %in% x[[tableName]]$fields$name)
  )
  ei <- lapply(
    x[[tableName]]$indexes,
    function(y) {
      identical(sort(y$fields), sort(fieldNames))
    }
  ) |>
    unlist() |>
    as.logical() |>
    which()
  if (length(ei) == 0) {
    warning("The index does not exists ==> no change")
    return(x)
  }
  x <- unclass(x)
  x[[tableName]]$indexes <- x[[tableName]]$indexes[-ei]
  return(RelDataModel(x))
}

##############################################################################@
#' Order index in a table in a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to modify (a single character)
#' @param order a vector of integers all in (1:number_of_indexes)
#'
#' @return A [RelDataModel]
#'
#' @export
#'
order_indexes <- function(
  x,
  tableName,
  order
) {
  stopifnot(is.RelDataModel(x))
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    tableName %in% names(x),
    is.numeric(order),
    length(order) == length(x[[tableName]]$indexes),
    all(order %in% 1:length(x[[tableName]]$indexes))
  )
  x <- unclass(x)
  x[[tableName]]$indexes <- x[[tableName]]$indexes[order]
  return(RelDataModel(x))
}

###############################################################################@
#' Set table index uniqueness in a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to modify (a single character)
#' @param fieldNames the names of the fields composing the index
#' @param unique a logical value
#'
#' @return A [RelDataModel]
#'
#' @export
#'
set_unique_index <- function(x, tableName, fieldNames, unique) {
  stopifnot(is.RelDataModel(x))
  fieldNames <- as.character(fieldNames)
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    tableName %in% names(x),
    all(fieldNames %in% x[[tableName]]$fields$name),
    is.logical(unique),
    length(unique) == 1,
    !is.na(unique)
  )
  ei <- lapply(
    x[[tableName]]$indexes,
    function(y) {
      identical(sort(y$fields), sort(fieldNames))
    }
  ) |>
    unlist() |>
    as.logical() |>
    which()
  if (length(ei) == 0) {
    warning("The index does not exists ==> no change")
    return(x)
  }
  x <- unclass(x)
  x[[tableName]]$indexes[[ei]]$unique <- unique
  return(RelDataModel(x))
}

###############################################################################@
#' Update field information in a table of a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to modify (a single character)
#' @param fieldName the name of the field to modify (a single character)
#' @param type the type of the field (a single character)
#' @param nullable if the field is nullable (a single logical)
#' @param unique if the values are unique (a single logical)
#' @param comment a description (a single character)
#'
#' @return A [RelDataModel]
#'
#' @export
#'
update_field <- function(
  x,
  tableName,
  fieldName,
  type = NULL,
  nullable = NULL,
  unique = NULL,
  comment = NULL
) {
  stopifnot(is.RelDataModel(x))
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    tableName %in% names(x),
    is.character(fieldName),
    length(fieldName) == 1,
    fieldName %in% x[[tableName]]$fields$name
  )
  curType <- x[[tableName]]$fields |>
    dplyr::filter(.data$name == fieldName) |>
    dplyr::pull("type")
  type <- ifelse(
    is.null(type),
    curType,
    type
  )
  nullable <- ifelse(
    is.null(nullable),
    x[[tableName]]$fields |>
      dplyr::filter(.data$name == fieldName) |>
      dplyr::pull("nullable"),
    nullable
  )
  unique <- ifelse(
    is.null(unique),
    x[[tableName]]$fields |>
      dplyr::filter(.data$name == fieldName) |>
      dplyr::pull("unique"),
    unique
  )
  comment <- ifelse(
    is.null(comment),
    x[[tableName]]$fields |>
      dplyr::filter(.data$name == fieldName) |>
      dplyr::pull("comment"),
    comment
  )
  type <- as.character(type)
  nullable <- as.logical(nullable)
  unique <- as.logical(unique)
  comment <- as.character(comment)
  stopifnot(
    is.character(type),
    length(type) == 1,
    !is.na(type),
    is.logical(nullable),
    length(nullable) == 1,
    !is.na(nullable),
    is.logical(unique),
    length(unique) == 1,
    !is.na(unique),
    is.character(comment),
    length(comment) == 1
  )

  x <- unclass(x)
  if (is.MatrixModel(x[[tableName]]) && curType %in% c("row", "column")) {
    ## row/column type ----
    if (!type %in% c("row", "column")) {
      stop("row and column can only be mutated in row and column")
    }
    if (nullable || unique) {
      stop("row and column cannot be nullable nor unique")
    }
    ## Updating field information ----
    d2CurType <- setdiff(c("row", "column"), curType)
    d2Type <- setdiff(c("row", "column"), type)
    d2fn <- x[[tableName]]$fields |>
      dplyr::filter(.data$type == !!d2CurType) |>
      dplyr::pull("name")
    x[[tableName]]$fields <- x[[tableName]]$fields |>
      dplyr::mutate(
        type = ifelse(
          .data$name == fieldName,
          !!type,
          ifelse(.data$name == d2fn, !!d2Type, .data$type)
        ),
        comment = ifelse(.data$name == fieldName, !!comment, .data$comment)
      )
  } else {
    ## Other type ----
    if (type %in% c("row", "column")) {
      stop("Cannot set this field as row or column")
    }
    if (is.MatrixModel(x[[tableName]]) && type == "base64") {
      stop("A matrix cannot store base64 documents")
    }
    check_types(type)
    ## Foreign keys ----
    if (type != curType) {
      refk <- lapply(
        x[[tableName]]$foreignKeys,
        function(y) {
          fieldName %in% y$key$from
        }
      ) |>
        unlist() |>
        as.logical() |>
        which()
      tefk <- lapply(
        x,
        function(y) {
          lapply(
            y$foreignKeys,
            function(z) {
              tableName == z$refTable &
                fieldName %in% z$key$to
            }
          ) |>
            unlist() |>
            as.logical() |>
            which()
        }
      )
      if (length(refk) > 0 || any(unlist(lapply(tefk, length)) > 0)) {
        stop(
          "Cannot change the type of a field involved in a foreign key.\n",
          "You should first remove the foreign key involving it."
        )
      }
    }
    ## Updating field information ----
    x[[tableName]]$fields <- x[[tableName]]$fields |>
      dplyr::mutate(
        type = ifelse(.data$name == fieldName, !!type, .data$type),
        nullable = ifelse(.data$name == fieldName, !!nullable, .data$nullable),
        unique = ifelse(.data$name == fieldName, !!unique, .data$unique),
        comment = ifelse(.data$name == fieldName, !!comment, .data$comment)
      )
  }

  return(RelDataModel(x))
}


###############################################################################@
#' Update the display of a table of a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param tableName the name of the table to modify (a single character)
#' @param px the position of the table: x value
#' @param py the position of the table: y value
#' @param color the color of the table
#' @param comment a table description/comment
#'
#' @return A [RelDataModel]
#'
#' @export
#'
update_table_display <- function(
  x,
  tableName,
  px = NULL,
  py = NULL,
  color = NULL,
  comment = NULL
) {
  stopifnot(is.RelDataModel(x))
  px <- ifelse(
    is.null(px),
    x[[tableName]]$display$x,
    px
  )
  py <- ifelse(
    is.null(py),
    x[[tableName]]$display$y,
    py
  )
  color <- ifelse(
    is.null(color),
    x[[tableName]]$display$color,
    color
  )
  comment <- ifelse(
    is.null(comment),
    x[[tableName]]$display$comment,
    comment
  )
  px <- as.numeric(px)
  py <- as.numeric(py)
  color <- as.character(color)
  comment <- as.character(comment)
  stopifnot(
    is.character(tableName),
    length(tableName) == 1,
    tableName %in% names(x),
    is.numeric(px),
    length(px) == 1,
    is.numeric(py),
    length(py) == 1,
    is.character(color),
    length(color) == 1,
    is.character(comment),
    length(comment) == 1
  )
  ## Updating table display ----
  x <- unclass(x)
  x[[tableName]]$display$x <- px
  x[[tableName]]$display$y <- py
  x[[tableName]]$display$color <- color
  x[[tableName]]$display$comment <- comment
  return(RelDataModel(x))
}

###############################################################################@
#' Copy fields from one table to another in a [RelDataModel]
#'
#' @param x a [RelDataModel]
#' @param from the name of the table from which the fields are taken
#' @param to the name of the table to which the fields are copied
#' @param fields the names of the fields to copy
#'
#' @return A [RelDataModel]
#'
#' @export
#'
copy_fields <- function(x, from, to, fields) {
  stopifnot(
    is.RelDataModel(x),
    is.character(from) &&
      length(from) == 1 &&
      !is.na(from) &&
      from %in% names(x),
    is.character(to) && length(to) == 1 && !is.na(to) && to %in% names(x),
    is.character(fields) &&
      length(fields) > 0 &&
      !any(is.na(fields)) &&
      all(fields %in% x[[from]]$fields$name) &&
      !any(fields %in% x[[to]]$fields$name)
  )
  for (f in fields) {
    ff <- dplyr::filter(x[[from]]$fields, .data$name == f)
    x <- add_field(
      x,
      tableName = to,
      name = f,
      type = ff$type,
      nullable = ff$nullable,
      unique = ff$unique,
      comment = ff$comment
    )
  }
  return(x)
}

###############################################################################@
#' Pre-compute [RelDataModel] layout when missing any x or y table position
#'
#' @param x a [RelDataModel]
#' @param layout character name of igraph layout function to use
#' (Default: "layout_nicely").
#' @param lengthMultiplier a numeric value to scale x and y coordinate
#' (default: 40*length(x))
#' @param force if TRUE autolayout even if all tables have coordinates
#' (default: FALSE)
#'
#' @return A [RelDataModel]
#'
#' @export
#'
auto_layout <- function(
  x,
  layout = "layout_nicely",
  lengthMultiplier = 40 * length(x),
  force = FALSE
) {
  stopifnot(is.RelDataModel(x))
  if (length(x) == 0) {
    return(x)
  }
  mn <- modelToVn(x)
  if (force) {
    mn$nodes$x <- NA
    mn$nodes$y <- NA
  }
  if (any(is.na(mn$nodes$x)) || any(is.na(mn$nodes$y))) {
    vp <- mn$nodes |>
      dplyr::select(-"x", -"y") |>
      visNetwork::visNetwork(mn$edges) |>
      visNetwork::visIgraphLayout(layout = layout, randomSeed = 2)
    x <- lapply(
      x,
      function(n) {
        i <- which(vp$x$nodes$id == n$tableName)
        n$display$x <- vp$x$nodes$x[i] * lengthMultiplier
        n$display$y <- vp$x$nodes$y[i] * lengthMultiplier
        return(n)
      }
    )
    class(x) <- c("RelDataModel", "list")
  }
  return(x)
}

###############################################################################@
#' Confront a [RelDataModel] to actual data
#'
#' @param x a [RelDataModel]
#' @param data a list of data frames to be confronted with the model.
#' @param paths a character vector with file paths taken into account if
#' the data is empty.
#' The file [basename] without extension
#' will be considered as the table name.
#' @param returnData a logical indicating if the data should be returned
#' with the report (default: FALSE).
#' @param verbose a single logical value indicating if some process information
#' should be displayed (default: TRUE)
#' @param n_max maximum number of records to read (default: Inf).
#' @param checks a character vector with the name of optional checks to be done
#' (Default: if n_max==Inf ==> all of
#' them c("unique", "not nullable", "foreign keys"),
#' else ==> none)
#' @param delim single character used to separate fields within a record
#' (default: "\\t")
#' @param ... supplementary parameters for the [readr::read_delim] function.
#'
#' @return A report as a list
#'
#' @example inst/examples/ex_read_json.R
#'
#' @export
#'
confront_data <- function(
  x,
  data = list(),
  paths = NULL,
  returnData = FALSE,
  verbose = TRUE,
  n_max = Inf,
  checks = if (n_max == Inf) {
    c("unique", "not nullable", "foreign keys")
  } else {
    as.character()
  },
  delim = "\t",
  ...
) {
  stopifnot(is.RelDataModel(x))

  read_td <- function(tm, tn) {
    if (is.MatrixModel(tm)) {
      ismm <- is_MM(paths[tn])
      if (ismm) {
        td <- read_named_MM(
          paths[tn],
          n_max = n_max
        )
      } else {
        r1 <- attr(ismm, "r1")
        cn <- r1 |>
          strsplit(split = delim) |>
          unlist()
        cn <- gsub("['`]", "", cn)
        cn <- gsub('["]', "", cn)

        vt <- tm$fields |>
          dplyr::filter(!.data$type %in% c("row", "column")) |>
          dplyr::pull("type")
        ctypes <- do.call(
          readr::cols,
          structure(
            list(
              readr::col_character(),
              .default = switch(
                vt,
                "integer" = readr::col_integer(),
                "numeric" = readr::col_double(),
                "logical" = readr::col_logical(),
                "character" = readr::col_character(),
                "Date" = readr::col_date(),
                "POSIXct" = readr::col_datetime()
              )
            ),
            .Names = c("___ROWNAMES___", ".default")
          )
        )
        td <- readr::read_delim(
          paths[tn],
          delim = delim,
          n_max = n_max,
          skip = 1,
          col_types = ctypes,
          col_names = c("___ROWNAMES___", cn[-1]),
          ...
        ) |>
          as.data.frame(stringsAsFactors = FALSE)
        stopifnot(
          !any(duplicated(colnames(td))),
          !any(duplicated(td[[1]]))
        )
        rownames(td) <- td[[1]]
        td <- as.matrix(td[, -1, drop = FALSE])
      }
    } else {
      td <- readr::read_delim(
        paths[tn],
        delim = delim,
        n_max = n_max,
        col_types = col_types(tm),
        ...
      )
    }
    return(td)
  }

  ## Optional checks ----
  if (length(checks) > 0) {
    checks <- match.arg(
      checks,
      c("unique", "not nullable", "foreign keys"),
      several.ok = TRUE
    )
  }

  ## Empty model and data
  if (length(x) == 0 && length(data) == 0 && length(paths) == 0) {
    toRet <- list(
      model = x,
      checks = checks,
      n_max = n_max,
      missingTables = NULL,
      suppTables = NULL,
      availableTables = NULL,
      constraints = list(),
      success = TRUE
    )
    if (verbose) {
      cat(format_confrontation_report(toRet))
    }
    return(invisible(toRet))
  }

  ## Data files ----
  if (length(data) == 0) {
    stopifnot(is.character(paths), length(paths) > 0)
    nf <- paths[which(!file.exists(paths))]
    if (length(nf) > 0) {
      stop(
        "Cannot find the following files:\n   - ",
        paste(nf, collapse = "\n   - ")
      )
    }
    names(paths) <- sub(
      pattern = "(\\.[[:alnum:]]+)(\\.gz)?$",
      replacement = "",
      x = basename(paths)
    )
    dtn <- names(paths)
  } else {
    dtn <- names(data)
    stopifnot(length(dtn) > 0)
  }

  ## Available tables ----
  missingTables <- setdiff(names(x), dtn)
  suppTables <- setdiff(dtn, names(x))
  availableTables <- intersect(names(x), dtn)
  toRet <- list(
    model = x,
    checks = checks,
    n_max = n_max,
    missingTables = missingTables,
    suppTables = suppTables,
    availableTables = availableTables,
    constraints = list(),
    success = length(missingTables) == 0 && length(suppTables) == 0
  )

  ## Table checks ----
  if (length(availableTables) > 0) {
    for (ti in 1:length(availableTables)) {
      tn <- availableTables[ti]
      if (verbose) {
        message(sprintf(
          'Processing "%s" (table %s / %s) ',
          tn,
          ti,
          length(availableTables)
        ))
      }
      tm <- x[[tn]]
      if (tn %in% names(data)) {
        td <- data[[tn]]
      } else {
        td <- read_td(tm, tn)
        if (returnData) {
          data[[tn]] <- td
        }
      }
      ### Simple table ----
      tr <- confront_table_data(x = tm, d = td, checks = checks)
      toRet$success <- toRet$success && tr$success
      ### Foreign keys ----
      tfk <- tm$foreignKeys
      if ("foreign keys" %in% checks && length(tfk) > 0) {
        fkr <- list()
        for (i in 1:length(tfk)) {
          tfki <- tfk[[i]]
          rtn <- tfki$refTable
          if (rtn %in% missingTables) {
            fkr[[i]] <- list(
              "success" = FALSE,
              "message" = "Missing reference table."
            )
            tr$success <- FALSE
            toRet$success <- FALSE
            next()
          }
          ##
          if (tfki$cardinality["fmin"] == 0 && tfki$cardinality["tmin"] == 0) {
            fkr[[i]] <- list("success" = TRUE)
            next()
          }
          ##
          if (rtn == tn) {
            rtd <- td
            rtm <- tm
          } else {
            rtm <- x[[rtn]]
            if (rtn %in% names(data)) {
              rtd <- data[[rtn]]
            } else {
              rtd <- read_td(rtm, rtn)
              if (returnData) {
                data[[rtn]] <- rtd
              }
            }
          }
          ##
          success <- TRUE
          message <- NULL
          if (tfki$cardinality["fmin"] > 0) {
            if (is.MatrixModel(tm) || is.MatrixModel(rtm)) {
              for (j in 1:length(tfki$key$from)) {
                ff <- tfki$key$from[j]
                if (is.MatrixModel(tm)) {
                  fft <- tm$fields$type[which(tm$fields$name == ff)]
                  if (fft == "row") {
                    ffv <- rownames(td)
                  } else {
                    ffv <- colnames(td)
                  }
                } else {
                  ffv <- unique(td[[ff]])
                }
                tf <- tfki$key$to[j]
                if (is.MatrixModel(rtm)) {
                  tft <- rtm$fields$type[which(rtm$fields$name == tf)]
                  if (tft == "row") {
                    tfv <- rownames(rtd)
                  } else {
                    tfv <- colnames(rtd)
                  }
                } else {
                  tfv <- unique(rtd[[tf]])
                }
                fkissue <- !all(tfv %in% ffv)
                if (fkissue) {
                  break()
                }
              }
            } else {
              . <- NULL
              mt <- dplyr::anti_join(
                dplyr::select(rtd, dplyr::all_of(tfki$key$to)) |>
                  dplyr::filter_all(dplyr::any_vars(!is.na(.))),
                dplyr::select(td, dplyr::all_of(tfki$key$from)),
                by = dplyr::all_of(stats::setNames(
                  tfki$key$from,
                  tfki$key$to
                ))
              )
              fkissue <- nrow(mt) > 0
            }
            if (fkissue > 0) {
              success <- FALSE
              message <- paste(
                c(
                  message,
                  sprintf(
                    "All keys of %s should be available in %s.",
                    rtn,
                    tn
                  )
                ),
                collapse = " "
              )
            }
          }
          if (tfki$cardinality["tmin"] > 0) {
            if (is.MatrixModel(tm) || is.MatrixModel(rtm)) {
              for (j in 1:length(tfki$key$from)) {
                ff <- tfki$key$from[j]
                if (is.MatrixModel(tm)) {
                  fft <- tm$fields$type[which(tm$fields$name == ff)]
                  if (fft == "row") {
                    ffv <- rownames(td)
                  } else {
                    ffv <- colnames(td)
                  }
                } else {
                  ffv <- unique(td[[ff]])
                }
                tf <- tfki$key$to[j]
                if (is.MatrixModel(rtm)) {
                  tft <- rtm$fields$type[which(rtm$fields$name == tf)]
                  if (tft == "row") {
                    tfv <- rownames(rtd)
                  } else {
                    tfv <- colnames(rtd)
                  }
                } else {
                  tfv <- unique(rtd[[tf]])
                }
                fkissue <- !all(ffv %in% tfv)
                if (fkissue) {
                  break()
                }
              }
            } else {
              . <- NULL
              mt <- dplyr::anti_join(
                dplyr::select(td, dplyr::all_of(tfki$key$from)) |>
                  dplyr::filter_all(dplyr::any_vars(!is.na(.))),
                dplyr::select(rtd, dplyr::all_of(tfki$key$to)),
                by = dplyr::all_of(stats::setNames(
                  tfki$key$to,
                  tfki$key$from
                ))
              )
              fkissue <- nrow(mt) > 0
            }
            if (fkissue) {
              success <- FALSE
              message <- paste(
                c(
                  message,
                  sprintf(
                    "All keys of %s should be available in %s.",
                    tn,
                    rtn
                  )
                ),
                collapse = " "
              )
            }
          }
          fkr[[i]] <- list(success = success, message = message)
          tr$success <- tr$success && success
          toRet$success <- toRet$success && success
        }
        tr$foreignKey <- fkr
      }
      toRet$constraints[[tn]] <- tr
    }
  }
  if (returnData) {
    toRet$data <- data
  }

  if (verbose) {
    cat(format_confrontation_report(toRet))
  }

  ## Return the results
  invisible(toRet)
}

###############################################################################@
#' Check if two [RelDataModel] are identical
#'
#' @param x a [RelDataModel]
#' @param y a [RelDataModel]
#' @param ... additional parameters for [identical_RelTableModel()]
#'
#' @return A logical: TRUE if the 2 models are identical
#'
#' @export
#'
identical_RelDataModel <- function(x, y, ...) {
  stopifnot(is.RelDataModel(x), is.RelDataModel(y))
  toRet <- length(x) == length(y) && all(sort(names(x)) == sort(names(y)))
  if (toRet && length(x) > 0) {
    for (i in 1:length(x)) {
      itoRet <- identical_RelTableModel(
        x[[names(x)[i]]],
        y[[names(x)[i]]],
        ...
      )
      if (!itoRet) {
        message(sprintf("Tables %s are different", names(x)[i]))
      }
      toRet <- toRet && itoRet
    }
  } else {
    if (length(x) > 0) {
      message("Not the same tables")
    }
  }
  return(toRet)
}


###############################################################################@
#' Get foreign keys in [RelDataModel]
#'
#' @param x a [RelDataModel]
#'
#' @return A tibble with the following fields:
#' - from: the origin of the key
#' - ff: the key fields in from
#' - to: the target of the key
#' - tf: the key fields in to
#' - fmin: minimum cardinality of from
#' - fmax: maximum cardinality of from
#' - tmin: minimum cardinality of to
#' - tmax: maximum cardinality of to
#'
#' @export
#'
get_foreign_keys.RelDataModel <- function(x) {
  do.call(rbind, lapply(x, get_foreign_keys))
}

#' Read a data model from an SQL file from the MySQL Workbench
#'
#' @param f the SQL file to read
#' @param typeRef the reference for type conversion
#' (Default: "MySQLWB"; see [list_type_ref()])
#' @param mysqlcomments if MySQL comments (starting with #) should be removed
#' (Default: TRUE)
#'
#' @details Database, table and field names should be surrounded by "`".
#'
#' @return A [RelDataModel] object
#'
#' @example inst/examples/ex_read_sql.R
#'
#' @export
#'
read_SQL_data_model <- function(f, typeRef = "MySQLWB", mysqlcomments = TRUE) {
  ############################################################################@
  ## Helpers ----

  ###################################################@
  getBrackPos <- function(x) {
    oBrackPos <- setdiff(gregexpr('[(]', x)[[1]], -1)
    if (length(oBrackPos) == 0) {
      return(data.frame(s = 1, e = 1)[-1, ])
    }
    cBrackPos <- setdiff(gregexpr('[)]', x)[[1]], -1)
    if (length(cBrackPos) != length(oBrackPos)) {
      stop("Closing brackets not matching opening brackets")
    }
    posBrack <- dplyr::tibble(
      pos = c(oBrackPos, cBrackPos),
      brack = c(rep("(", length(oBrackPos)), rep(")", length(cBrackPos)))
    ) |>
      dplyr::arrange(.data$pos) |>
      dplyr::mutate(oc = ifelse(.data$brack == "(", 1, -1)) |>
      dplyr::mutate(ib = cumsum(.data$oc))
    e <- which(posBrack$ib == 0)
    s <- c(1, e[-length(e)] + 1)
    return(dplyr::tibble(
      s = posBrack$pos[s],
      e = posBrack$pos[e]
    ))
  }

  ###################################################@
  getVarNamePos <- function(x) {
    varNamePos <- gregexpr('`[^`]*`', x)[[1]]
    if (varNamePos[1] == -1) {
      return(dplyr::tibble(s = 1, e = 1)[-1, ])
    }
    return(dplyr::tibble(
      s = varNamePos,
      e = varNamePos + attr(varNamePos, "match.length") - 1
    ))
  }

  ###################################################@
  splitByPos <- function(x, pos, keepPos = c('NO', 'right', 'left')) {
    keepPos <- match.arg(keepPos)
    stopifnot(all(pos > 0), all(pos <= nchar(x)), length(x) == 1)
    if (length(pos) == 0) {
      return(x)
    }
    einc <- switch(
      keepPos,
      "NO" = -1,
      "right" = 0,
      "left" = -1
    )
    sinc <- switch(
      keepPos,
      "NO" = +1,
      "right" = +1,
      "left" = 0
    )
    toRet <- list()
    s <- 1
    for (i in 1:length(pos)) {
      p <- pos[i]
      e <- p + einc
      toRet[[i]] <- substr(x, s, e)
      s <- p + sinc
    }
    e <- nchar(x)
    if (s <= e) {
      toRet[[i + 1]] <- substr(x, s, e)
    } else {
      toRet[[i + 1]] <- ""
    }
    return(unlist(toRet))
  }

  ###################################################@
  getSubStatements <- function(x, sep = ",") {
    varNamePos <- getVarNamePos(x)
    brackPos <- getBrackPos(x)
    toExcl <- union(
      unlist(apply(varNamePos, 1, function(x) x[1]:x[2])),
      unlist(apply(brackPos, 1, function(x) x[1]:x[2]))
    )
    splitPos <- setdiff(gregexpr(sep, x)[[1]], c(-1, toExcl))
    toRet <- splitByPos(x, splitPos, keepPos = "NO")
    toRet <- sub("^[[:space:]]*", "", toRet)
    toRet <- sub("[[:space:]]*$", "", toRet)
    return(toRet)
  }

  ###################################################@
  parseCreateTableStatement <- function(cs, typeRef) {
    ## Table ----
    header <- sub("[[:space:]]*[(].*", "", cs)
    p <- gregexpr('(`[^`]*`[.])?`[^`]*`', header)[[1]]
    tableFullName <- substr(header, p, p + attr(p, "match.length") - 1) |>
      strsplit(split = "[.]") |>
      unlist()
    tableFullName <- gsub("`", "", tableFullName)
    if (length(tableFullName) == 1) {
      tableFullName <- c("", tableFullName)
    }
    dbName <- tableFullName[1]
    tableName <- tableFullName[2]
    tableFullName <- paste0("`", dbName, "`.`", tableName, "`")

    ##
    body <- sub("[^(]*[(]", "", cs)
    body <- sub("[)][^)]*$", "", body)
    subStatements <- getSubStatements(body, sep = ",")

    ## Fields ----
    fields <- grep("^[`]", subStatements, value = TRUE)
    types <- sub(" .*$", "", sub("^.*` *", "", fields))
    nullable <- rep(TRUE, length(fields))
    nullable[grep("NOT NULL", fields)] <- FALSE
    # comPos <- gregexpr(" *COMMENT *'[^']*'", fields)
    comPos <- gregexpr(" *COMMENT *'.*'", fields)
    comments <- rep("", length(fields))
    for (i in 1:length(comPos)) {
      p <- comPos[[i]]
      if (p != -1) {
        toAdd <- substr(fields[i], p, p + attr(p, 'match.length') - 1)
        toAdd <- sub(" *COMMENT *'", "", toAdd)
        toAdd <- sub("'$", "", toAdd)
        toAdd <- gsub('\\\\"', '"', toAdd)
        toAdd <- gsub("\\\\'", "'", toAdd)
        comments[i] <- toAdd
      }
    }
    fields <- dplyr::tibble(
      name = sub("[`].*$", "", sub("[`]", "", fields)),
      type = conv_type_ref(types, from = typeRef),
      nullable = nullable,
      comment = comments
    )

    ## Primary key ----
    primaryKey <- grep("^PRIMARY KEY ", subStatements, value = TRUE)
    primaryKey <- sub("^PRIMARY KEY [(]", "", primaryKey)
    primaryKey <- sub("[)]$", "", primaryKey)
    if (length(primaryKey) > 0) {
      primaryKey <- getSubStatements(gsub('`', '', primaryKey), sep = ",")
    } else {
      primaryKey <- NULL
    }

    ## Foreign keys ----
    foreignKeys <- grep(
      "^CONSTRAINT `[^`]*` FOREIGN KEY [(]",
      subStatements,
      value = TRUE
    )
    foreignKeys <- do.call(
      c,
      lapply(
        foreignKeys,
        function(x) {
          fk <- sub(
            '[)].*$',
            "",
            sub("^CONSTRAINT `[^`]*` FOREIGN KEY [(]", "", x)
          )
          fk <- gsub('`', '', getSubStatements(fk, sep = ","))
          ref <- sub(
            "^CONSTRAINT `[^`]*` FOREIGN KEY [(].*[)] REFERENCES ",
            "",
            x
          )
          p <- gregexpr('(`[^`]*`[.])?`[^`]*`', ref)[[1]]
          refTableFullName <- substr(
            ref,
            p,
            p + attr(p, "match.length") - 1
          ) |>
            strsplit(split = "[.]") |>
            unlist()
          refTableFullName <- gsub("`", "", refTableFullName)
          if (length(refTableFullName) == 1) {
            refTableFullName <- c("", refTableFullName)
          }
          refDB <- refTableFullName[1]
          refTable <- refTableFullName[2]
          refTableFullName <- paste0("`", refDB, "`.`", refTable, "`")

          refKeys <- gsub(
            '`',
            '',
            getSubStatements(
              sub(
                '^.*[(]',
                "",
                sub('[)].*$', "", ref)
              ),
              sep = ","
            )
          )
          toRet <- dplyr::tibble(
            from = fk,
            to = refKeys
          )
          toRet <- list(list(
            refDB = refDB,
            refTable = refTable,
            key = toRet
          ))
          names(toRet) <- refTableFullName
          return(toRet)
        }
      )
    )

    ## Indexes ----
    indexes <- grep("^INDEX ", subStatements, value = TRUE)
    indexes <- sub("[)][^)]*$", "", sub("^INDEX [^(]*[(]", "", indexes))
    indexes <- lapply(
      indexes,
      function(x) {
        toRet <- getSubStatements(x, sep = ",")
        if (toRet[1] == "") {
          return(NULL)
        }
        toRet <- do.call(
          rbind,
          lapply(
            toRet,
            function(y) {
              toRet <- getSubStatements(y, sep = " ")
              toRet <- gsub('`', '', toRet)
              return(toRet)
            }
          )
        )
        colnames(toRet) <- c("field", "order")
        toRet <- list(fields = toRet[, "field"], unique = FALSE)
        return(toRet)
      }
    )
    if (length(indexes) > 0) {
      indexes <- indexes[which(!unlist(lapply(indexes, is.null)))]
    } else {
      indexes <- NULL
    }

    ## Unique indexes ----
    uindexes <- grep("^UNIQUE INDEX ", subStatements, value = TRUE)
    uindexes <- sub(
      "[)][^)]*$",
      "",
      sub("^UNIQUE INDEX [^(]*[(]", "", uindexes)
    )
    uindexes <- unlist(lapply(
      uindexes,
      function(x) {
        toRet <- getSubStatements(x, sep = ",")
        if (toRet[1] == "") {
          return(NULL)
        }
        toRet <- do.call(
          rbind,
          lapply(
            toRet,
            function(y) {
              toRet <- getSubStatements(y, sep = " ")
              toRet <- gsub('`', '', toRet)
              return(toRet)
            }
          )
        )
        colnames(toRet) <- c("field", "order")
        toRet <- toRet[, "field"]
        return(toRet)
      }
    ))
    fields <- fields |>
      dplyr::mutate(unique = ifelse(.data$name %in% uindexes, TRUE, FALSE))

    toRet <- list(list(
      dbName = dbName,
      tableName = tableName,
      fields = fields,
      primaryKey = primaryKey,
      foreignKeys = foreignKeys,
      indexes = indexes, #c(indexes, uindexes),
      display = list(
        x = as.numeric(NA),
        y = as.numeric(NA),
        color = as.character(NA),
        comment = as.character(NA)
      )
    ))
    names(toRet) <- tableFullName
    return(toRet)
  }

  ############################################################################@
  ## Main ----

  typeRef <- match.arg(typeRef, list_type_ref())
  txt <- readLines(f) |>
    paste(collapse = "\n")
  txt <- gsub("/\\*((?!\\*/)(.|\n))*\\**/", "", txt, perl = T) |>
    strsplit(split = "\n") |>
    unlist()
  txt <- sub("--.*$", "", txt)
  if (mysqlcomments) {
    txt <- sub("#.*$", "", txt)
  }
  txt <- sub("^[[:space:]]*", "", txt)
  txt <- txt[which(txt != "")] |>
    paste(collapse = " ") |>
    strsplit(split = ";") |>
    unlist()
  txt <- sub("^[[:space:]]*", "", txt)
  txt <- sub("[[:space:]]*$", "", txt)
  txt <- gsub("[[:space:]]+", " ", txt)
  tdTxt <- grep("^CREATE TABLE", txt, ignore.case = TRUE, value = TRUE)
  unsupported <- grep("`", tdTxt, value = TRUE, invert = TRUE)
  if (length(unsupported) > 0) {
    stop(
      sprintf(
        '%s statement%s %s not supported because without "`":\n',
        length(unsupported),
        ifelse(length(unsupported) > 1, "s", ""),
        ifelse(length(unsupported) > 1, "are", "is")
      ),
      paste("   -", unsupported) |> paste(collapse = "\n")
    )
  }
  toRet <- do.call(
    c,
    lapply(
      tdTxt,
      parseCreateTableStatement,
      typeRef = typeRef
    )
  )
  dbNames <- unlist(lapply(toRet, function(x) x$dbName))
  if (length(unique(dbNames)) > 1) {
    stop(
      "Several database names are mentioned in the SQL model. ",
      "There should be only one."
    )
  }
  toRet <- lapply(toRet, function(x) x[which(names(x) != "dbName")])
  toRet <- lapply(toRet, RelTableModel)
  toRet <- RelDataModel(toRet, checkFK = TRUE)
  return(toRet)
}

#' @describeIn read_SQL_data_model
#'
#' Deprecated version of read_SQL_data_model
#'
#' @param ... params for `read_SQL_data_model`
#'
#' @export
readSQLDataModel <- function(...) {
  warning("Deprecated. Use read_SQL_data_model instead")
  read_SQL_data_model(...)
}

###############################################################################@
#' Get a foreign key table from an object
#'
#' @param x a [RelTableModel] or a [RelDataModel]
#'
#' @return A tibble with the following fields:
#' - from: the origin of the key
#' - ff: the key fields in from
#' - to: the target of the key
#' - tf: the key fields in to
#' - fmin: minimum cardinality of from
#' - fmax: maximum cardinality of from
#' - tmin: minimum cardinality of to
#' - tmax: maximum cardinality of to
#'
#' @export
#'
get_foreign_keys <- function(x) {
  UseMethod("get_foreign_keys", x)
}


###############################################################################@
#' Lengths of object elements
#'
#' @param x an object. If there is no method implemented for this object,
#' the [base::lengths()] function is used.
#' @param use.names logical indicating if the result should
#' inherit the names from x.
#'
#' @return A non-negative integer of length length(x), except when any element
#' has a length of more than 2^31 - 1 elements, when it returns a double vector.
#' When use.names is true, the names are taken from the names on x, if any.
#'
#' @seealso [base::lengths()]
#'
#' @export
lengths <- function(x, use.names = TRUE) {
  UseMethod("lengths", x)
}
#' @export
lengths.default <- function(x, use.names = TRUE) {
  base::lengths(unclass(x), use.names = use.names)
}

---
title: "Modeling Relational Data in R with ReDaMoR"
subtitle: "ReDaMoR version: `r packageVersion('ReDaMoR')`"
author: "[Patrice Godard](mailto:patrice.godard@gmail.com?Subject=ReDaMoR)"
date: "`r format(Sys.time(), '%B %d %Y')`"
always_allow_html: true
output:
  html_document:
    keep_md: no
    number_sections: yes
    self_contained: yes
    theme: cerulean
    toc: yes
    toc_float: yes
  pdf_document:
    toc: yes
    number_sections: yes
package: ReDaMoR (version `r packageVersion('ReDaMoR')`)
vignette: >
    %\VignetteIndexEntry{ReDaMoR - Tutorial}
    %\VignetteEncoding{UTF-8}
    %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---

```{r setup, include = FALSE}
library(knitr)
library(ReDaMoR)
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
cranRef <- function(x){
  sprintf(
    "[%s](https://CRAN.R-project.org/package=%s): %s",
    x, x, packageDescription(x)$Title
  )
}
```

<div style="width:200px;">
![](img/ReDaMoR.png){alt="logo"}
</div>

<!----------------------------------------------------------------------------->
<!----------------------------------------------------------------------------->
# Introduction


In R, data are often stored in data frames which are tables in which each row represents a record and each column a variable.
Because data frames are highly used, they have been improved in different objects such as [tibble](https://tibble.tidyverse.org/), [data.table](https://rdatatable.gitlab.io/data.table/) or [AnnotatedDataFrame](https://bioconductor.org/packages/release/bioc/html/Biobase.html).

However, in many projects, the data do not fit in only one table but they are organized in different data frames, each of them dealing with a specific concept.
These tables are often related to each other by some variables.
Depending on the diversity of the concepts, it can be difficult to remember what these tables represent and how they are connected.
Fortunately, because they are tables, a set of data frames can be directly documented using a relational data model.

The [datamodelr](https://github.com/bergant/datamodelr) R package provides tools to document relational data. The generated data models are leveraged by the [dm](https://github.com/cynkra/dm) R package to interact more easily with relational data.

Here we present the ReDaMoR package which also allows the manipulation of relational data models in R but with an approach quite different from the one implemented in datamodelr.
It provides functions to create, import and save relational data models.
These functions are accessible through a graphical user interface made with [Shiny](https://shiny.posit.co/).

The main features of ReDaMoR are the following:

- Create data model from scratch with a graphical user interface
- Save data model in a json format
- Import SQL data model generated with [MySQL Workbench](https://www.mysql.com/products/workbench/)
- Document data types as R types (character, numeric, logical, ...)
- Add comments to tables and fields to improve model documentation
- Check and auto-correct model constraints (types, keys, uniqueness, mandatory fields)
- Confront data to the model to check their compatibility

The ReDaMoR R package is licensed under [GPL-3](https://www.gnu.org/licenses/gpl-3.0.en.html)
and its source code is available on [github](https://github.com/patzaw/ReDaMoR).

<!----------------------------------------------------------------------------->
<!----------------------------------------------------------------------------->
# Installation

<!---->
## From CRAN
<!----------->

```{r, eval=FALSE}
install.packages("ReDaMoR")
```

<!---->
## Dependencies
<!----------->

The following R packages available on CRAN are required:

- `r cranRef("dplyr")`
- `r cranRef("visNetwork")`
- `r cranRef("readr")`
- `r cranRef("shiny")`
- `r cranRef("shinyjs")`
- `r cranRef("DT")`
- `r cranRef("rintrojs")`
- `r cranRef("colourpicker")`
- `r cranRef("jsonlite")`
- `r cranRef("markdown")`
- `r cranRef("rstudioapi")`
- `r cranRef("crayon")`
- `r cranRef("Matrix")`
<!-- - utils: [R](https://www.r-project.org/) core package -->
<!-- - graphics: [R](https://www.r-project.org/) core package -->
<!-- - stats: [R](https://www.r-project.org/) core package -->

<!---->
## From github
<!----------->

```{r, eval=FALSE}
devtools::install_github("patzaw/ReDaMoR")
```

<!----------------------------------------------------------------------------->
<!----------------------------------------------------------------------------->
# Creating and modifying relational data using the graphical user interface

<!---->
## Running the app
<!--------------->

The Shiny app is launched with the following command:

```{r, eval=FALSE}
library(ReDaMoR)
m <- model_relational_data()
```

When the ***Done*** button is clicked, the model is returned in R environment.
Because the interface can be closed accidentally, the model is also autosaved and can be recovered using the `recover_RelDataModel()` function.
The recovered model can be provided as `modelInput` when calling `model_relational_data()`.
For example:

```{r, eval=FALSE}
m <- model_relational_data(recover_RelDataModel())
```

<!---->
## Model formalism
<!------------------------------------------------->

### Data model example: the Human Phenotype Ontology
<!------------------------------------------------->

A data model example is provided within the package.
It represents data extracted from the [Human Phenotype Ontology (HPO)](https://hpo.jax.org/)<sup>([1](#hpoRef))</sup> and for which a subset is also provided within the ReDaMoR package (more details provided in the [Confronting data section](#confronting-data)).

This example can be imported from the Shiny app when clicking on the ***Import*** button and then on the ***Try an example*** link.
It can be also be loaded and displayed (and edited) using the following commands:

```{r}
hpo_model <- read_json_data_model(
  system.file("examples/HPO-model.json", package="ReDaMoR")
)
plot(hpo_model)
```


```{r}
## Edit the model
# m <- model_relational_data(hpo_model)
```

The view is rendered by the [visNetwork package](https://datastorm-open.github.io/visNetwork/). It means that it can take advantages of all the functions provided by the visNetwork package including the use of the model in [Shiny apps](https://datastorm-open.github.io/visNetwork/shiny.html).

### Tables
<!------->

Each box represents a table.
The header of the box corresponds to the name of the table and the following lines document each field:

- The **type** of the field is given between curly brackets
- A star indicates that the field is **unique**
- Fields surrounded by parentheses are **optional**
- Fields in bold are involved in the **primary key**
- **Indexes** are also indicated between the curly brackets when they exist:
  - idx. correspond to regular indexes
  - uidx. correspond to unique indexes
  
When the cursor is over a box, table and field **comments** are displayed.

### Foreign keys
<!------------->

Each arrow represent a foreign key:

- The start of the arrow corresponds to the table with the **foreign key**.
- The end of the arrow corresponds to the **referenced table**.
- When the mouse is over the arrow, the **fields** involved in the foreign key are displayed along with the minimum and maximum **cardinalities** of the key on both sides (between brackets below the table name).

<!---->
## App organization
<!---------------->

<div style="width:100%;">
![shiny app](img/model_relational_data-Overview-Divided.png)
</div>

The app is divided in three main parts:

- ![a part](img/ica.png){width=30px} The **main menu** at the top of the page to execute general actions on the model
- ![b part](img/icb.png){width=30px} The **model view** on the left of the page to interact with the model
- ![c part](img/icc.png){width=30px} The **edition view** on the right of the page to execute contextual edition of the model

### The main menu
<!-------------->

General actions can be executed on the model, using the buttons in the main menu:

- Import json or sql (Only sql models produced by MySQL Workbench are supported) files
- Export json or html files (a png file can also be saved by right clicking on the model view)
- Add a table

### The model view
<!--------------->

The model view provides a view of the data model rendered by the [visNetwork package](https://datastorm-open.github.io/visNetwork/).
Tables can be selected by clicking on it or by searching them using the box just above the model view.
Autodraw and autofit capabilities are also provided in this area.

### The edition view
<!----------------->

In the edition view the user can:

- Change the color of a table
- Duplicate or remove tables
- Add and remove foreign keys
- Edit table information
  - table name
  - table comment
  - fields
  - primary key
  - indexes

### Contextual help tour
<!--------------------->

An help tour can be launched when clicking on the ![help icon](img/bhelp.png){width=40px} button in the main menu.
This help tour is contextual: its content depends on the state of the app.
  
### Keyboard shortcuts
<!------------------->

Some *common* keyboard shortcuts are implemented:

- **Ctrl+Z** / **Ctrl+Shift+Z** for undo and redo
- **Enter** for validating a new name
- **del** for deleting tables or foreign keys
- **F2** for renaming tables or editing foreign key cardinalities

## Adding foreign keys {-}

In the interface, when you select one or two tables (holding down the ctrl key)
you can create a foreign key by clicking on the “Key”  button
(See in the figure below surrounded in blue).

<div style="width:100%;">
![foreign key icon](img/Add-key-button.png)
</div>

The following interface should appear:

<div style="width:100%;">
![foreign key interface](img/Add-key-interface.png)
</div>

For each table you can select the fields participating to the relationship
(surrounded in blue) and add them with the “+” button (surrounded in red).
You can also change the cardinality of the 2 sides of the relationship
(surrounded in yellow). You can also change the direction of the relationship
with the arrow button (surrounded in red). When you've finished,
you can click on the “Add” button (surrounded in purple) to create
the foreign key.
 
After having created a relationship you can still change its cardinalities
by selecting the relationship of interest and by clicking on the edit “keys”
button. If you need to apply more changes, you'll have to delete the key
("Discard keys" button) and recreate it.

<!----------------------------------------------------------------------------->
<!----------------------------------------------------------------------------->
# Confronting data

Data can be confronted to the model using the `confront_data()` function.
During this process the following checks are performed:

- All the tables in the model and only those are **available**.
- For each table:
  - All the columns in the model and only those are **available**.
  - The **type** of each column corresponds to the type in the model.
  - Not any *NA* values are found in **non-nullable** columns.
  - There is no duplicated value in **unique** columns.
  - There is no duplicated value in **unique indexes**.
- **Foreign keys** are tested according to their cardinality: all values in one table should be available in the other table if the corresponding minimum cardinality is greater than 0. 

A **subset** (500 phenotypes among more than 14&nbsp;000 in the original resource) of the [HPO](https://hpo.jax.org/)<sup>([1](#hpoRef))</sup> is provided within the package and can be directly confronted to the data model.
The function prints a report with messages about the global success of the data model confrontation and additional failure or warning messages for relevant tables.

```{r}
confrontation_report <- confront_data(
  hpo_model,
  path=list.files(
    system.file("examples/HPO-subset", package="ReDaMoR"),
    full.names=TRUE
  ),
  returnData=TRUE
)
```

This report can also be formatted using markdown and included in a document such as this one.

<div style="background-color:#FFFACD; border-top:5px solid black; border-bottom:5px solid black; margin-top:50px; margin-bottom:50px">

```{r, results='asis'}
# view_confrontation_report(confrontation_report) # Use RStudio viewer
format_confrontation_report_md(
  confrontation_report,
  title="Example: Confrontation with original data",
  level=1, numbered=FALSE
) |>
  cat()
```

</div>

The `returnData` argument is used to keep the data in the confrontation report.

```{r}
hpo_tables <- confrontation_report$data
```

In the example below the data are altered and confronted again to the data model.
The confrontation report shows the discrepancies between the model and the data in order to facilitate the correction of one or the other.

<div style="background-color:#FFFACD; border-top:5px solid black; border-bottom:5px solid black; margin-top:50px; margin-bottom:50px">

```{r, results='asis'}
hpo_tables$HPO_diseases <- hpo_tables$HPO_diseases |> slice(1:100)
hpo_tables$HPO_synonyms[1:10, "synonym"] <- NA
hpo_tables$HPO_hp <- hpo_tables$HPO_hp |> mutate(level=as.character(level))
confront_data(hpo_model, hpo_tables, verbose=FALSE) |>
  format_confrontation_report_md(
    title="Example: Confrontation with altered data",
    level=1, numbered=FALSE
  ) |>
    cat()
```

</div>

<!---->
## Drafting a data model from data frames
<!-------------------------------------->

Beside confronting a model to data, a model can also be drafted from a set of tables.
Only column names and types are documented during this process.
Uniqueness, mandatory and key constraints still need to be documented by the user to complete the model. From version 0.8.0, the function `guess_constraints()` guesses unique and non-nullable fields, along with potential foreign keys, based on the provided or existing tables.
The guessed constraints should be carefully reviewed, especially the foreign keys. This can be easily done with `model_relational_data()`.
The `auto_layout()` function used below requires the suggested package `r cranRef("igraph")`.


```{r}
hpo_tables <- confrontation_report$data

## Build the model from a list of data frames ----
new_model <- df_to_model(
   list=names(hpo_tables), envir=as.environment(hpo_tables)
)
## Guess constraints and auto layout ----
new_model <- guess_constraints(new_model, data = hpo_tables) |>
   auto_layout(lengthMultiplier=250)

## Plot the model ----
new_model |>
   plot()
```


```{r}
# model_relational_data(new_model)
```


<!----------------------------------------------------------------------------->
<!----------------------------------------------------------------------------->
# Support of matrices

Matrices (2 dimensions) are supported by ReDaMoR (>= 0.5).
A matrix is modeled as a table
with 3 and only 3 fields: 2 of types *row* and *column* and the third
of your choice. The *row* and *column* compose the primary index of this
special table, but they cannot be unique independently (a cell in a matrix
is always identified by its row and its column). The example below shows a model
of a *gene_expression* matrix: the rows correspond to genes, the column to
samples and the value to the gene expression value in each sample.
Genes and samples are further described in the *genes* and *samples* tables.

```{r}
ge_model <- read_json_data_model(
  system.file("examples/GE-model.json", package="ReDaMoR")
)
plot(ge_model)
```

<!----------------------------------------------------------------------------->
<!----------------------------------------------------------------------------->
# TKCat: Tailored Knowledge Catalog

The aim of [TKCat](https://patzaw.github.io/TKCat/) (Tailored Knowledge Catalog)
is to facilitate the management
of data from knowledge resources that are frequently used alone or together in
research environments. In TKCat, knowledge resources are manipulated as modeled
database (MDB) objects. These objects provide access to the data tables along
with a general description of the resource and a detail data model generated
with ReDaMoR documenting the tables, their fields and their relationships.
These MDB are then gathered in catalogs that can be easily explored an shared.
TKCat provides tools to easily subset, filter and combine MDBs and create new
catalogs suited for specific needs.

A modeled database (MDB) in TKCat gathers the following information:

-   General database information including a mandatory *name* and optionally the
    following fields: *title*, *description*, *url*, *version* and *maintainer*.
-   A data model created using the [ReDaMoR][redamor] package.
-   A list of tables corresponding to reference concepts shared by different
    MDBs. The way these concepts are identified is defined in specific documents
    called collections.
-   The data themselves organized according to the data model.

`MDB` objects can be subsetted, combined, filtered, joined and merged.
[TKCat](https://patzaw.github.io/TKCat/) supports different implementations
of `MDB` objects such as `memoMDB` (data in memory), `fileMDB` (data in files)
and `chMDB` (data in [ClickHouse](https://clickhouse.com/) database).
More information is provided in the
[TKCat user guide](https://patzaw.github.io/TKCat/articles/TKCat.html).

<!----------------------------------------------------------------------------->
<!----------------------------------------------------------------------------->
# Acknowledgments

This work was entirely supported by [UCB Pharma](https://www.ucb.com/).
(Early Solutions department)


<!----------------------------------------------------------------------------->
<!----------------------------------------------------------------------------->
# References

<p id="hpoRef">**(1)**</a> [Sebastian Köhler, Leigh Carmody, Nicole Vasilevsky, Julius O B Jacobsen, et al. **Expansion of the Human Phenotype Ontology (HPO) knowledge base and resources.** *Nucleic Acids Research.* (2018) doi: 10.1093/nar/gky1105](https://doi.org/10.1093/nar/gky1105)</p>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{set_primary_key}
\alias{set_primary_key}
\title{Set the primary key a table in a \link{RelDataModel}}
\usage{
set_primary_key(x, tableName, fieldNames)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{tableName}{the name of the table to modify (a single character)}

\item{fieldNames}{the names of the fields to include in the primary key}
}
\value{
A \link{RelDataModel}
}
\description{
Set the primary key a table in a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelTableModel.R
\name{identical_RelTableModel}
\alias{identical_RelTableModel}
\title{Check if two \link{RelTableModel} are identical}
\usage{
identical_RelTableModel(x, y, includeDisplay = TRUE)
}
\arguments{
\item{x}{a \link{RelTableModel}}

\item{y}{a \link{RelTableModel}}

\item{includeDisplay}{a single logical (default: TRUE) indicating if
the display should be included in the comparison}
}
\value{
A logical: TRUE if the 2 models are identical
}
\description{
Check if two \link{RelTableModel} are identical
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TYPEDEF.R
\name{list_type_ref}
\alias{list_type_ref}
\title{List supported types references}
\usage{
list_type_ref()
}
\description{
List supported types references
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{set_unique_index}
\alias{set_unique_index}
\title{Set table index uniqueness in a \link{RelDataModel}}
\usage{
set_unique_index(x, tableName, fieldNames, unique)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{tableName}{the name of the table to modify (a single character)}

\item{fieldNames}{the names of the fields composing the index}

\item{unique}{a logical value}
}
\value{
A \link{RelDataModel}
}
\description{
Set table index uniqueness in a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{RelDataModel}
\alias{RelDataModel}
\title{Create a RelDataModel object}
\usage{
RelDataModel(l, checkFK = TRUE, createFKIndex = FALSE)
}
\arguments{
\item{l}{the list of table models (\link{RelTableModel} objects)}

\item{checkFK}{a logical indicating if foreign keys should be checked
(default: TRUE)}

\item{createFKIndex}{should an index be create for the foreign keys
(default: FALSE)}
}
\value{
A RelDataModel object.
}
\description{
Create a RelDataModel object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{remove_foreign_key}
\alias{remove_foreign_key}
\title{Remove a foreign key between two tables}
\usage{
remove_foreign_key(x, fromTable, fromFields, toTable, toFields)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{fromTable}{the name of the referencing table}

\item{fromFields}{the name of the referencing fields}

\item{toTable}{the name of the referenced table}

\item{toFields}{the names of the referenced fields}
}
\value{
A \link{RelDataModel}
}
\description{
Remove a foreign key between two tables
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TYPEDEF.R
\name{norm_type_ref}
\alias{norm_type_ref}
\title{Normalize type names}
\usage{
norm_type_ref(x, typeRef, ignore.case = TRUE)
}
\arguments{
\item{x}{a character vector to normalize}

\item{typeRef}{a character vector of length one: the type
reference (\link{list_type_ref})}

\item{ignore.case}{should case be ignored (default: TRUE)}
}
\description{
Normalize type names
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{identical_RelDataModel}
\alias{identical_RelDataModel}
\title{Check if two \link{RelDataModel} are identical}
\usage{
identical_RelDataModel(x, y, ...)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{y}{a \link{RelDataModel}}

\item{...}{additional parameters for \code{\link[=identical_RelTableModel]{identical_RelTableModel()}}}
}
\value{
A logical: TRUE if the 2 models are identical
}
\description{
Check if two \link{RelDataModel} are identical
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{check_foreign_keys}
\alias{check_foreign_keys}
\title{Check the availability of foreign keys}
\usage{
check_foreign_keys(x)
}
\arguments{
\item{x}{a \link{RelDataModel} object}
}
\value{
Nothing. The function throws an error if there is an issue with
foreign keys.
}
\description{
Check the availability of foreign keys
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/df_to_model.R
\name{df_to_model}
\alias{df_to_model}
\title{Create a \link{RelDataModel} object from column names of data frames}
\usage{
df_to_model(..., list = character(), pos = -1, envir = as.environment(pos))
}
\arguments{
\item{...}{the data frame objects, as names (unquoted) or character strings
(quoted)}

\item{list}{a character vector naming data frame objects}

\item{pos}{where to get the objects.
By default, uses the current environment.
See ‘details’ for other possibilities.}

\item{envir}{the environment to use. See ‘details’.}
}
\value{
A \link{RelDataModel} object.
}
\description{
Create a \link{RelDataModel} object from column names of data frames
}
\details{
The pos argument can specify the environment from which to get
the objects in any of several ways:
as an integer (the position in the search list);
as the character string name of an element in the search list;
or as an environment.
The envir argument is an alternative way to specify an environment,
but is primarily there for back compatibility.
}
\examples{
## Read data files ----
to_read <- list.files(
  system.file("examples/HPO-subset", package = "ReDaMoR"),
  full.names = TRUE
)
hpo_tables <- list()
for (f in to_read) {
  hpo_tables[[sub("[.]txt$", "", basename(f))]] <- readr::read_tsv(f)
}
## Build the model from a list of data frames ----
new_model <- df_to_model(
  list = names(hpo_tables),
  envir = as.environment(hpo_tables)
)
## Guess constraints and auto layout ----
new_model <- guess_constraints(new_model, data = hpo_tables) |>
  auto_layout(lengthMultiplier = 250)

## Plot the model ----
new_model |>
  plot()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TYPEDEF.R
\name{conv_type_ref}
\alias{conv_type_ref}
\title{Convert a set of types from or to R supported types}
\usage{
conv_type_ref(x, from = NULL, to = NULL, ignore.case = TRUE)
}
\arguments{
\item{x}{a character vector of types to be converted. If from is not null,
x should be a set of valid types in the from reference. If to is not null,
x should be a set of supported R types (SUPPTYPES).}

\item{from}{a character vector of length one: the type reference
(\link{list_type_ref}) of x}

\item{to}{a character vector of length one: the targeted type reference
(\link{list_type_ref})}

\item{ignore.case}{should case be ignored when converting `from``
type reference (default: TRUE)}
}
\description{
Convert a set of types from or to R supported types
}
\details{
Only \code{from} XOR \code{to} should be set
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readSQLDataModel.R
\name{read_SQL_data_model}
\alias{read_SQL_data_model}
\alias{readSQLDataModel}
\title{Read a data model from an SQL file from the MySQL Workbench}
\usage{
read_SQL_data_model(f, typeRef = "MySQLWB", mysqlcomments = TRUE)

readSQLDataModel(...)
}
\arguments{
\item{f}{the SQL file to read}

\item{typeRef}{the reference for type conversion
(Default: "MySQLWB"; see \code{\link[=list_type_ref]{list_type_ref()}})}

\item{mysqlcomments}{if MySQL comments (starting with #) should be removed
(Default: TRUE)}

\item{...}{params for \code{read_SQL_data_model}}
}
\value{
A \link{RelDataModel} object
}
\description{
Read a data model from an SQL file from the MySQL Workbench
}
\details{
Database, table and field names should be surrounded by "`".
}
\section{Functions}{
\itemize{
\item \code{readSQLDataModel()}: Deprecated version of read_SQL_data_model

}}
\examples{
## Read the model ----
hpo_from_sql <- read_SQL_data_model(
  system.file("examples/HPO-model.sql", package = "ReDaMoR")
)
## Confront to data ----
confrontation_report <- confront_data(
  hpo_from_sql,
  path = list.files(
    system.file("examples/HPO-subset", package = "ReDaMoR"),
    full.names = TRUE
  ),
  returnData = TRUE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format_confrontation_report.R
\name{format_confrontation_report_md}
\alias{format_confrontation_report_md}
\title{Format confrontation report in markdown format}
\usage{
format_confrontation_report_md(
  cr,
  title = "Model",
  level = 0,
  numbered = TRUE,
  bgSuccess = "green",
  txSuccess = "black",
  bgFailure = "red",
  txFailure = "white",
  bgMessage = "#FFBB33",
  txMessage = "white"
)
}
\arguments{
\item{cr}{the confrontation report from \link{confront_data}}

\item{title}{a character with a single value corresponding to the report}

\item{level}{rmarkdown level in document hierarchy (default:0 ==> highest).
It should be an integer between 0 and 4.}

\item{numbered}{a logical. If TRUE (default) the sections are part of
document numbering.}

\item{bgSuccess}{background color for SUCCESS}

\item{txSuccess}{text color for SUCCESS}

\item{bgFailure}{background color for FAILURE}

\item{txFailure}{text color for FAILURE}

\item{bgMessage}{background color for a warning message}

\item{txMessage}{text color for a warning message}
}
\description{
Format confrontation report in markdown format
}
\examples{
## Read the model ----
hpo_from_sql <- read_SQL_data_model(
  system.file("examples/HPO-model.sql", package = "ReDaMoR")
)
## Confront to data ----
confrontation_report <- confront_data(
  hpo_from_sql,
  path = list.files(
    system.file("examples/HPO-subset", package = "ReDaMoR"),
    full.names = TRUE
  ),
  verbose = FALSE,
  returnData = TRUE
)
## Show the report in console ----
format_confrontation_report(confrontation_report) |> cat()
## Format the report using markdown ----
format_confrontation_report_md(confrontation_report) |> cat()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{rename_table}
\alias{rename_table}
\title{Rename a table in a \link{RelDataModel}}
\usage{
rename_table(x, old, new)
}
\arguments{
\item{x}{a \link{RelDataModel} object}

\item{old}{a single character corresponding to the table name to change}

\item{new}{the new table name}
}
\value{
A \link{RelDataModel}
}
\description{
Rename a table in a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/named_MM.R
\name{read_named_MM_header}
\alias{read_named_MM_header}
\title{Read the header of a named sparse matrix in MatrixMarket text format}
\usage{
read_named_MM_header(file, guess_max = 20)
}
\arguments{
\item{file}{the file to read}

\item{guess_max}{the number of lines to read to find the header.
(4 should be sufficient. Default: 20)}
}
\value{
A list with the following fields:
\itemize{
\item rownames: a character vector with the matrix row names
\item colnames: a character vector with the matrix column names
\item rows: the number of matrix rows
\item columns: the number of matrix columns
\item values: the number of values in the matrix
\item header_length: the number of lines in the header
}
}
\description{
Read the header of a named sparse matrix in MatrixMarket text format
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{rename_field}
\alias{rename_field}
\title{Rename an existing field in a \link{RelDataModel} table}
\usage{
rename_field(x, tableName, current, new)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{tableName}{the name of the table to modify (a single character)}

\item{current}{the current name of the field to modify (a single character)}

\item{new}{the new name of the field (a single character)}
}
\value{
A \link{RelDataModel}
}
\description{
Rename an existing field in a \link{RelDataModel} table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelTableModel.R
\name{format.RelTableModel}
\alias{format.RelTableModel}
\title{Format a \link{RelTableModel} object for printing}
\usage{
\method{format}{RelTableModel}(x, ...)
}
\arguments{
\item{x}{a \link{RelTableModel} object}

\item{...}{for generics compatibility (not used)}
}
\value{
A single character
}
\description{
Format a \link{RelTableModel} object for printing
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{copy_fields}
\alias{copy_fields}
\title{Copy fields from one table to another in a \link{RelDataModel}}
\usage{
copy_fields(x, from, to, fields)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{from}{the name of the table from which the fields are taken}

\item{to}{the name of the table to which the fields are copied}

\item{fields}{the names of the fields to copy}
}
\value{
A \link{RelDataModel}
}
\description{
Copy fields from one table to another in a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/named_MM.R
\name{is_MM}
\alias{is_MM}
\title{Identify if a file is in MatrixMarket text format}
\usage{
is_MM(file)
}
\arguments{
\item{file}{the file to read}
}
\value{
A logical. If FALSE, the first line of the file is returned as
an attribute named "r1": \code{attr(is_MM, "r1")}
}
\description{
Identify if a file is in MatrixMarket text format
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelTableModel.R
\name{is.RelTableModel}
\alias{is.RelTableModel}
\title{Check if the object is  a \link{RelTableModel} object}
\usage{
is.RelTableModel(x)
}
\arguments{
\item{x}{any object}
}
\value{
A single logical: TRUE if x is a \link{RelTableModel} object
}
\description{
Check if the object is  a \link{RelTableModel} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{remove_index}
\alias{remove_index}
\title{Remove an index from a table in a \link{RelDataModel}}
\usage{
remove_index(x, tableName, fieldNames)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{tableName}{the name of the table to modify (a single character)}

\item{fieldNames}{the names of the fields composing the index}
}
\value{
A \link{RelDataModel}
}
\description{
Remove an index from a table in a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{add_index}
\alias{add_index}
\title{Add an index to a table in a \link{RelDataModel}}
\usage{
add_index(x, tableName, fieldNames, unique)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{tableName}{the name of the table to modify (a single character)}

\item{fieldNames}{the names of the fields to include in the index}

\item{unique}{a logical indicating if the indexed values are unique}
}
\value{
A \link{RelDataModel}
}
\description{
Add an index to a table in a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{add_field}
\alias{add_field}
\title{Add a field to a table in a \link{RelDataModel}}
\usage{
add_field(x, tableName, name, type, nullable, unique, comment)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{tableName}{the name of the table to modify (a single character)}

\item{name}{the name of the field to add (a single character)}

\item{type}{the type of the field (a single character)}

\item{nullable}{if the field is nullable (a single logical)}

\item{unique}{if the values are unique (a single logical)}

\item{comment}{a description (a single character)}
}
\value{
A \link{RelDataModel}
}
\description{
Add a field to a table in a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_data_model.R
\name{read_json_data_model}
\alias{read_json_data_model}
\title{Read a data model from JSON}
\usage{
read_json_data_model(txt)
}
\arguments{
\item{txt}{a JSON string, URL or file}
}
\description{
Read a data model from JSON
}
\examples{
## Read the model ----
hpo_model <- read_json_data_model(
  system.file("examples/HPO-model.json", package = "ReDaMoR")
)
## Confront to data ----
confrontation_report <- confront_data(
  hpo_model,
  path = list.files(
    system.file("examples/HPO-subset", package = "ReDaMoR"),
    full.names = TRUE
  ),
  returnData = TRUE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{add_foreign_key}
\alias{add_foreign_key}
\title{Add a foreign key between two tables}
\usage{
add_foreign_key(
  x,
  fromTable,
  fromFields,
  toTable,
  toFields,
  fmin = 0L,
  fmax = -1L,
  tmin = 1L,
  tmax = 1L
)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{fromTable}{the name of the referencing table}

\item{fromFields}{the name of the referencing fields}

\item{toTable}{the name of the referenced table}

\item{toFields}{the names of the referenced fields}

\item{fmin}{from minimum cardinality (default: 0L)}

\item{fmax}{from maximum cardinality (default: -1L ==> Infinite)}

\item{tmin}{to minimum cardinality (default: 1L)}

\item{tmax}{to maximum cardinality (default: 1L)}
}
\value{
A \link{RelDataModel}
}
\description{
Add a foreign key between two tables
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TYPEDEF.R
\name{as_type}
\alias{as_type}
\title{Convert an object into a specific type}
\usage{
as_type(x, type)
}
\arguments{
\item{x}{an object to convert}

\item{type}{the targeted type}
}
\description{
Convert an object into a specific type
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelTableModel.R
\name{correct_constraints}
\alias{correct_constraints}
\title{Correct the constraints of a table to make them consistent}
\usage{
correct_constraints(x, createPKIndex = FALSE)
}
\arguments{
\item{x}{a \link{RelTableModel} object}

\item{createPKIndex}{should an index be create for the primary key
(default: FALSE)}
}
\description{
Correct the constraints of a table to make them consistent
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{confront_data}
\alias{confront_data}
\title{Confront a \link{RelDataModel} to actual data}
\usage{
confront_data(
  x,
  data = list(),
  paths = NULL,
  returnData = FALSE,
  verbose = TRUE,
  n_max = Inf,
  checks = if (n_max == Inf) {
     c("unique", "not nullable", "foreign keys")
 } else
    {
     as.character()
 },
  delim = "\\t",
  ...
)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{data}{a list of data frames to be confronted with the model.}

\item{paths}{a character vector with file paths taken into account if
the data is empty.
The file \link{basename} without extension
will be considered as the table name.}

\item{returnData}{a logical indicating if the data should be returned
with the report (default: FALSE).}

\item{verbose}{a single logical value indicating if some process information
should be displayed (default: TRUE)}

\item{n_max}{maximum number of records to read (default: Inf).}

\item{checks}{a character vector with the name of optional checks to be done
(Default: if n_max==Inf ==> all of
them c("unique", "not nullable", "foreign keys"),
else ==> none)}

\item{delim}{single character used to separate fields within a record
(default: "\\t")}

\item{...}{supplementary parameters for the \link[readr:read_delim]{readr::read_delim} function.}
}
\value{
A report as a list
}
\description{
Confront a \link{RelDataModel} to actual data
}
\examples{
## Read the model ----
hpo_model <- read_json_data_model(
  system.file("examples/HPO-model.json", package = "ReDaMoR")
)
## Confront to data ----
confrontation_report <- confront_data(
  hpo_model,
  path = list.files(
    system.file("examples/HPO-subset", package = "ReDaMoR"),
    full.names = TRUE
  ),
  returnData = TRUE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/df_to_model.R
\name{guess_constraints}
\alias{guess_constraints}
\title{Guess \link{RelDataModel} constraints based on the provided or existing tables}
\usage{
guess_constraints(
  x,
  data = NULL,
  env = parent.frame(n = 1),
  constraints = c("unique", "not nullable", "foreign keys")
)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{data}{a named list of tables. All names of x should exist in data. If
NULL, the data are taken from env.}

\item{env}{the R environment in which to find the tables}

\item{constraints}{the type of constraints to guess}
}
\value{
A \link{RelDataModel}
}
\description{
Guess \link{RelDataModel} constraints based on the provided or existing tables
}
\details{
The guessed constraints should be carefully review, especially the foreign
keys.

Complex foreign keys involving multiple fields are not guessed.
}
\examples{
## Read data files ----
to_read <- list.files(
  system.file("examples/HPO-subset", package = "ReDaMoR"),
  full.names = TRUE
)
hpo_tables <- list()
for (f in to_read) {
  hpo_tables[[sub("[.]txt$", "", basename(f))]] <- readr::read_tsv(f)
}
## Build the model from a list of data frames ----
new_model <- df_to_model(
  list = names(hpo_tables),
  envir = as.environment(hpo_tables)
)
## Guess constraints and auto layout ----
new_model <- guess_constraints(new_model, data = hpo_tables) |>
  auto_layout(lengthMultiplier = 250)

## Plot the model ----
new_model |>
  plot()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/named_MM.R
\name{read_named_MM}
\alias{read_named_MM}
\title{Read a named sparse matrix in MatrixMarket text format}
\usage{
read_named_MM(
  file,
  skip = 0,
  n_max = Inf,
  class = c("dgCMatrix", "tibble"),
  guess_max = 20
)
}
\arguments{
\item{file}{the file to read}

\item{skip}{the number of records to skip (default: 0)}

\item{n_max}{the maximum number of records to read (default: Inf)}

\item{class}{the class of object to return. By default a "dgCMatrix".
If "tibble" is chosen, the sparse matrix is returned as a tibble with
3 columns: i (row index), j (column index) and x (values) and
an "header" attribute containing the matrix rownames and colnames.}

\item{guess_max}{the number of lines to read to find the header.
(see \code{\link[=read_named_MM_header]{read_named_MM_header()}})}
}
\value{
By default a dgCMatrix.
If the "tibble" class is chosen, the sparse matrix is returned as
a tibble with 3 columns: i (row index), j (column index) and x (values) and
an "header" attribute containing the matrix rownames and colnames.
}
\description{
Read a named sparse matrix in MatrixMarket text format
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{fromDBM}
\alias{fromDBM}
\title{Convert a list of 5 normalized tibbles in a \link{RelDataModel} object}
\usage{
fromDBM(dbm)
}
\arguments{
\item{dbm}{a list with the following tibbles:
\itemize{
\item \strong{tables}: The tables in the model with the following information
\itemize{
\item \strong{name}: the name of the table
\item \strong{x}: the x coordinate of the table in the model drawing
(NA ==> position undefined)
\item \strong{y}: the y coordinate of the table in the model drawing
(NA ==> position undefined)
\item \strong{color}: the color of the table in the model drawing
(NA ==> undefined)
\item \strong{comment}: comment about the table
}
\item \strong{fields}: The fields in the model with the following information
\itemize{
\item \strong{name}: the name of the field
\item \strong{type}: the type of the field
\item \strong{nullable}: a logical indicating if the field can be null
\item \strong{comment}: comment about the field
\item \strong{table}: the name of the table to which the field belongs
}
\item \strong{primaryKeys}: The primary keys in the model with the
following information
\itemize{
\item \strong{table}: the name of the relevant table
\item \strong{field}: the name of the field participating to the primary key
}
\item \strong{foreignKeys}: The foreign keys in the model with the
following information
\itemize{
\item \strong{table}: the name of the referring table
\item \strong{fki}: the identifier of the foreign key (by referring table)
\item \strong{field}: the name of the referring field
\item \strong{refTable}: the name of the referred table
\item \strong{refField}: the name of the referred field
}
\item \strong{indexes}: The indexes in the model with the following information
\itemize{
\item \strong{table}: the name of the relevant table
\item \strong{idx}: the identifier of the index (by table)
\item \strong{field}: the name of the field participating to the index
\item \strong{unique}: a logical indicating if the field is unique
}
}}
}
\value{
A \link{RelDataModel} object
}
\description{
Convert a list of 5 normalized tibbles in a \link{RelDataModel} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{remove_field}
\alias{remove_field}
\title{Remove a field from a table in a \link{RelDataModel}}
\usage{
remove_field(x, tableName, fieldName, rmForeignKeys = FALSE)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{tableName}{the name of the table to modify (a single character)}

\item{fieldName}{the name of the field to remove (a single character)}

\item{rmForeignKeys}{a single logical indicating if the corresponding foreign
keys should be removed. If FALSE (default), the function will throw an
error if it encounter a foreign key using the field.}
}
\value{
A \link{RelDataModel}
}
\description{
Remove a field from a table in a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{update_table_display}
\alias{update_table_display}
\title{Update the display of a table of a \link{RelDataModel}}
\usage{
update_table_display(
  x,
  tableName,
  px = NULL,
  py = NULL,
  color = NULL,
  comment = NULL
)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{tableName}{the name of the table to modify (a single character)}

\item{px}{the position of the table: x value}

\item{py}{the position of the table: y value}

\item{color}{the color of the table}

\item{comment}{a table description/comment}
}
\value{
A \link{RelDataModel}
}
\description{
Update the display of a table of a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{lengths}
\alias{lengths}
\title{Lengths of object elements}
\usage{
lengths(x, use.names = TRUE)
}
\arguments{
\item{x}{an object. If there is no method implemented for this object,
the \code{\link[base:lengths]{base::lengths()}} function is used.}

\item{use.names}{logical indicating if the result should
inherit the names from x.}
}
\value{
A non-negative integer of length length(x), except when any element
has a length of more than 2^31 - 1 elements, when it returns a double vector.
When use.names is true, the names are taken from the names on x, if any.
}
\description{
Lengths of object elements
}
\seealso{
\code{\link[base:lengths]{base::lengths()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_relational_data.R
\name{clean_autosaved_RelDataModels}
\alias{clean_autosaved_RelDataModels}
\title{Remove all autosaved \link{RelDataModel}}
\usage{
clean_autosaved_RelDataModels()
}
\description{
Remove all autosaved \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelTableModel.R
\name{confront_table_data}
\alias{confront_table_data}
\title{Confront a \link{RelTableModel} to actual data}
\usage{
confront_table_data(x, d, checks = c("unique", "not nullable"))
}
\arguments{
\item{x}{a \link{RelTableModel}}

\item{d}{a data frame or a matrix for matrix model}

\item{checks}{a character vector with the name of optional checks to be done
(Default: all of them c("unique", "not nullable"))}
}
\value{
A report as a list
}
\description{
Confront a \link{RelTableModel} to actual data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TYPEDEF.R
\name{SUPPTYPES}
\alias{SUPPTYPES}
\title{Supported R types}
\usage{
SUPPTYPES
}
\description{
Supported R types
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{auto_layout}
\alias{auto_layout}
\title{Pre-compute \link{RelDataModel} layout when missing any x or y table position}
\usage{
auto_layout(
  x,
  layout = "layout_nicely",
  lengthMultiplier = 40 * length(x),
  force = FALSE
)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{layout}{character name of igraph layout function to use
(Default: "layout_nicely").}

\item{lengthMultiplier}{a numeric value to scale x and y coordinate
(default: 40*length(x))}

\item{force}{if TRUE autolayout even if all tables have coordinates
(default: FALSE)}
}
\value{
A \link{RelDataModel}
}
\description{
Pre-compute \link{RelDataModel} layout when missing any x or y table position
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{update_field}
\alias{update_field}
\title{Update field information in a table of a \link{RelDataModel}}
\usage{
update_field(
  x,
  tableName,
  fieldName,
  type = NULL,
  nullable = NULL,
  unique = NULL,
  comment = NULL
)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{tableName}{the name of the table to modify (a single character)}

\item{fieldName}{the name of the field to modify (a single character)}

\item{type}{the type of the field (a single character)}

\item{nullable}{if the field is nullable (a single logical)}

\item{unique}{if the values are unique (a single logical)}

\item{comment}{a description (a single character)}
}
\value{
A \link{RelDataModel}
}
\description{
Update field information in a table of a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format_confrontation_report.R
\name{view_confrontation_report}
\alias{view_confrontation_report}
\title{View confrontation report in rstudio viewer}
\usage{
view_confrontation_report(cr, ...)
}
\arguments{
\item{cr}{the confrontation report from \link{confront_data}}

\item{...}{additional params for
the \code{\link[=format_confrontation_report_md]{format_confrontation_report_md()}} function}
}
\description{
View confrontation report in rstudio viewer
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.RelDataModel.R
\name{plot.RelDataModel}
\alias{plot.RelDataModel}
\title{Plot a \link{RelDataModel} object}
\usage{
\method{plot}{RelDataModel}(x, ...)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{...}{additional parameters:
\itemize{
\item \strong{color} default table background color
\item \strong{border} border color (single character)
\item \strong{highlightBorder} color of highlighted borders
}}
}
\description{
This function draw a visNetwork of the \link{RelDataModel}.
}
\examples{
## Read the model ----
hpo_model <- read_json_data_model(
  system.file("examples/HPO-model.json", package = "ReDaMoR")
)
## Plot the model ----
plot(hpo_model)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{c.RelDataModel}
\alias{c.RelDataModel}
\title{Merge \link{RelDataModel} objects}
\usage{
\method{c}{RelDataModel}(..., checkFK = TRUE)
}
\arguments{
\item{...}{\link{RelDataModel} objects}

\item{checkFK}{a logical indicating if foreign keys should be checked
(default: TRUE)}
}
\value{
A \link{RelDataModel} objects
}
\description{
Merge \link{RelDataModel} objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{add_table}
\alias{add_table}
\title{Add a table to a \link{RelDataModel}}
\usage{
add_table(x, newTable)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{newTable}{the name of the new table or a \link{RelTableModel}}
}
\value{
A \link{RelDataModel}
}
\description{
Add a table to a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelTableModel.R
\name{RelTableModel}
\alias{RelTableModel}
\title{Create a RelTableModel object}
\usage{
RelTableModel(
  l = NULL,
  tableName,
  fields,
  primaryKey = NULL,
  foreignKeys = NULL,
  indexes = NULL,
  display = list(x = as.numeric(NA), y = as.numeric(NA), color = as.character(NA),
    comment = as.character(NA))
)
}
\arguments{
\item{l}{DEPRECATED. A named list with the function parameters.
If \code{NULL} (default) the function parameters are used. If not \code{NULL}, the
function parameters are ignored and taken from l.}

\item{tableName}{a character vector of length one}

\item{fields}{a tibble with the following columns:
\itemize{
\item \emph{name}: character
\item \emph{type}: character
\item \emph{nullable}: logical (optional, defaults to TRUE)
\item \emph{unique}: logical (optional, defaults = FALSE)
\item \emph{comment}:  character (optional, defaults to NA_character_)
}}

\item{primaryKey}{a character vector of any length. All
values should be in fields$name}

\item{foreignKeys}{a list of foreign keys. Each foreign key is defined
as a list with the following elements:
\itemize{
\item \emph{refTable}: a character vector of length one (the referenced table)
\item \emph{key}: a tibble with a "from" and a "to" columns
\item (\emph{cardinality}): an optional integer vector with 4 values:
\itemize{
\item fmin: from minimum cardinality
\item fmax: from maximum cardinality
\item tmin: to minimum cardinality
\item tmax: to maximum cardinality
}
}}

\item{indexes}{a list of indexes. Each index is defined by
3 columns:
\itemize{
\item \emph{field}: character (all in fields$name)
\item \emph{order}: character
\item \emph{unique}: logical
}}

\item{display}{a list gathering:
\itemize{
\item \emph{x}: single numeric value for the x position of the table
\item \emph{y}: single numeric value for the y position of the table
\item \emph{color}: single character value corresponding to the color of the table
\item \emph{comment}: single character value with some description of the table
}}
}
\value{
A RelTableModel object.
}
\description{
Create a RelTableModel object
}
\details{
When defining a matrix, 3 and only 3 fields must be defined:
2 of types 'row' and 'column' and the 3rd of your choice. In this case
primaryKey is defined automatically as the combination of row and column.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelTableModel.R
\name{is.MatrixModel}
\alias{is.MatrixModel}
\title{Check if the object is a \link{RelTableModel} matrix object}
\usage{
is.MatrixModel(x)
}
\arguments{
\item{x}{any object}
}
\value{
A single logical: TRUE if x is a \link{RelTableModel} matrix object
}
\description{
A matrix model is a special \link{RelTableModel} object with 3 and only 3 fields:
2 of types 'row' and 'column' and the 3rd of your choice.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{order_fields}
\alias{order_fields}
\title{Order fields in a table in a \link{RelDataModel}}
\usage{
order_fields(x, tableName, order)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{tableName}{the name of the table to modify (a single character)}

\item{order}{a vector of integers all in (1:number_of_fields)}
}
\value{
A \link{RelDataModel}
}
\description{
Order fields in a table in a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelTableModel.R
\name{col_types}
\alias{col_types}
\title{Get the types of the columns of a \link{RelTableModel} object}
\usage{
col_types(x)
}
\arguments{
\item{x}{a \link{RelTableModel} object}
}
\value{
A col_spec object with the type of each column
}
\description{
Get the types of the columns of a \link{RelTableModel} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelTableModel.R
\name{index_table}
\alias{index_table}
\title{List indexes of a \link{RelTableModel} object}
\usage{
index_table(x)
}
\arguments{
\item{x}{a \link{RelTableModel} object}
}
\value{
A tibble with the following columns:
\itemize{
\item \strong{index}: an integer corresponding to the index number
\item \strong{field}: a character corresponding to field belonging to the index
\item \strong{unique}: a logical indicating the uniqueness of the field
}
}
\description{
List indexes of a \link{RelTableModel} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generics.R
\name{get_foreign_keys}
\alias{get_foreign_keys}
\title{Get a foreign key table from an object}
\usage{
get_foreign_keys(x)
}
\arguments{
\item{x}{a \link{RelTableModel} or a \link{RelDataModel}}
}
\value{
A tibble with the following fields:
\itemize{
\item from: the origin of the key
\item ff: the key fields in from
\item to: the target of the key
\item tf: the key fields in to
\item fmin: minimum cardinality of from
\item fmax: maximum cardinality of from
\item tmin: minimum cardinality of to
\item tmax: maximum cardinality of to
}
}
\description{
Get a foreign key table from an object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.RelDataModel.R
\name{modelToVn}
\alias{modelToVn}
\title{VisNetwork representation of a \link{RelDataModel} object}
\usage{
modelToVn(
  model,
  color = "lightgrey",
  border = "black",
  highlightBorder = "orange"
)
}
\arguments{
\item{model}{a \link{RelDataModel}}

\item{color}{default table background color}

\item{border}{border color (single character)}

\item{highlightBorder}{color of highlighted borders

Internal function}
}
\description{
VisNetwork representation of a \link{RelDataModel} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/json_data_model.R
\name{write_json_data_model}
\alias{write_json_data_model}
\title{Write a data model in a JSON file}
\usage{
write_json_data_model(x, path)
}
\arguments{
\item{x}{the model to be written}

\item{path}{file on disk}
}
\description{
Write a data model in a JSON file
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_relational_data.R
\name{list_autosaved_RelDataModel}
\alias{list_autosaved_RelDataModel}
\title{List autosaved \link{RelDataModel}}
\usage{
list_autosaved_RelDataModel()
}
\description{
List autosaved \link{RelDataModel}
}
\seealso{
\code{\link[=clean_autosaved_RelDataModels]{clean_autosaved_RelDataModels()}} to clean this list.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_relational_data.R
\name{model_relational_data}
\alias{model_relational_data}
\title{Relational data modeler GUI}
\usage{
model_relational_data(
  modelInput = RelDataModel(list()),
  fromR = interactive(),
  defaultColor = "#D9D9D9",
  availableColors = c("#9BC8FE", "#F67FC4", "#C6BDF1", "#DFFB86", "#F8DEC3", "#8FE6E0",
    "#FEFE8F", "#FAC6DC", "#A9ECC9"),
  example = system.file("examples/HPO-model.json", package = utils::packageName()),
  forceIntro = FALSE
)
}
\arguments{
\item{modelInput}{the \link{RelDataModel} to start from}

\item{fromR}{a logical indicating if the application is launched from R}

\item{defaultColor}{a single color indicating the default table color}

\item{availableColors}{a character of possible colors for tables}

\item{example}{a file path to an sql or json model}

\item{forceIntro}{if TRUE the help tour start when the application
is launched (default: FALSE)}
}
\value{
The \link{RelDataModel} designed with the GUI.
}
\description{
Relational data modeler GUI
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/format_confrontation_report.R
\name{format_confrontation_report}
\alias{format_confrontation_report}
\title{Format confrontation report for printing in console}
\usage{
format_confrontation_report(cr, title = "Model")
}
\arguments{
\item{cr}{the confrontation report from \link{confront_data}}

\item{title}{a character with a single value corresponding to the report
title (e.g. database/model name)}
}
\description{
Format confrontation report for printing in console
}
\examples{
## Read the model ----
hpo_from_sql <- read_SQL_data_model(
  system.file("examples/HPO-model.sql", package = "ReDaMoR")
)
## Confront to data ----
confrontation_report <- confront_data(
  hpo_from_sql,
  path = list.files(
    system.file("examples/HPO-subset", package = "ReDaMoR"),
    full.names = TRUE
  ),
  verbose = FALSE,
  returnData = TRUE
)
## Show the report in console ----
format_confrontation_report(confrontation_report) |> cat()
## Format the report using markdown ----
format_confrontation_report_md(confrontation_report) |> cat()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelTableModel.R
\name{get_foreign_keys.RelTableModel}
\alias{get_foreign_keys.RelTableModel}
\title{Get foreign keys from \link{RelTableModel}}
\usage{
\method{get_foreign_keys}{RelTableModel}(x)
}
\arguments{
\item{x}{a \link{RelTableModel}}
}
\value{
A tibble with the following fields:
\itemize{
\item from: the origin of the key
\item ff: the key fields in from
\item to: the target of the key
\item tf: the key fields in to
\item fmin: minimum cardinality of from
\item fmax: maximum cardinality of from
\item tmin: minimum cardinality of to
\item tmax: maximum cardinality of to
}
}
\description{
Get foreign keys from \link{RelTableModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_relational_data.R
\name{recover_RelDataModel}
\alias{recover_RelDataModel}
\title{Recover an autosaved \link{RelDataModel}}
\usage{
recover_RelDataModel(name = NA)
}
\arguments{
\item{name}{The name of the autosaved \link{RelDataModel} to bring back.
Available autosaved \link{RelDataModel} can be listed using
the \code{\link[=list_autosaved_RelDataModel]{list_autosaved_RelDataModel()}}. If NA (default) the latest model is
returned.}
}
\description{
Recover an autosaved \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{[.RelDataModel}
\alias{[.RelDataModel}
\title{Subset a \link{RelDataModel}}
\usage{
\method{[}{RelDataModel}(x, i, rmForeignKeys = FALSE, ...)
}
\arguments{
\item{x}{the \link{RelDataModel} objcet}

\item{i}{the index or the names of the elements to extract}

\item{rmForeignKeys}{if TRUE, remove foreign keys which are not
available after extraction. If FALSE (default) the function will throw an
error if any foreign keys does not exist in the extracted RelDataModel.}

\item{...}{additional arguments for the \code{\link{RelDataModel}} function.}
}
\description{
Subset a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RelDataModel.R
\name{remove_table}
\alias{remove_table}
\title{Remove a table from a \link{RelDataModel}}
\usage{
remove_table(x, tableName, rmForeignKeys = FALSE)
}
\arguments{
\item{x}{a \link{RelDataModel}}

\item{tableName}{the name of the table to remove}

\item{rmForeignKeys}{if TRUE, remove foreign keys which are not
available after extraction. If FALSE (default) the function will throw an
error if any foreign keys does not exist in the extracted RelDataModel.}
}
\value{
A \link{RelDataModel}
}
\description{
Remove a table from a \link{RelDataModel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TYPEDEF.R
\name{check_types}
\alias{check_types}
\title{Check if a set of types is supported}
\usage{
check_types(x)
}
\arguments{
\item{x}{a character vector of types to be checked}
}
\description{
Check if a set of types is supported
}
